-- KILOVIEW G2 products Audio/Video front end.
local ADV7611 = require("chips.adv7611")
require( "sii902x" )
local periphery = require( "periphery" )
local hwChk = require( "kiloview.G2.hardwareCheck" )
----------------------------------------------------------------
local _M = {
	deviceType = "G2",
	deviceName = "HDMI Full-HD H.264 ENCODER",
	vendor = "KILOVIEW Electronics Co., Ltd.",
	chip = nil,
	sii902x = nil,
	currentAudioSrc = "DIGITAL",
	audSelIO = nil,
	
	lastVideoFormatVersion = 0,
	lastVideoFormat = {},
	lastAudioFormatVersion = 0,
	lastAudioFormat = {},
	analogAudioFormat = {
		_valid = true,
		audio_src = "LINE",
		sampling = 48000, --TODO: How can dynamic support?
		audio_signal_type = 3, --AKA analog
		channels = 2,
		bit_width = 16
	},
	lastOrgAudioFormat = {}
}
----------------------------------------------------------------
local function AlsaSelect( src )
	if src == "MIC" then
		os.execute( [[/usr/bin/amixer cset name='Capture Source' 1]] )
		--os.execute( [[/usr/bin/amixer cset name='Line Input Switch' off]] )
		--os.execute( [[/usr/bin/amixer cset name='Mic Input Switch' on]] )
	else
		os.execute( [[/usr/bin/amixer cset name='Capture Source' 0]] )
		--os.execute( [[/usr/bin/amixer cset name='Line Input Switch' on]] )
		--os.execute( [[/usr/bin/amixer cset name='Mic Input Switch' off]] )
	end
end
----------------------------------------------------------------
local function CompareFormatInfo( a, b )
	if a == b then
		return false
	end
	
	local k,v
	for k,v in pairs(a) do
		if v ~= b[k] then
			return true
		end
	end
	
	for k,v in pairs(b) do
		if v ~= a[k] then
			return true
		end
	end
	
	return false
end
----------------------------------------------------------------
function _M:Init()
	--1st, Check the hardware version, if >=1.7, that need create SII902x as buddy.
	hwChk.Check()
	if hwChk.hardwareVersionCode >= 0x17 then
		self.sii902x = SII902X.createNew( "/dev/i2c-0" )
		if not self.sii902x then
			print( "ERROR: Could not create HDMI loopback buddy!" )
		end
	end

	self.audSelIO = periphery.GPIO( 73, "out" )
	if not self.audSelIO then
		print( "WARNING: Could not open GPIO73 controller for audio digital/analog selection!" )
	else
		self.audSelIO:write( false )	-- 0: Digital(Default); 1: Analog
	end
	
	self.chip = ADV7611( "/dev/i2c-0", 0x98 )
	if not self.chip then
		print( "ERROR: Initialize G2 AVFE: Could not create ADV7611 controller!" )
		return false
	end
	self.chip:Reset()

	if self.sii902x then
		self.sii902x:RefreshEDID() --Try read EDID first!
		self.sii902x:SetVideoFormat( "custom" )	-- As default first.
		self.sii902x:SetAudioFormat( 48 ) -- 48KHz sample as default.
		self.sii902x:UpdateVideoSet()
		self.sii902x:UpdateAudioSet()
	end

	return true
end
----------------------------------------------------------------

function _M:StepCheck( enableSwitchNextSource )
	if self.chip then
		self.chip:Step( enableSwitchNextSource )

		if self.sii902x then
			if self.chip.infoFrame_CEA_ID ~= self.last_CEA_ID then
				self.last_CEA_ID = self.chip.infoFrame_CEA_ID
				if self.last_CEA_ID and self.last_CEA_ID >= 0 then
					self.sii902x:SetVideoFormat( self.last_CEA_ID )
				else
					self.sii902x:SetVideoFormat( "custom" )
				end
				self.sii902x:UpdateVideoSet()
			end

			if CompareFormatInfo( self.chip.orgAudioFormat, self.lastOrgAudioFormat ) then
				self.lastOrgAudioFormat = {}
				local k,v
				for k, v in pairs(self.chip.orgAudioFormat) do
					self.lastOrgAudioFormat[k] = v
				end

				self.sii902x:SetAudioFormat( self.lastOrgAudioFormat.sample and (self.lastOrgAudioFormat.sample/1000) or 48, {
					mclk_fs = self.lastOrgAudioFormat.fs or 256
				})
				self.sii902x:UpdateAudioSet()
			end

			self.sii902x:Poll()
		else --No sii902x
			if CompareFormatInfo( self.chip.orgAudioFormat, self.lastOrgAudioFormat ) then
				self.lastOrgAudioFormat = {}
				local k,v
				for k, v in pairs(self.chip.orgAudioFormat) do
					self.lastOrgAudioFormat[k] = v
				end
			end
		end

		if CompareFormatInfo( self.chip.videoFormat, self.lastVideoFormat ) then	--Different
			self.lastVideoFormatVersion = self.lastVideoFormatVersion + 1
			self.lastVideoFormat = {}
			local k,v
			for k,v in pairs(self.chip.videoFormat) do
				self.lastVideoFormat[k] = v
			end
		end
		
		if CompareFormatInfo( self.chip.audioFormat, self.lastAudioFormat) then -- Different audio
			self.lastAudioFormatVersion = self.lastAudioFormatVersion + 1
			self.lastAudioFormat = {}
			local k,v
			for k, v in pairs(self.chip.audioFormat) do
				self.lastAudioFormat[k] = v
			end
		end
	end
end
----------------------------------------------------------------
function _M:ReportDetectingStatus( srcType )
	if srcType == "video" then
		return true, "HDMI"
	elseif srcType == "audio" then
		return true, self.currentAudioSrc
	end
end
----------------------------------------------------------------
function _M:ReportVideoFormat( curVersion )
	--if not curVersion or curVersion ~= self.lastVideoFormatVersion then
		return self.lastVideoFormatVersion, self.lastVideoFormat
	--else
	--	return self.lastVideoFormatVersion, nil	--Don't return format
	--end
end
----------------------------------------------------------------
function _M:ReportAudioFormat( curVersion )
	if self.savedAudioSrc ~= self.currentAudioSrc then
		self.savedAudioSrc = self.currentAudioSrc
		self.lastAudioFormatVersion = self.lastAudioFormatVersion + 1
		self.analogAudioFormat.audio_src = self.currentAudioSrc
	end

	if self.currentAudioSrc == "DIGITAL" then
		return self.lastAudioFormatVersion, self.lastAudioFormat
	else
		return self.lastAudioFormatVersion, self.analogAudioFormat
	end
end
----------------------------------------------------------------
function _M:SelectVideoSource( src )
	return "HDMI"
end
----------------------------------------------------------------
local function PatchI2SWordLen(len)
	if hwChk.hardwareVersionSwCode < 0x20 then
		local file = io.open( "/proc/k2e_audio/wlen", "w" )
		if file then
			file:write( "" .. len )
			file:close()
		end
	end
end

function _M:SelectAudioSource( src )
	--TODO: Select audio source (SDI digital/Analog line-in/Analog Mic-in)
	src = src:upper()
	if src == "DIGITAL" or src == "HDMI" or src == "AUTO" or src == "DEFAULT" then
		self.currentAudioSrc = "DIGITAL"
		if self.audSelIO then
			self.audSelIO:write( true )		-- 1: Digital
		end
		PatchI2SWordLen(32)
	elseif src == "ANALOG" or src == "LINE" then
		self.currentAudioSrc = "LINE"
		if self.audSelIO then
			self.audSelIO:write( false )		-- 0: Analog
		end
		AlsaSelect( "LINE" )
		PatchI2SWordLen(64)
	elseif src == "MIC" then
		self.currentAudioSrc = "MIC"
		if self.audSelIO then
			self.audSelIO:write( false )		-- 0: Analog
		end
		AlsaSelect( "MIC" )
		PatchI2SWordLen(64)
	end
	return self.currentAudioSrc
end
----------------------------------------------------------------
function _M:SetVideoParameter( pmName, pmValue, src )
	--Nothing
end
----------------------------------------------------------------
function _M:SetAudioParameter( pmName, pmValue, src )
	if not src then
		src = self.currentAudioSrc
	end
	
	if src == "DIGITAL" or src == "HDMI" then
		return --Nothing
	end
	
	if pmName == "gain" then	--Adjust analog audio gain
		if src == "ANALOG" or src == "LINE" then
			pmValue = tonumber(pmValue)	--As DB
			local dbSteps = math.floor(pmValue/0.75)	--For AIC23 chip, value 1 means 0.75db
			if dbSteps > 8 then
				dbSteps = 8
			elseif dbSteps < -23 then
				dbSteps = -23
			end
			
			local dbVal = 23 + dbSteps
			os.execute( [[/usr/bin/amixer cset name='Line Input Volume' ]] .. dbVal )
		elseif src == "MIC" then
			pmValue = tonumber(pmValue)
			if pmValue < 3 then	-- <3db, as default mic gain (0db)
				os.execute( [[/usr/bin/amixer cset name='Mic Booster Switch' off]] )
			else	-- >=3db, Mic boost (so with 20db gain)
				os.execute( [[/usr/bin/amixer cset name='Mic Booster Switch' on]] )
			end
		end
	elseif pmName == "line_bypass" then
		if pmValue == 0 then
			os.execute( [[/usr/bin/amixer cset name='Output Mixer Line Bypass Switch' off]] )
		else
			os.execute( [[/usr/bin/amixer cset name='Output Mixer Line Bypass Switch' on]] )
		end
	end
end
----------------------------------------------------------------
function _M:GetSupportVideoSources()
	return { ["HDMI"] = "HDMI Video Input" }
end
----------------------------------------------------------------
function _M:GetSupportAudioSources()
	return { ["HDMI"] = "HDMI Digital Audio", ["LINE"] = "Analog Line Input" --[[, ["MIC"] = "Analog MIC Input"]] }
end
----------------------------------------------------------------

return _M

