-----------------------------------------------------------------------------------------------------
-- ADV7611 chip controller.
-- Copyright (C) 2016, KILOVIEW Elecontrics Co., Ltd.
--
-- Author: 	Jacob Zuo, 2016/6/4
-----------------------------------------------------------------------------------------------------

local periphery = require "periphery"
local edid = require "edid"	--EDID
local I2C = periphery.I2C
local GPIO = periphery.GPIO
--For debug only.
--local pretty = require "pl.pretty"

local MAIN_FREQ = 28636360

local hdtv_formats = require "hdtv_formats"
--local vga_formats = require "vga_formats"

hdtv_formats.SetMainFreq( MAIN_FREQ )
--vga_formats.SetMainFreq( MAIN_FREQ )

--Init both HDTV and VGA format standards
hdtv_formats.Init( "STDI" )
--vga_formats.Init()

--[[
local CLOCK_PHASE = 0x97
local config_phase_file = io.open("/etc/hdmi_phase.conf","r")
if config_phase_file then
	local l = config_phase_file:read("*l")
	config_phase_file:close()
	if l then
		CLOCK_PHASE = tonumber(l) or 0x85
	end
end
--]]

------------------------------------------------------
-- The factory function to create a ADV7611 handle object
------------------------------------------------------
local function CreateADV7611( i2c_dev, baseAddr )
	i2c_dev = i2c_dev or "/dev/i2c-0"
	
	local BASE_MAP = baseAddr or 0x98
	local CEC_MAP=0x80
	local INFOFRAME_MAP=0x7c
	local DPLL_MAP=0x4c
	local CP_MAP=0x44
	local HDMI_MAP=0x68
	local KSV_MAP=0x64
	local EDID_MAP=0x6c

	print( "Creating ADV7611 control object at 0x98@" .. i2c_dev .. "..." )

	------------------------------------------------------
	-- Object
	------------------------------------------------------
	local ADV7611 = {
		currentSource = "HDMI",
		currentSignal = false,
		freeRun = false,
		currentFormat = nil,
		forceReassignFormat = false,
		lastFormatRecord = {},
		videoFormat = { _valid = false },
		audioFormat = { _valid = false, audio_src = "HDMI", audio_signal_type = 0 }, --type 0 aka. HDMI
		orgAudioFormat = { _valid = false },
		nostandardFormatInfo = {},
		i2c_handle = nil,
	}

	-------------------------------------------------------
	-- I2C and GPIO openning
	-------------------------------------------------------
	local done,err = pcall( function() ADV7611.i2c_handle = I2C( i2c_dev ) end )
	if not done then
		if type(err) == "table" then
			print( "ERROR: Could not open I2C device for ADV7611 controlling. ErrMsg:", err.message or "---" )
		else
			print( "ERROR: Could not open I2C device for ADV7611 controlling. ErrMsg:", err )
		end
		return nil, err
	end
	
	-------------------------------------------------------
	-- Object method: Register read
	-------------------------------------------------------
	function ADV7611:ReadReg( map, addr, itemArr )
		if not self.i2c_handle then
			return nil, "Invalid I2C handle"
		end
		if not itemArr or type(itemArr) ~= "table" then
			itemArr = { 0xff, flags = I2C.I2C_M_RD }
		else
			if #itemArr < 1 then
				itemArr[1] = 0xff
			end
			if not itemArr.flags then
				itemArr.flags = I2C.I2C_M_RD
			else
				itemArr.flags = itemArr.flags | I2C.I2C_M_RD
			end
		end

		local msg = { {addr}, itemArr }
		local done,err = pcall( function() self.i2c_handle:transfer( map>>1, msg ) end )
		if not done then
			if type(err) == "table" then
				--print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", map, addr), err.message or "---" )
				return nil, err.message or "---"
			else
				--print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", map, addr), err )
				return nil, err
			end
		else
			return itemArr
		end
	end

	------------------------------------------------------
	-- Register write
	------------------------------------------------------
	function ADV7611:WriteReg( map, addr, itemArr )
		if not self.i2c_handle then
			return false, "Invalid I2C handle"
		end
		if not itemArr or type(itemArr) ~= "table" or #itemArr < 1 then
			return false, "Invalid data"
		end

		table.insert( itemArr, 1, addr )
		local msg = { itemArr }
		local done,err = pcall( function() self.i2c_handle:transfer( map>>1, msg ) end )
		if not done then
			if type(err) == "table" then
				--print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", map, addr), err.message or "---" )
				return nil, err.message or "---"
			else
				--print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", map, addr), err )
				return nil, err
			end
		else
			return true
		end
	end

	------------------------------------------------------
	
	function ADV7611:_FillRegBT656()
		--CP Map
		--44 6C 00 ; ADI recommended setting
		self:WriteReg( CP_MAP, 0x6c, {0x00} )

		--self:WriteReg( CP_MAP, 0xba, {0x00} )

		self:WriteReg( HDMI_MAP, 0x00, {0x00} )
		--Enable clock terminator for port A
		self:WriteReg( HDMI_MAP, 0x83, {0xfe} )
		--HDMI Map
		--68 6F 0C ; ADI recommended setting
		--68 85 1F ; ADI recommended setting
		--68 87 70 ; ADI recommended setting
		--68 57 DA ; ADI recommended setting
		--68 58 01 ; ADI recommended setting
		--68 03 98 ; Set DIS_I2C_ZERO_COMPR 0x03[7]=1
		--68 4C 44 ; Set NEW_VS_PARAM 0x44[2]=1
		self:WriteReg( HDMI_MAP, 0x6f, {0x0c} )
		self:WriteReg( HDMI_MAP, 0x85, {0x1f} )
		self:WriteReg( HDMI_MAP, 0x87, {0x70} )
		self:WriteReg( HDMI_MAP, 0x57, {0xda} )
		self:WriteReg( HDMI_MAP, 0x58, {0x01} )
		self:WriteReg( HDMI_MAP, 0x03, {0x98} )
		self:WriteReg( HDMI_MAP, 0x4c, {0x44} )

	    self:WriteReg( HDMI_MAP, 0x96, {0x01} )

		--ADI recommended setting
		--0x68 0x9b 0x03
		self:WriteReg( HDMI_MAP, 0x9b, {0x03} )

		--Following writes are recommended for non-fast switching applications:
		--68 C1 01 ; ADI recommended setting
		--68 C2 01 ; ADI recommended setting
		--68 C3 01 ; ADI recommended setting
		--68 C4 01 ; ADI recommended setting
		--68 C5 01 ; ADI recommended setting
		--68 C6 01 ; ADI recommended setting
		--68 C7 01 ; ADI recommended setting
		--68 C8 01 ; ADI recommended setting
		--68 C9 01 ; ADI recommended setting
		--68 CA 01 ; ADI recommended setting
		--68 CB 01 ; ADI recommended setting
		--68 CC 01 ; ADI recommended setting
		self:WriteReg( HDMI_MAP, 0xc1, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc2, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc3, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc4, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc5, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc6, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc7, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc8, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc9, {0x01} )
		self:WriteReg( HDMI_MAP, 0xca, {0x01} )
		self:WriteReg( HDMI_MAP, 0xcb, {0x01} )
		self:WriteReg( HDMI_MAP, 0xcc, {0x01} )

		--Video input standard
		--self:WriteReg( BASE_MAP, 0x00, {0x02} )   --Default
		--PRIM_MODE
	    --self:WriteReg( BASE_MAP, 0x01, {0x06} )

	    --#Address 0x02:
	    --#INP_COLOR_SPACE[3:0] : 1111----
	    --#ALT_GAMMA            : ----0---
	    --#OP_656_RANGE         : -----1-- ( 0=0-255, 1=16-235 )
	    --#RGB_OUT              : ------0- ( 0=YUV, 1=RGB )
	    --#ALT_DATA_SAT         : -------0
	    self:WriteReg( BASE_MAP, 0x02, {0xf4} )
	    --#OP_FORMAT_SEL(0x03): 80=16bit BT.656 MODE.
	    self:WriteReg( BASE_MAP, 0x03, {0x80} )
	    --#Output Signal(HS/VS/FIELD)
	    self:WriteReg( BASE_MAP, 0x06, {0xa0} )
	    --#POWER_DOWN(0x0c): 01 [1 = power_down, 0 = power_on ] 00010
	    self:WriteReg( BASE_MAP, 0x0c, {0x42} )
	    --#98 14 7F ; Max Drive Strength
	    --self:WriteReg( BASE_MAP, 0x14, {0x7f} )
		self:WriteReg( BASE_MAP, 0x14, {0x6a} )

	    --#98 15 80 ; Disable Tristate of Pins
	    self:WriteReg( BASE_MAP, 0x15, {0x80} )
	end
	
	------------------------------------------------------
	function ADV7611:_FillRegBT601()	
		return self:_FillRegBT656()
--[[		
		--CP Map
		--44 6C 00 ; ADI recommended setting
		self:WriteReg( CP_MAP, 0x6c, {0x00} )

		--self:WriteReg( CP_MAP, 0xba, {0x00} )

		self:WriteReg( HDMI_MAP, 0x00, {0x00} )
		--Enable clock terminator for port A
		self:WriteReg( HDMI_MAP, 0x83, {0xfe} )
		--HDMI Map
		--68 6F 0C ; ADI recommended setting
		--68 85 1F ; ADI recommended setting
		--68 87 70 ; ADI recommended setting
		--68 57 DA ; ADI recommended setting
		--68 58 01 ; ADI recommended setting
		--68 03 98 ; Set DIS_I2C_ZERO_COMPR 0x03[7]=1
		--68 4C 44 ; Set NEW_VS_PARAM 0x44[2]=1
		self:WriteReg( HDMI_MAP, 0x6f, {0x0c} )
		self:WriteReg( HDMI_MAP, 0x85, {0x1f} )
		self:WriteReg( HDMI_MAP, 0x87, {0x70} )
		self:WriteReg( HDMI_MAP, 0x57, {0xda} )
		self:WriteReg( HDMI_MAP, 0x58, {0x01} )
		self:WriteReg( HDMI_MAP, 0x03, {0x98} )
		self:WriteReg( HDMI_MAP, 0x4c, {0x44} )

	    self:WriteReg( HDMI_MAP, 0x96, {0x01} )

		--ADI recommended setting
		--0x68 0x9b 0x03
		self:WriteReg( HDMI_MAP, 0x9b, {0x03} )

		--Following writes are recommended for non-fast switching applications:
		--68 C1 01 ; ADI recommended setting
		--68 C2 01 ; ADI recommended setting
		--68 C3 01 ; ADI recommended setting
		--68 C4 01 ; ADI recommended setting
		--68 C5 01 ; ADI recommended setting
		--68 C6 01 ; ADI recommended setting
		--68 C7 01 ; ADI recommended setting
		--68 C8 01 ; ADI recommended setting
		--68 C9 01 ; ADI recommended setting
		--68 CA 01 ; ADI recommended setting
		--68 CB 01 ; ADI recommended setting
		--68 CC 01 ; ADI recommended setting
		self:WriteReg( HDMI_MAP, 0xc1, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc2, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc3, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc4, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc5, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc6, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc7, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc8, {0x01} )
		self:WriteReg( HDMI_MAP, 0xc9, {0x01} )
		self:WriteReg( HDMI_MAP, 0xca, {0x01} )
		self:WriteReg( HDMI_MAP, 0xcb, {0x01} )
		self:WriteReg( HDMI_MAP, 0xcc, {0x01} )

		--Video input standard
		--self:WriteReg( BASE_MAP, 0x00, {0x02} )   --Default
		--PRIM_MODE
	    --self:WriteReg( BASE_MAP, 0x01, {0x06} )

	    --#Address 0x02:
	    --#INP_COLOR_SPACE[3:0] : 1111----
	    --#ALT_GAMMA            : ----0---
	    --#OP_656_RANGE         : -----1-- ( 0=0-255, 1=16-235 )
	    --#RGB_OUT              : ------0- ( 0=YUV, 1=RGB )
	    --#ALT_DATA_SAT         : -------0
	    self:WriteReg( BASE_MAP, 0x02, {0xf4} )
	    --#OP_FORMAT_SEL(0x03): 80=16bit BT.656 MODE.
	    self:WriteReg( BASE_MAP, 0x03, {0x80} )
	    --#Output Signal(HS/VS/FIELD)
	    self:WriteReg( BASE_MAP, 0x06, {0xa0} )
	    --#POWER_DOWN(0x0c): 01 [1 = power_down, 0 = power_on ] 00010
	    self:WriteReg( BASE_MAP, 0x0c, {0x42} )
	    --#98 14 7F ; Max Drive Strength
	    --self:WriteReg( BASE_MAP, 0x14, {0x7f} )
		self:WriteReg( BASE_MAP, 0x14, {0x6a} )

	    --#98 15 80 ; Disable Tristate of Pins
	    self:WriteReg( BASE_MAP, 0x15, {0x80} )		
	]]
		
	end

	------------------------------------------------------
	-- Soft reset the chip!
	------------------------------------------------------
	function ADV7611:Reset()
		self:_Reset()
	end

	function	ADV7611:_Reset()	--Soft reset
		self:WriteReg( BASE_MAP, 0xff, {0x80, flags=I2C.I2C_M_IGNORE_NAK} )
		local i, powerup
		powerup = false
		for i=1,100 do
			periphery.sleep_ms(10)
			local rslt = self:ReadReg( BASE_MAP, 0xff, { 0xff } )
			if rslt and (rslt[1] & 0x80) == 0 then
				powerup = true
				print( "Chip reset DONE." )
				break
			end
		end
	
		if not powerup then
			print( "***ERROR*** This chip looks could not power up after reset!" )
		end

		periphery.sleep_ms(100)

		self:WriteReg( BASE_MAP, 0xf4, {CEC_MAP} )
		self:WriteReg( BASE_MAP, 0xf5, {INFOFRAME_MAP} )
		self:WriteReg( BASE_MAP, 0xf8, {DPLL_MAP} )
		self:WriteReg( BASE_MAP, 0xf9, {KSV_MAP} )
		self:WriteReg( BASE_MAP, 0xfa, {EDID_MAP} )
		self:WriteReg( BASE_MAP, 0xfb, {HDMI_MAP} )
		self:WriteReg( BASE_MAP, 0xfd, {CP_MAP} )
		
		self:_InitInterrupts()
		self:_InitEDID()	--Init EDID
		
		self:WriteReg( CP_MAP, 0x3e, {0x80} )	--Enable color adjust.
		self:_FillRegBT656()
		self:_ConfigureHDMIFormat(nil, true)	--Force default blue
		
		self.currentSignal = false
		self.freeRun = false
		self.currentFormat = nil
		self.forceReassignFormat = false
		self.lastFormatRecord = {}		--Clear recorded format detection informations.
		--self.videoFormat = {_valid = false}
		self.audioFormat = { _valid = false, audio_src = "HDMI", audio_signal_type = 0 }
		self.orgAudioFormat = { _valid = false }
		--NOTE: Because of free run, create a 720p60 format.
		self.videoFormat = {_valid = true}
		local formatInfo = hdtv_formats.Standard_Match( 4 )
		self:_RecordVideoFormatInfo( "HDMI", formatInfo )
		self.lastFormatRecord.infoFrameId = 4
	end

	------------------------------------------------------
	-- Init the ADV7611 interrupts 
	------------------------------------------------------
	
	function ADV7611:_InitInterrupts()
		--TODO: Nothing yet.
	end
	
	------------------------------------------------------
	--EDID configuration
	------------------------------------------------------
	function ADV7611:_InitEDID()
		--[7 6 5 4] -> Detect +5V to assert HPA delay
		--         [3] -> HPA_OVR_TERM
		--            [ 2 1 ] -> 00: only EDID; 01: only +5V; 10: EDID&+5V; 11: EDID&+5V&Manual
		--					  [0] -> 0 : auto HPA, 1 : Manual.
		self:WriteReg( HDMI_MAP, 0x6c, {0x34} );
		
		self:WriteReg( KSV_MAP, 0x77, {0} );	 --Why? Ref. the sample script, but no documented.
		self:WriteReg( KSV_MAP, 0x74, {0} );	 --Disable internal EDID accessing, by document.
		
		local i
		for i=1, #edid, 8 do
			self:WriteReg( EDID_MAP, i-1, { edid[i], edid[i+1], edid[i+2], edid[i+3], edid[i+4], edid[i+5], edid[i+6], edid[i+7] 
															  } )
		end
		
		self:WriteReg( KSV_MAP, 0x74, { 1 } )
		
		periphery.sleep_ms(5)		--Wait 5ms for Calc CRC
		
--[[		
		for i=1, 256, 8 do
			local v = self:ReadReg( EDID_MAP, i-1, { 0, 0, 0, 0, 0, 0, 0, 0 } )
			print( string.format( "%02x %02x %02x %02x %02x %02x %02x %02x", table.unpack(v) ) )
		end
]]		
	end
	
	function ADV7611:_CheckSignal(src)
		--NOTE: HDMI: 0x04[1] - PLL lock
		local freeRun = false
		local rslt = self:ReadReg( CP_MAP, 0xff, {0x00} )
		if rslt and (rslt[1]&0x10 ~= 0) then
			freeRun = true
			--print( "FREE RUN")
		end
		
		rslt = self:ReadReg( HDMI_MAP, 0x04, {0x00} )
		if rslt then
			if (rslt[1] & 0x02) ~= 0 then
				--NOTE: HDMI:0x07[7|5] - VERT_FILTER_LOCKED | DE_REGEN_FILTER_LOCKED
				rslt = self:ReadReg( HDMI_MAP, 0x07, {0x00} )
				if rslt and (rslt[1]&0x20) ~= 0 then
					return true, freeRun
				else
					return false, freeRun
				end
			else
				return false, freeRun
			end
		else
			return false, freeRun
		end
	end

	------------------------------------------------------
	
	function ADV7611:ForceUpdate()
		self.lastFormatRecord = {}		--Clear the recorded format information.
	end

	------------------------------------------------------
	-- A step handler for ADV7611 processing.
	-- To fit the LUA working mode, you need call the 
	-- 'Step' periodically ( suggest per 200~500ms).
	-- This routine should do :
	--	* Signal status checking
	--	*	Auto format detecting
	--	*	Configue the chip to make it output BT.656/BT.601 signals
	--
	------------------------------------------------------
	
	function ADV7611:Step()
		self.currentSignal, self.freeRun = self:_CheckSignal()	--Check current signal
		if self.currentSignal then
			self.videoFormat.signal_locked = true	--Assume it be, but when detecting, it may changed
			self.videoFormat.free_run = false
			self:_DetectCurrentFormat()
			self:_ConfigureCurrentFormat()
		elseif self.freeRun then
			self.videoFormat.signal_locked = true
			self.videoFormat.free_run = true
		else
			self.videoFormat.signal_locked = false
			self.videoFormat.free_run = false
		end
	end


	------------------------------------------------------
	-- Detect current signal format
	------------------------------------------------------
	
	function ADV7611:_DetectCurrentFormat(src)
		self:_DetectHDMIAudio(src)
		return self:_DetectHDMIFormat(src)
	end
	
	------------------------------------------------------
	-- Configure signal format
	------------------------------------------------------
	
	function ADV7611:_ConfigureCurrentFormat(src)
		if not self.videoFormat._lastFormatChanged and not self.forceReassignFormat then
			return true
		end
		
		return self:_ConfigureHDMIFormat(src)
	end
	
	------------------------------------------------------
	-- According detected format information, record it to Amba based formation context.
	------------------------------------------------------
	
	function ADV7611:_RecordVideoFormatInfo(src, formatInfo)
		if type(formatInfo) ~= "table" then
			return false
		end

		--pretty.dump( formatInfo )
		
		if self.videoFormat._formatInfo ~= formatInfo or (src and src ~= self.videoFormat.video_src) then
			self.videoFormat._lastFormatChanged = true
			self.videoFormat._formatInfo = formatInfo
			
			if src then
				self.videoFormat.video_src = src
			end
			
			--TODO:
			self.videoFormat.cap_start_x = 0
			self.videoFormat.cap_start_y = 0
			
			self.videoFormat.cap_cap_w = formatInfo.hAct
			self.videoFormat.line_width = formatInfo.hAct
			
			self.videoFormat.total_width = formatInfo.hTotal
			
			if formatInfo.vAct then
				self.videoFormat.cap_cap_h = formatInfo.vAct
				self.videoFormat.total_height = formatInfo.vTotal
				
				if formatInfo.interlaced then
					self.videoFormat.f0_height = formatInfo.vAct/2
					self.videoFormat.f1_height = formatInfo.vAct/2
				else
					self.videoFormat.f0_height = formatInfo.vAct
					self.videoFormat.f1_height = formatInfo.vAct
				end
			elseif formatInfo.vAct0 then
				self.videoFormat.f0_height = formatInfo.vAct0
				if formatInfo.interlaced then
					if formatInfo.vAct1 then
						self.videoFormat.f1_height = formatInfo.vAct1
						self.videoFormat.cap_cap_h = formatInfo.vAct0 + formatInfo.vAct1
						if self.videoFormat.cap_cap_h >= 486 and self.videoFormat.cap_cap_h <= 488 then
							self.videoFormat.cap_cap_h = 480	--This is a special format
						else
							self.videoFormat.cap_cap_h = self.videoFormat.cap_cap_h & 0xfffffffe
						end
						self.videoFormat.total_height = self.videoFormat.cap_cap_h	--??? total_height for interlace?
					else
						self.videoFormat.f1_height = formatInfo.vAct0
						self.videoFormat.cap_cap_h = formatInfo.vAct0 * 2
						self.videoFormat.total_height = formatInfo.vAct0 * 2
					end
				else
					self.videoFormat.f1_height = formatInfo.vAct0
					self.videoFormat.cap_cap_h = formatInfo.vAct0
					self.videoFormat.total_height = formatInfo.vTotal0
				end
			else
				self.videoFormat.f0_height = 0
				self.videoFormat.f1_height = 0
				self.videoFormat.cap_cap_h = 0	--This is wrong.
				self.videoFormat.total_height  = 0
			end
			
			--TODO:
			self.videoFormat.sync_start = 0
			
			if formatInfo.video_system then
				self.videoFormat.video_system = formatInfo.video_system
			else
				self.videoFormat.video_system = _G.AMBA_VIDEO_SYSTEM_AUTO or 0x0
			end
			
			if AMBA_VIDEO_FPS then
				self.videoFormat.frame_rate = AMBA_VIDEO_FPS(formatInfo.frameRate)
			else
				self.videoFormat.frame_rate = AMBA_VIDEO_FPS_60 or formatInfo.frameRate
			end
			
			self.videoFormat.frame_rate_value = formatInfo.frameRate
			
			if formatInfo.aspect_ratio then
				self.videoFormat.aspect_ratio = formatInfo.aspect_ratio
			else
				self.videoFormat.aspect_ratio = _G.AMBA_VIDEO_RATIO_AUTO or 0
			end
			
			--TODO: Here, need fix this to support 8 bits/ BT.601 options.
			self.videoFormat.input_type = _G.AMBA_VIDEO_TYPE_YUV_656 or 2
			self.videoFormat.bit_resolution = 16
			if formatInfo.interlaced then
				self.videoFormat.video_format = _G.AMBA_VIDEO_FORMAT_INTERLACE or 1
				self.videoFormat.input_format = _G.AMBA_VIN_INPUT_FORMAT_YUV_422_INTLC or 1
				self.videoFormat.interlaced = 1
			else
				self.videoFormat.video_format = _G.AMBA_VIDEO_FORMAT_PROGRESSIVE or 2
				self.videoFormat.input_format = _G.AMBA_VIN_INPUT_FORMAT_YUV_422_PROG or 2
				self.videoFormat.interlaced = 0
			end
			
			--TODO:
			self.videoFormat.hs_front_porch = 0
			self.videoFormat.hs_pulse_width = 0
			self.videoFormat.hs_back_porch = 0
			--TODO:
			self.videoFormat.f0_vs_front_porch = 0
			self.videoFormat.f0_vs_pulse_width = 0
			self.videoFormat.f0_vs_back_porch = 0
			--TODO:
			self.videoFormat.f1_vs_front_porch = 0
			self.videoFormat.f1_vs_pulse_width = 0
			self.videoFormat.f1_vs_back_porch = 0
		
			if formatInfo.tmdsFreq and formatInfo.tmdsFreq > 1000000 then
				self.videoFormat.pixel_freq = formatInfo.tmdsFreq 
			elseif formatInfo.pixelFreq and formatInfo.pixelFreq > 1000000 then
				self.videoFormat.pixel_freq = formatInfo.pixelFreq
			else
				self.videoFormat.pixel_freq = 0
			end
			
			self.videoFormat.format_code = formatInfo.formatId	-- Make the formatId as format-code.
			self.videoFormat.format_name = formatInfo.formatName or (self.videoFormat.cap_cap_w .. "x" .. self.videoFormat.cap_cap_h)
			--self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_DIGITAL or 3
			return true
		else
			return true	--No changed.
		end
	end

	------------------------------------------------------
	-- Detect HDMI Audio
	------------------------------------------------------
	function ADV7611:_DetectHDMIAudio(src)
		self.orgAudioFormat.withAudio = false
		self.audioFormat._valid = false
		local rslt
		rslt = self:ReadReg( HDMI_MAP, 0x04, {0x00} )
		
		if rslt and (rslt[1] & 0x01) ~= 0 then
			self.orgAudioFormat.pll_locked = true
		else
			self.orgAudioFormat.pll_locked = false
			return
		end


		rslt = self:ReadReg( BASE_MAP, 0x65, {0x00} )
		if rslt and (rslt[1]&0x80) ~= 0 then
			self.orgAudioFormat.withAudio = true
			self.audioFormat._valid = true
		else
			return
		end

		rslt = self:ReadReg( DPLL_MAP, 0xb5, {0x00} )
		if rslt then
			self.orgAudioFormat.fs = (rslt[1]+1)*128
			self.audioFormat.fs = self.orgAudioFormat.fs
		end

		rslt = self:ReadReg( HDMI_MAP, 0x36, {0x00,0x00,0x00,0x00,0x00} )
		if rslt then
			local sampleFlag = rslt[4] & 0x0f
			if sampleFlag == 0 then
				self.orgAudioFormat.sample = 44100
			elseif sampleFlag == 2 then
				self.orgAudioFormat.sample = 48000
			elseif sampleFlag == 3 then
				self.orgAudioFormat.sample = 32000
			elseif sampleFlag == 8 then
				self.orgAudioFormat.sample = 88200
			elseif sampleFlag == 9 then
				self.orgAudioFormat.sample = 768000
			elseif sampleFlag == 10 then
				self.orgAudioFormat.sample = 96000
			elseif sampleFlag == 12 then
				self.orgAudioFormat.sample = 176000
			elseif sampleFlag == 14 then
				self.orgAudioFormat.sample = 192000
			else
				self.orgAudioFormat.sample = 48000
			end	
			self.audioFormat.sampling = self.orgAudioFormat.sample

			if (rslt[5]&0x01) ~= 0 then
				self.orgAudioFormat.maxSampleSize = 20
			else
				self.orgAudioFormat.maxSampleSize = 24
			end

			--TODO:
			self.audioFormat.channels = 2
			self.audioFormat.bit_width = 16
		end
	end

	------------------------------------------------------
	-- Detect HDMI AVI info frame
	------------------------------------------------------
	
	function ADV7611:_DetectHDMIInfoFrame(src)
		local rslt
		self.infoFrameChecked = false
		self.infoFrame_CEA_ID = nil

		rslt = self:ReadReg( BASE_MAP, 0x60, {0x00} )	--HDMI Raw status
		if rslt and (rslt[1] & 0x01) ~= 0 then	--AVI Info detected
			--IO:0x88[4] - The AVI checksum error status
			rslt = self:ReadReg( BASE_MAP, 0x88, {0x00} )
			if not rslt then
				return false
			--TODO: else if (rslt[1]&0x10) ~= 0 then ...
			end
			
			local i
			local veri_ok = false
			for i=1,3 do
				rslt = self:ReadReg( INFOFRAME_MAP, 0xe1, {0x00} )	--Read info-frame version
				if rslt and rslt[1] >= 2 then
					veri_ok = true
					break
				end
			end
			
			if not veri_ok then
				return false
			end
			
			self.infoFrameChecked = true
			rslt = self:ReadReg( INFOFRAME_MAP, 0x04, {0x00} )	--Read AVIINFODATA4 (id code)
			if rslt then
				local idCode = rslt[1] & 0x7f
				self.infoFrame_CEA_ID = idCode

				if idCode ~= self.lastFormatRecord.infoFrameId then
					local formatInfo = hdtv_formats.Standard_Match( idCode )
					if formatInfo then
						if self:_RecordVideoFormatInfo( src, formatInfo ) then
							self.lastFormatRecord.infoFrameId = idCode
							return true
						else
							return false
						end
					else --of matched.
						return false
					end
				else --Last record is this one, so no do again.
					return true
				end
			else
				return false
			end
		else
			return false
		end
	end
	
	------------------------------------------------------
	-- STDI checking
	------------------------------------------------------
	
	function ADV7611:_STDICheck()
		local interlaced = false	
		local rslt = self:ReadReg( CP_MAP, 0xb1, {0x00, 0x00} )
		local  BL = rslt[1] & 0x3f
		BL = ( BL << 8) + rslt[2]

		rslt = self:ReadReg( CP_MAP, 0xb3, {0x00} )
		local LCVS = (rslt[1] & 0xf8) >> 3

		rslt = self:ReadReg( CP_MAP, 0xa3, {0x00, 0x00} )
		local LCF = rslt[1] & 0x0f
		LCF = (LCF << 8) + rslt[2]

		rslt = self:ReadReg( CP_MAP, 0xb8, {0x00, 0x00} )
		local FCL = (rslt[1] & 0x1f) << 8
		FCL =  FCL +rslt[2]
		
		rslt = self:ReadReg( CP_MAP, 0xb1, {0x00} )
		interlaced = ((rslt[1]&0x40) ~= 0)
		
		self.videoFormat._DebugInfo = "E:" ..  BL .. "/" .. LCVS .. "/" .. LCF .. "/" .. FCL ..  (interlaced and "/I" or "/P")
		--print( self.videoFormat._DebugInfo )
		return true, BL, LCF, FCL, LCVS, interlaced
	end
	
	------------------------------------------------------
	-- Assume HDMI customized format
	------------------------------------------------------
	function ADV7611:_AssumeHDMICustomFormat(src, BL, LCF, FCL, LCVS, interlaced)
		local refreshFreq = 0
		if FCL and FCL > 0 then
			refreshFreq = MAIN_FREQ/(FCL*256)
			if refreshFreq <= 24 then
				refreshFreq = 24
			elseif refreshFreq <= 26 then
				refreshFreq = 25
			elseif refreshFreq <= 31 then
				refreshFreq = 30
			elseif refreshFreq <=51 then
				refreshFreq = 50
			elseif refreshFreq <=61 then
				refreshFreq = 60
			elseif refreshFreq <=71 then
				refreshFreq = 70
			elseif refreshFreq <= 76 then
				refreshFreq = 75
			elseif refreshFreq <= 85 then
				refreshFreq = 85
			end	--Else ignore restrict.
		end
		
		if refreshFreq <= 0 then
			return false
		end
		
		local rslt = self:ReadReg( HDMI_MAP, 0x51, {0x00, 0x00} )
		local tmdsFreq = rslt[1]*2
		if (rslt[2]&0x80) ~= 0 then
			tmdsFreq = tmdsFreq + 1
		end
		
		tmdsFreq = tmdsFreq * 1000000

		local actW = 0
		local totalW = 0
		local actH = 0
		local totalH = 0
		local actH2 = 0
		local totalH2 = 0
		
		rslt = self:ReadReg( HDMI_MAP, 0x07, {0x00,0x00} )
		if rslt then
			actW = ((rslt[1]&0x0f) << 8) + rslt[2]
		end
		
		rslt = self:ReadReg( HDMI_MAP, 0x1e, {0x00, 0x00} )
		if rslt then
			totalW = ((rslt[1]&0x0f) << 8) + rslt[2]
		end
		
		rslt = self:ReadReg( HDMI_MAP, 0x26, {0x00, 0x00} )
		if rslt then
			totalH = ((rslt[1]&0x1f) << 8) + rslt[2]
		end
		
		rslt = self:ReadReg( HDMI_MAP, 0x28, {0x00, 0x00} )
		if rslt then
			totalH2 = ((rslt[1]&0x1f) << 8) + rslt[2]
		end
		
		rslt = self:ReadReg( HDMI_MAP, 0x09, {0x00, 0x00} )
		if rslt then
			actH = ((rslt[1]&0x0f) << 8) + rslt[2]
		end
	
		rslt = self:ReadReg( HDMI_MAP, 0x0b, {0x00, 0x00} )
		if rslt then
			actH2 = ((rslt[1]&0x0f) << 8) + rslt[2]
		end
		
--[[
	/*HS*/
	adv7611_read_reg(vdev, HDMI_MAP, 0x20, &regdata1);
	adv7611_read_reg(vdev, HDMI_MAP, 0x21, &regdata2);
	pinfo->hs_front_porch = (regdata1 & 0x1f) << 8;
	pinfo->hs_front_porch |= regdata2;

	adv7611_read_reg(vdev, HDMI_MAP, 0x22, &regdata1);
	adv7611_read_reg(vdev, HDMI_MAP, 0x23, &regdata2);
	pinfo->hs_plus_width = (regdata1 & 0x1f) << 8;
	pinfo->hs_plus_width |= regdata2;

	adv7611_read_reg(vdev, HDMI_MAP, 0x24, &regdata1);
	adv7611_read_reg(vdev, HDMI_MAP, 0x25, &regdata2);
	pinfo->hs_back_porch = (regdata1 & 0x1f) << 8;
	pinfo->hs_back_porch |= regdata2;
]]

		if actW < 160 or actH < 90 or refreshFreq < 23 then
			return false
		end
		
		local formatName
		if interlaced then
			formatName = actW .. "x" .. (actH + actH2) .. "@" .. refreshFreq .. "i"
		else
			formatName = actW .. "x" .. actH .. "@" .. refreshFreq .. " Hz"
		end
		
		local v = self.nostandardFormatInfo[formatName]
		if not v then	--Create this format information
			v = {
				formatId = formatName,
				formatName = formatName,
				frameRate = refreshFreq,
				hTotal = totalW,
				hAct = actW,
				vTotal0 = totalH,
				vAct0 = actH,
				vTotal1 = totalH2,
				vAct1 = actH2,
				interlaced = interlaced,
				rateMode = 1,	--That is VIRATE_HD
				tmdsFreq = tmdsFreq,
				spec_amba_mode = 0,	--That is AMBA_VIDEO_MODE_AUTO
				pixelRepeat = false,	--TODO:
				aspect_ratio = 0, --That is ASPECT_RATIO_AUTO
				video_system = 0, --That is AMBA_VIDEO_SYSTEM_AUTO
			}
			self.nostandardFormatInfo[formatName] = v
		end
		
		return self:_RecordVideoFormatInfo( src, v )
	end

	------------------------------------------------------
	-- Detect HDMI input format
	------------------------------------------------------
	
	function ADV7611:_DetectHDMIFormat(src)
		local rslt = self:ReadReg( HDMI_MAP, 0x05, {0x00} )		--Check HDCP
		local hdcp = false
		if rslt and (rslt[1]&0x40) ~= 0 then
			hdcp = true
			self.videoFormat.hdcp_content = 1
		else
			self.videoFormat.hdcp_content = 0	--It may contains an error status, but, no matter.
		end
		
		if rslt and (rslt[1]&0x80) ~= 0 then
			self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_HDMI or 1
		else
			self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_DVI or 0
		end
		
		self.videoFormat._valid = false
		self.videoFormat.signal_locked = true
		
		--Check info frame first. NOTE: Some HDMI source does not provide a valid 
		--format Id.
		--[[
		if self:_DetectHDMIInfoFrame(src) then
			--Clear all recorded BL/LCF/FCL/LCVS/interlaced
			self.videoFormat._valid = true
			self.lastFormatRecord.BL = nil
			self.lastFormatRecord.LCF = nil
			self.lastFormatRecord.FCL = nil
			self.lastFormatRecord.LCVS = nil
			self.lastFormatRecord.interlaced = nil
			return true
		end
		]]

		self.lastFormatRecord.infoFrameId = -1		--Invalid recorded infoFrameId
		
		--So sorry, I need check special formats
		local done, BL, LCF, FCL, LCVS, interlaced = self:_STDICheck()
		
		if done and ( BL ~= self.lastFormatRecord.BL 
								or LCF ~= self.lastFormatRecord.LCF
								or FCL ~= self.lastFormatRecord.FCL
								or LCVS ~= self.lastFormatRecord.LCVS
								or interlaced ~= self.lastFormatRecord.interlaced
							)
		then
			rslt = hdtv_formats.STDI_Match( LCF, LCVS, FCL, BL, interlaced )
			if rslt then
				if self:_RecordVideoFormatInfo( src, rslt ) then
					self.videoFormat._valid = true
					self.lastFormatRecord.BL = BL
					self.lastFormatRecord.LCF = LCF
					self.lastFormatRecord.FCL = FCL
					self.lastFormatRecord.LCVS = LCVS
					self.lastFormatRecord.interlaced = interlaced
					return true
				end
			end
			
			--[[
			rslt = vga_formats.STDI_Match( LCF, LCVS, FCL, BL, interlaced )	--Check VGA standards
			if rslt then
				if self:_RecordVideoFormatInfo( src, rslt ) then
					self.videoFormat._valid = true
					self.lastFormatRecord.BL = BL
					self.lastFormatRecord.LCF = LCF
					self.lastFormatRecord.FCL = FCL
					self.lastFormatRecord.LCVS = LCVS
					self.lastFormatRecord.interlaced = interlaced
					return true
				end
			end
			]]
			
			if self:_AssumeHDMICustomFormat( src, BL, LCF, FCL, LCVS, interlaced ) then
					self.videoFormat._valid = true
					self.lastFormatRecord.BL = BL
					self.lastFormatRecord.LCF = LCF
					self.lastFormatRecord.FCL = FCL
					self.lastFormatRecord.LCVS = LCVS
					self.lastFormatRecord.interlaced = interlaced
					return true
			else
				return false	--Not detect
			end
		elseif done then
			self.videoFormat._valid = true
			return true	--Already recorded
		end

		return false
		
	end

	------------------------------------------------------
	-- Configure HDMI format
	------------------------------------------------------
	
	function ADV7611:_ConfigureHDMIFormat(src, defaultBlue)
		self.videoFormat._lastFormatChanged = false
		local CLOCK_PHASE = 0x83
	
		if defaultBlue then
			print( "Configure ADV7611 output blue screen @ 720p60." )
			self:WriteReg( CP_MAP, 0xba, {0x01} )
			self:WriteReg( CP_MAP, 0xc9, {0x2c} )
			self:WriteReg( CP_MAP, 0xbf, {0x12} )	--//0x13, force free run.
			self:WriteReg( BASE_MAP, 0x01, {0x05} )
			self:WriteReg( BASE_MAP, 0x00, {0x13} )
			self:WriteReg( BASE_MAP, 0x19, {CLOCK_PHASE} )	--LLC DLL phase (0x83?0x89?)
			self:WriteReg( BASE_MAP, 0x33, {0x40} )	--LLC DLL enable
			return true
		end
	
		if self.videoFormat._formatInfo and self.videoFormat._formatInfo.spec_adv7611_mode then	--Enable free run
			self:WriteReg( CP_MAP, 0xba, {0x01} )
			self:WriteReg( CP_MAP, 0xc9, {0x2c} )
		else
			self:WriteReg( CP_MAP, 0xba, {0x00} ) --Disable freefun
			self:WriteReg( CP_MAP, 0xc9, {0x2d} ) --Disable AUTO_PARAM_BUF ( NO_AUTO_PARAM_BUF=1 )
		end
		--Free run color options (using default blue)
		self:WriteReg( CP_MAP, 0xbf, {0x12} )	--//0x13, force free run.
		
		local data = 0	--Frame rate code
		if self.videoFormat.frame_rate_value then
				local v = self.videoFormat.frame_rate_value
				if v >= 59 then
					data = 0
				elseif v >= 49 then
					data = 1
				elseif v >= 29 then
					data = 2
				elseif v >=25 then
					data = 3
				elseif v >= 24 then
					data = 4
				elseif v >= 23 then
					data = 3
				end
		end
		data  = data << 4
		if self.videoFormat._formatInfo and self.videoFormat._formatInfo.rateMode and self.videoFormat._formatInfo.rateMode ~= hdtv_formats.VIRATE_VGA then
			data = data + 5
		else
			data = data + 6
		end
		
		self:WriteReg( BASE_MAP, 0x01, {data} )
		
		if self.videoFormat._formatInfo and self.videoFormat._formatInfo.spec_adv7611_mode then
			self:WriteReg( BASE_MAP, 0x00, {self.videoFormat._formatInfo.spec_adv7611_mode} )
		else
			self:WriteReg( BASE_MAP, 0x00, {0x00} )
		end

		--TODO: the threshold value 100M is correct?
		if self.videoFormat._formatInfo and self.videoFormat._formatInfo.tmdsFreq and self.videoFormat._formatInfo.tmdsFreq >= 100000000 then
			CLOCK_PHASE = 0x97
		end
	
		self:WriteReg( BASE_MAP, 0x19, {CLOCK_PHASE} )	--LLC DLL phase (0x83?0x89?)
		self:WriteReg( BASE_MAP, 0x33, {0x40} )	--LLC DLL enable

		return true
	end
	
	------------------------------------------------------
	
	ADV7611:WriteReg( BASE_MAP, 0xf4, {CEC_MAP} )
	ADV7611:WriteReg( BASE_MAP, 0xf5, {INFOFRAME_MAP} )
	ADV7611:WriteReg( BASE_MAP, 0xf8, {DPLL_MAP} )
	ADV7611:WriteReg( BASE_MAP, 0xf9, {KSV_MAP} )
	ADV7611:WriteReg( BASE_MAP, 0xfa, {EDID_MAP} )
	ADV7611:WriteReg( BASE_MAP, 0xfb, {HDMI_MAP} )
	ADV7611:WriteReg( BASE_MAP, 0xfd, {CP_MAP} )
	
	--ADV7611:_Reset()
	
	return ADV7611
end --END of CreateADV7611()

return CreateADV7611

