local CONFIG_FILE = "/data/configs/osd.conf"
local OSD_IMAGE_LOC = "/data/osd_images"

require "configuredb"
require "osd_config"

local getset = require "kiloview.getSetHelper"
local osdAdapter = require "mod_OSD_adapter"
local _dbg = require "coaf.debug"

---------------------------------------------------
_G.osd = CONFIGS.InstanceOf( OSD_Template, "osd" )
_G.OSD_CONFIG_LOADING = true	--Lock to avoid trigger events
_G.osd:LoadFile( CONFIG_FILE )
_G.OSD_CONFIG_LOADING = false
_G.osd:ResetChanges()
---------------------------------------------------

local _M = {
	osdgroups = {
	},

	AreaInfos = {},
	FontMap = {
		["vera"] = 0,
		["lucida"] = 1,
		["song"] = 2,
		["default"] = 3
	},
	FontFileMap = {
		["vera"] = "Vera.ttf",
		["lucida"] = "Lucida.ttf",
		["song"] = "gbsn00lp.ttf",
		["default"] = "Vera.ttf"
	},
	DefaultFontIndex = 0,

	dynamic_data = {
		coordinates = {
			latitude = nil,
			latitude_hemisphere = nil,
			longitude = nil,
			longitude_hemisphere = nil
		},

		altitude = nil,

		speed = nil,

		bearing = nil
	}
}

---------------------------------------------------

local OSD_SVC = {
	Identifier = "OSD",
	FriendlyName = "Encoder OSD service",
	Version = "1.0",
	ConfigPrefix = "osd",
}

function OSD_SVC:Activate()
	--Nothing
end

function OSD_SVC:Start()
	--Nothing
end

function OSD_SVC:Stop()
	--Nothing
end

function OSD_SVC:Reset()
--TODO: Nothing
end

function OSD_SVC:Terminate()
--TODO: Nothing
end

function OSD_SVC:Control(action,params)
	if action == "activate" then
		self:Activate()
		return 0, {}
	elseif action == "start" then
		self:Start()
		return 0, {}
	elseif action == "stop" then
		self:Stop()
		return 0, {}
	elseif action == "Reset" then
		self:Reset()
		return 0, {}
	elseif action == "terminate" then
		self:Terminate()
		return 0, {}
	elseif action == "update" then
		_M.UpdateOSD()
		return 0, {}
	else
		return -1, {}
	end
end

function OSD_SVC:Set(params)
	local cnt = getset.SetConfigureValues( self.ConfigPrefix, params )
	return cnt
end

function OSD_SVC:SetEnv(params)
	local cnt = getset.SetValues( "_ENV", params )
	return cnt
end

function OSD_SVC:Get(inParams)
	local cnt = getset.GetValues( self.ConfigPrefix, inParams )
	return cnt,inParams
end

function OSD_SVC:GetEnv(inParams)
	local cnt = getset.GetValues( "_ENV", inParams )
	return cnt,inParams
end

function OSD_SVC:RestoreDefault()
	--TODO:
	return 0
end

function OSD_SVC:LoadConfigureFile(filename)
	CONFIGS.BeginUpdate()
	local r = pcall( dofile, filename )
	CONFIGS.Submit()
	if r then
		return 1
	else
		return 0
	end
end

function OSD_SVC:UpdateDynamicData(value)
	if type(value) == "table" then
		local update_types = {}
		local k,v
		for k,v in pairs(value) do
			table.insert(update_types,k)
			_M.dynamic_data[k] = v
		end

		local i
		for i=1, #_M.osdgroups do
			_M.osdgroups[i]:UpdateOSDByTypes(update_types)
		end

		--Render immediately.
		_M.PeriodRender()
	end
end

----------------------------------------------
--Return ready or not, streamW, streamH, isRotate
local function CheckStreamInfo(streamId)
	local encoder = require "encoder"
	return encoder.CheckStreamStatus(streamId)
end
----------------------------------------------

local osdgroup_meta = {}

function osdgroup_meta:Init( width, height )
	if self.started then
		self.osdObj:Reset()
	end

	self.items = {}

	self.osdObj:ChangeSize( width, height, 0)
	self.width = width
	self.height = height
	self.started = true
end

function osdgroup_meta:Destroy()
	if self.started then
		self.osdObj:Destroy()
		self.osdObj = nil
	end
	collectgarbage()
end

local function format_coordinates(format)
	local lati_dgree
	local lati_min
	local lati_min_f
	local lati_hemi
	local longi_dgree
	local longi_min
	local longi_min_f
	local longi_hemi

	if _M.dynamic_data and type(_M.dynamic_data.coordinates) == "table" then
		local coord = _M.dynamic_data.coordinates
		if coord.latitude then
			lati_dgree = tonumber(coord.latitude:sub(1,2))
			lati_min = tonumber(coord.latitude:sub(3,4))
			lati_min_f = tonumber("0" .. coord.latitude:sub(5))
		end
		lati_hemi = coord.latitude_hemisphere or ""

		if coord.longitude then
			longi_dgree = tonumber(coord.longitude:sub(1,3))
			longi_min = tonumber(coord.longitude:sub(4,5))
			longi_min_f = tonumber("0" .. coord.longitude:sub(6))
		end
		longi_hemi = coord.longitude_hemisphere or ""
	end

	if not lati_dgree or not longi_dgree then
		return "Coordinates: N/A"
	end

	if format == "dm" then
		return string.format("%02d° %02d", lati_dgree, lati_min or 0) .. 
			   string.format("%.4f' ", lati_min_f or 0):sub(2) .. lati_hemi .. "  " ..
			   string.format("%03d° %02d", longi_dgree, longi_min or 0) ..
			   string.format("%.4f' ", longi_min_f or 0):sub(2) .. longi_hemi
	elseif format == "d" then
		lati_min = (lati_min or 0) + (lati_min_f or 0)
		lati_min = lati_min/60
		lati_dgree = lati_dgree + lati_min
		longi_min = (longi_min or 0) + (longi_min_f or 0)
		longi_min = longi_min/60
		longi_dgree = longi_dgree + longi_min
		return string.format("%02d", math.floor(lati_dgree)) .. 
			   string.format("%.7f° ", lati_dgree-math.floor(lati_dgree)):sub(2) .. lati_hemi .. "  " ..
			   string.format("%03d", math.floor(longi_dgree)) ..
			   string.format("%.7f° ", longi_dgree-math.floor(longi_dgree)):sub(2) .. longi_hemi
	else --type(format) ~= "string" or format == "" or format == "dms" then
		lati_min_f = (lati_min_f or 0)*60
		longi_min_f = (longi_min_f or 0)*60
		return string.format("%02d° %02d' ", lati_dgree, lati_min or 0) .. 
			   string.format("%.4f\" ", lati_min_f) .. lati_hemi .. "  " ..
			   string.format("%03d° %02d' ", longi_dgree, longi_min or 0) ..
			   string.format("%.4f\" ", longi_min_f) .. longi_hemi
	end
end

local function format_altitude(format)
	if format == "f" then
		if _M.dynamic_data and _M.dynamic_data.altitude then
			return "AMSL: " .. string.format( "%.3f", (tonumber(_M.dynamic_data.altitude) or 0)*3.28084 ) .. " F"
		else
			return "AMSL: --"
		end
	else
		if _M.dynamic_data and _M.dynamic_data.altitude then
			return "AMSL: " .. string.format( "%.3f", (tonumber(_M.dynamic_data.altitude) or 0) ) .. " M"
		else
			return "AMSL: --"
		end
	end
end

local function format_speed(format)
	if format == "kph" then
		if _M.dynamic_data and _M.dynamic_data.speed then
			return string.format( "%.3f", (tonumber(_M.dynamic_data.speed) or 0)*1.852 ) .. " KPH"
		else
			return "-- KPH"
		end
	elseif format == "mph" then
		if _M.dynamic_data and _M.dynamic_data.speed then
			return string.format( "%.3f", (tonumber(_M.dynamic_data.speed) or 0)*1.1508 ) .. " MPH"
		else
			return "-- MPH"
		end
	else
		if _M.dynamic_data and _M.dynamic_data.speed then
			return string.format( "%.3f", (tonumber(_M.dynamic_data.speed) or 0) ) .. " KT"
		else
			return "-- KT"
		end
	end
end

local function format_bearing(format)
	return "Bearing: " .. (_M.dynamic_data and _M.dynamic_data.bearing or "--") .. " deg"
end

function osdgroup_meta:UpdateOSDArea( areaId, glbCfg, itemCfg )
	if self.forAmbaOverlay then
		local ready, streamW, streamH, isRotate, rotateDeg = CheckStreamInfo(self.streamId)
		--print( "***Check stream:", i, ready, streamW, streamH, isRotate )
		if glbCfg.enabled == 0 or not ready then
			if glbCfg.enabled == 0 then
				self.osdObj:DropArea(areaId)
			end
			return
		end

		if glbCfg.norotate == 0 and isRotate then
			self.width = streamH or 0
			self.height = streamW or 0
		else
			self.width = streamW or 0
			self.height = streamH or 0
		end

		if rotateDeg and rotateDeg ~= 0 and glbCfg.norotate ~= 0 then
			self.osdObj:ChangeSize(self.width, self.height, 360-rotateDeg)
		else
			self.osdObj:ChangeSize(self.width, self.height, 0)
		end
	end

	if self.width <=0 or self.height <=0 then
		_dbg.Warn( "OSD", "The stream of " .. self.streamId .. " with invalid size! So could not update the areas." )
		return
	end

	local text
	local osd_type = itemCfg.osd_type -- or itemCfg.type
	local imageId = nil

	if itemCfg.show_when ~= "always" then
		if ( itemCfg.show_when == "nosignal" and itemCfg.showing == 1 ) or
		   ( itemCfg.show_when == "signal" and itemCfg.showing == 0 ) then
			osd_type = "none"
		end
	end
	self.osdObj:DropArea(areaId)
	if glbCfg.enabled == 0 or not osd_type or osd_type == "none" or osd_type == "" then
		return
	--elseif not itemCfg.dynamic_show then
	--	self.osdObj:EnableArea(areaId, false)
	--	return
	elseif osd_type == "time" then
		if itemCfg.format == "" then
			text = "%Y-%m-%d %H:%M:%S"
		else
			text = itemCfg.format
		end
	elseif osd_type == "date" then
		if itemCfg.format == "" then
			text = "%Y-%m-%d"
		else
			text = itemCfg.format
		end
	elseif osd_type == "clock" then
		if itemCfg.format == "" then
			text = "%H:%M:%S"
		else
			text = itemCfg.format
		end
	elseif osd_type == "coordinates" then
		text = format_coordinates(itemCfg.format)
	elseif osd_type == "altitude" then
		text = format_altitude(itemCfg.format)
	elseif osd_type == "speed" then
		text = format_speed(itemCfg.format)
	elseif osd_type == "bearing" then
		text = format_bearing(itemCfg.format)
	elseif osd_type == "image" then
		osd_type = "image"
		imageId = itemCfg.image_id
		text = "~image"
	else
		osd_type = "text"
		text = itemCfg.text
	end

	local thisAreaW, thisAreaH
	local xOffset = 16
	local yOffset = 16

	self.osdObj:ChangeAreaType(areaId, osd_type )
	if not text or text == "" then --Sure, image also with a text '~image'
		self.osdObj:EnableArea(areaId, false)
		return
	else
		self.osdObj:EnableArea(areaId, true)
	end

	if osd_type == "image" then
		thisAreaW, thisAreaH = self.osdObj:LoadImageByID(areaId, imageId)
		thisAreaW = thisAreaW or 0
		thisAreaH = thisAreaH or 0
	else
		local line_space = itemCfg.line_space or "x0.5"
		local line_space_ratio = true
		if line_space:match("^x") then
			line_space = tonumber(line_space:sub(2)) or 0.5
		else
			line_space_ratio = false
			line_space = tonumber(line_space) or 8
		end

		local fontSize = 32
		local screenGridW, screenGridH
		local imageFile
		local font

		if itemCfg.size <= 2 then
			local screenW = 40
			if itemCfg.size <= 0 then
				screenW = 80
			elseif itemCfg.size == 1 then
				screenW = 60
			end

			screenGridW = math.floor( self.width/screenW )
			fontSize = screenGridW
		else
			fontSize = itemCfg.size
			screenGridW = fontSize
		end

		if line_space_ratio then
			screenGridH = math.floor(screenGridW * (1+line_space))
		else
			screenGridH = screenGridW + line_space
		end

		local fntname = string.lower( itemCfg.font )
		font = _M.FontMap[fntname] or _M.DefaultFontIndex or 0

		self.osdObj:SetAreaFont( areaId, { font = font, size=fontSize, bold=itemCfg.bold or 0, italic=itemCfg.italic or 0, outline=itemCfg.outline } )
		self.osdObj:SetAreaFontColor( areaId, itemCfg.font_color )
		self.osdObj:SetAreaOutlineColor( areaId, itemCfg.outline_color )
		self.osdObj:SetAreaLineHeight( areaId, screenGridH )
		self.osdObj:SetAreaText( areaId, text )
		--Calculate lines and area height, according to screen grid height (but not text rendering size)
		local lineCnt = 0
		local flag
		for flag in text:gmatch("(\n)") do
			lineCnt = lineCnt + 1
		end

		thisAreaW, thisAreaH = self.osdObj:TellAreaSize( areaId )
		thisAreaH = (lineCnt+1)*screenGridH
		--Should return back restricted size.
		thisAreaW, thisAreaH = self.osdObj:ChangeAreaSize( areaId, thisAreaW, thisAreaH )
	end

	local x, y
	if itemCfg.position == "left-top" then
		x = xOffset
		y = yOffset
	elseif itemCfg.position == "top-center" then
		x = math.floor((self.width - thisAreaW)/2)
		y = yOffset
	elseif itemCfg.position == "right-top" then
		x = self.width - thisAreaW - xOffset
		y = yOffset
	elseif itemCfg.position == "left-bottom" then
		x = xOffset
		y = self.height - yOffset - thisAreaH
	elseif itemCfg.position == "bottom-center" then
		x = math.floor((self.width - thisAreaW)/2)
		y = self.height - yOffset - thisAreaH
	elseif itemCfg.position == "right-bottom" then
		x = self.width - thisAreaW - xOffset
		y = self.height - yOffset - thisAreaH
	elseif itemCfg.position == "center" then
		x = math.floor((self.width - thisAreaW)/2)
		y = math.floor((self.height - thisAreaH)/2)
	else
		local posX = itemCfg.x or 0 --itemCfg.customX or 0
		local posY = itemCfg.y or 0 --itemCfg.customY or 0
		--x = xOffset + posX          --* screenGridW
		--y = yOffset + posY          --* screenGridH
		x = posX
		y = posY
	end

	--Debug only
	--print("********* stream-Id:", self.streamId, " areaId:", areaId, "Position:", itemCfg.position, "x,y:", x, y, text)

	self.osdObj:SetAreaPositionType( areaId, itemCfg.position )
	self.osdObj:SetAreaPosition( areaId, x, y )
end

function osdgroup_meta:UpdateOSD(itemId)
	local glbCfg = {
		enabled = _G.osd.stream[ self.streamId ].enabled,
		norotate = _G.osd.stream[ self.streamId ].norotate,
		-- gridW = _G.osd.stream[ self.streamId ].screen_w,
		-- gridH = _G.osd.stream[ self.streamId ].screen_h,
		gridShiftX = 0.5,
		gridShiftY = 0.5
	}

	local i
	for i=0, 15 do
		--if self.forAmbaOverlay or not itemId or itemId == i then
		if not itemId or itemId == i then
			local itemCfg = _G.osd.stream[ self.streamId ].items[i]
			if itemCfg then
				self:UpdateOSDArea( i, glbCfg, itemCfg )
			end
		end
	end

	self.osdObj:Update()
end

function osdgroup_meta:UpdateOSDByTypes(types)
	if type(types) ~= "table" then
		types = {types}
	end

	local glbCfg = {
		enabled = _G.osd.stream[ self.streamId ].enabled,
		-- gridW = _G.osd.stream[ self.streamId ].screen_w,
		-- gridH = _G.osd.stream[ self.streamId ].screen_h,
		gridShiftX = 0.5,
		gridShiftY = 0.5
	}

	local i, j
	local updated
	for i=0, 15 do
		local itemCfg = _G.osd.stream[ self.streamId ].items[i]
		if itemCfg then
			for j=1, #types do
				if itemCfg.osd_type == types[j] then
					updated = true
					self:UpdateOSDArea( i, glbCfg, itemCfg )
					break
				end
			end
		end
	end

	if updated then
		self.osdObj:Update()
	end
end

osdgroup_meta.__index = osdgroup_meta

---------------------------------------------------

function _M.Init(  cosvc )
	--_M.UpdateOSD()
	if cosvc then
		cosvc:ExposeObject("/OSD", OSD_SVC)
	end
	print( "OSD Service online." )


	print( "Loading fonts..." )
	local fontMapFnx = loadfile( "/usr/share/fonts/font_patch.lua" )
	if type(fontMapFnx) == "function" then
		local r, fontMap = pcall(fontMapFnx)
		if r and type(fontMap) == "table" then
			_M.FontMap = {}
			local k,v
			for k,v in pairs(fontMap) do
				_M.FontMap[v.key] = v.index
				_M.FontFileMap[v.key] = k
				if v.default then
					_M.DefaultFontIndex = v.index
				end
				print( "##", v.index, k, v.locale )
			end
			print( "Fonts mapping loaded." )
		else
			print( "WARNING: Bad fonts mapping file, ignore it." )
		end
	else
		print("WARNING: No valid fonts mapping file, used default." )
	end

	--Map font files.
	local k,v
	for k,v in pairs(_M.FontMap) do
		osdAdapter.fontMap[v+1] = _M.FontFileMap[k]
	end

	local i
	for i=0, 2 do
		local obj = {
			streamId = i,
			forAmbaOverlay = true,
			osdObj = osdAdapter.createNew("amba_iav", {id=i} )
		}
		setmetatable(obj, osdgroup_meta)
		table.insert( _M.osdgroups, obj )
	end

	for i=1, #_M.osdgroups do
		if _M.osdgroups[i].forAmbaOverlay then
			_M.osdgroups[i]:Init(1920, 1080)
	--	else
	--		_M.osdgroups[i]:Init(720, 576)
		end
	end

	_M.UpdateOSD()
end

function _M.UpdateOSD(itemId)
	local i
	for i=1, #_M.osdgroups do
		_M.osdgroups[i]:UpdateOSD(itemId)
	end
	osdAdapter.ForceRefresh()
end

function _M.PeriodRender()
	osdAdapter.PeriodRender()
end

return _M
