local getset = require "kiloview.getSetHelper"
local PTZ_M = require "ptz"
local CONFIG_FILE = "/data/configs/ptz.conf"
local lfs = require "lfs"
local socket = require "socket"
local ev = require "ev"

require "configuredb"
require "ptz_config"

---------------------------------------------------
_G.ptz = CONFIGS.InstanceOf( PTZ_Template, "ptz" )
_G.PTZ_CONFIG_LOADING = true	--Lock to avoid trigger events
_G.ptz:LoadFile( CONFIG_FILE )
_G.PTZ_CONFIG_LOADING = false
_G.ptz:ResetChanges()

---------------------------------------------------

local _M = {
	ptz_obj = nil,
	serial_rx_watcher = nil
}

---------------------------------------------------
local SOCKET_SVC = {
	version = "1.0",
	svc_socket = nil,
	loop = nil,

	clients = {
	}
}
---------------------------------------------------
 
local function StartSerialRxMonitor()
	if not _M.ptz_obj then
		print( "SERIAL_SERVICE: Invalid serial/ptz object while starting Rx watcher." )
		return
	end

	if _M.serial_rx_watcher then
		print( "SERIAL_SERVICE: The serial Rx watcher is already started." )
		return
	end

	_M.serial_rx_watcher = ev.IO.new( function( loop, watcher )
		local ser = _M.ptz_obj and _M.ptz_obj:GetSerialObject() or nil
		if not ser then
			print( "SERIAL_SERVICE: Invalid serial/ptz object, so closed the watcher!" )
			watcher:stop(loop)
			_M.serial_rx_watcher = nil
			return
		end

		local data
		local status, err = pcall( function() data = ser:read(128,0) end )
		if not status then
			print( "SERIAL_SERVICE: Read serial port error! Stopped the watcher!" )
			watcher:stop(loop)
			_M.serial_rx_watcher = nil
			return
		end

		if data and #data > 0 then -- Got data!
			local k,v
			for k,v in pairs( SOCKET_SVC.clients ) do
				if v.stage == "established" and not v.disableRx then
					v:Send( data )
				end
			end
		end
	end, _M.ptz_obj:GetSerialObject().fd, ev.READ )

	if not _M.serial_rx_watcher then
		print( "SERIAL_SERVICE: Warning! Fail to start serial Rx watcher!" )
		return
	end

	_M.serial_rx_watcher:start( SOCKET_SVC.loop )
end


local function StopSerialRxMonitor()
	if _M.serial_rx_watcher then
		if _M.serial_rx_watcher:is_pending() then
			_M.serial_rx_watcher:clear_pending()
		end
		_M.serial_rx_watcher:stop( SOCKET_SVC.loop )
		_M.serial_rx_watcher = nil
	end
end

---------------------------------------------------


local PTZ_SVC = {
	Identifier = "PTZ",
	FriendlyName = "PTZ Service",
	Version = "1.0",
	ConfigPrefix = "ptz",
}

function PTZ_SVC:Activate()
	--Nothing
end

function PTZ_SVC:Start()
end

function PTZ_SVC:Stop()
end

function PTZ_SVC:Reset()
--TODO: Nothing
end

function PTZ_SVC:Terminate()
	if _M.ptz_obj then
		StopSerialRxMonitor()
		_M.ptz_obj:Close()
		_M.ptz_obj = nil
	end
	
	os.exit(0)
end

function PTZ_SVC:Control(action,params)
	if action == "activate" then
		self:Activate()
		return 0, {}
	elseif action == "start" then
		self:Start()
		return 0, {}
	elseif action == "stop" then
		self:Stop()
		return 0, {}
	elseif action == "Reset" then
		self:Reset()
		return 0, {}
	elseif action == "terminate" then
		self:Terminate()
		return 0, {}
	elseif action == "set-address" then
		if _M.ptz_obj then
			local addr = 1
			if params then
				if params["address"] then
					addr = tonumber(params["address"])
				elseif params["addr"] then
					addr = tonumber(params["addr"])
				end
			end
			_M.ptz_obj:SetAddress(addr)
			return 0,{}
		else
			return -1,{}
		end
	elseif action == "load-protocol" then
		if _M.ptz_obj then
			_M.ptz_obj:LoadProtocol( params and params["protocol"] or "pelco_d" )
			return 0,{}
		else
			return -1,{}
		end
	elseif action == "set-speed" then
		if _M.ptz_obj then
			_M.ptz_obj:SetSpeed( params )
			return 0,{}
		else
			return -1, {}
		end
	elseif action == "save-preset" then
		if _M.ptz_obj then
			--TODO: Mark presets
			_M.ptz_obj:SavePreset( params and params["id"] or 0 )
			return 0,{}
		else
			return -1, {}
		end
	elseif action == "remove-preset" then
		if _M.ptz_obj then
			--TODO: Mark presets
			_M.ptz_obj:RemovePreset( params and params["id"] or 0 )
			return 0,{}
		else
			return -1, {}
		end
	elseif action == "load-preset" then
		if _M.ptz_obj then
			--TODO: Mark presets
			_M.ptz_obj:LoadPreset( params and params["id"] or 0 )
			return 0,{}
		else
			return -1, {}
		end
	else
		if _M.ptz_obj then
			local arg1 = nil
			local arg2 = nil
			if params then
				if params["speed"] then
					arg1 = tonumber(params["speed"])
				elseif params["pan_speed"] then
					arg1 = tonumber(params["pan_speed"])
				elseif params["pan"] then
					arg1 = tonumber(params["pan"])
				elseif params["id"] then
					arg1 = tonumber(params["id"])
				end

				if params["tilt_speed"] then
					arg2 = tonumber(params["tilt_speed"])
				elseif params["tilt"] then
					arg2 = tonumber(params["tilt"])
				elseif params["speed2"] then
					arg2 = tonumber(params["speed2"])
				end
			end
			local r = _M.ptz_obj:DoAction( action, _M.ptz_obj:GetActionSpeedArgs(arg1,arg2) )
			if r then
				return 0, {}
			else
				return -1, {}
			end
		else
			return -1, {}
		end
	end
end

function PTZ_SVC:Set(params)
	local cnt = getset.SetConfigureValues( self.ConfigPrefix, params )
	return cnt
end

function PTZ_SVC:SetEnv(params)
	local cnt = getset.SetValues( "_ENV", params )
	return cnt
end

function PTZ_SVC:Get(inParams)
	local cnt = getset.GetValues( self.ConfigPrefix, inParams )
	return cnt,inParams
end

function PTZ_SVC:GetEnv(inParams)
	local cnt = getset.GetValues( "_ENV", inParams )
	return cnt,inParams
end

function PTZ_SVC:RestoreDefault()
	--TODO:
	return 0
end

function PTZ_SVC:LoadConfigureFile(filename)
	CONFIGS.BeginUpdate()
	local r = pcall( dofile, filename )
	CONFIGS.Submit()
	if r then
		return 1
	else
		return 0
	end
end

---------------------------------------------------
function _M.OpenPTZ()
	_M.ptz_obj = PTZ_M.CreateNew({	device = ptz.device.device, 
									baudrate = ptz.device.baudrate, 
									databits = ptz.device.databits, 
									parity = ptz.device.parity, 
									stopbits = ptz.device.stopbits, 
									xonxoff = (ptz.device.xonxoff == 1),
									rtscts = (ptz.device.rtscts == 1)
								})
end

---------------------------------------------------

function _M.ConfigPTZ()
	if _M.ptz_obj then
		StartSerialRxMonitor()
		_M.ptz_obj:SetAddress( ptz.options.address )
		_M.ptz_obj:LoadProtocol( ptz.options.protocol )
		_M.ptz_obj:SetSpeed( {
			pan = ptz.options.pan_speed,
			tilt = ptz.options.tilt_speed,
			zoom = ptz.options.zoom_speed,
			focus = ptz.options.focus_speed,
			iris = ptz.options.iris_speed
		})
	end
end

---------------------------------------------------
-- Period check to create the PTZ object.
---------------------------------------------------

function _M.Check()
	local attr = lfs.attributes( ptz.device.device )
	if not attr then --Device is not exist.
		if _M.ptz_obj then
			StopSerialRxMonitor()
			_M.ptz_obj:Close()
			_M.ptz_obj = nil
		end
		return
	end

	if attr and not _M.ptz_obj then
		_M.OpenPTZ()
		_M.ConfigPTZ()
	end
end

---------------------------------------------------
function _M.Config()
	if _M.ptz_obj then
		StopSerialRxMonitor()
		_M.ptz_obj:Close()
		_M.ptz_obj = nil
	end

	_M.OpenPTZ()
	_M.ConfigPTZ()
end

---------------------------------------------------
--Call it while configure changed.
function _M.UpdateSocketService(en,port)
	if not en or en == 0 then
		SOCKET_SVC.Stop()
	else
		SOCKET_SVC.Start(port)
	end
end

---------------------------------------------------

function _M.Init( cosvc, loop )
	SOCKET_SVC.loop = loop

	_M.Config()
	if cosvc then
		cosvc:ExposeObject( "/PTZ", PTZ_SVC )
	end
	print( "PTZ CoAF service online." )

	if _G.ptz.socket_svc.enable ~= 0 then
		SOCKET_SVC.Start( _G.ptz.socket_svc.port )
	end
end

---------------------------------------------------

--CMDs:
--  SER_OPEN [default [115200,8,N,1]]	=> OK ttyS1 115200,8,N,1 or OK none or ERROR <msg>
--  SER_CLOSE							=> OK or ERROR <msg>
--  PING							=> OK PONG
--  SELF				=> OK 192.168.5.1:33333
--	CLIENTS				=> OK 3 192.168.3.1:12341 192.168.1.5:11071 #192.168.5.1:33333
--	KILL <client|*|+>	=> OK or ERROR <msg>
--	INFO				=> OK <version> or ERROR <msg>
--	BYE					=> closed
--	START [TR/TO/RO]	=> Enter tranceiver

local function ProcessClientNego( ctx, cmd )
	local tokens = {}
	local item
	cmd = cmd .. " "
	for item in cmd:gmatch("(.-) ") do
		if item ~= "" then
			table.insert( tokens, item )
		end
	end

	cmd = tokens[1]
	if not cmd or cmd == "" then
		ctx:Send( "ERROR Bad format\r\n" )
		return
	end
	
	cmd = cmd:upper()
	table.remove( tokens, 1 )

	if cmd == "SER_OPEN" then
		if tokens[1] then
			CONFIGS.BeginUpdate()
			if tokens[1]:lower() == "default" then
				_G.ptz.device.device = "/dev/ttyS1"
			else
				_G.ptz.device.device = "/dev/" .. tokens[1]
			end

			if tokens[2] then
				local sets = tokens[2]:upper() .. ","
				local cnt = 1
				for item in sets:gmatch("(.-),") do
					if cnt == 1 then
						_G.ptz.device.baudrate = tonumber(item) or 115200
					elseif cnt == 2 then
						_G.ptz.device.databits = tonumber(item) or 8
					elseif cnt == 3 then
						if item == 'N' then
							_G.ptz.device.parity = 'none'
						elseif item == 'O' then
							_G.ptz.device.parity = 'odd'
						elseif item == 'E' then
							_G.ptz.device.parity = 'even'
						end
					elseif cnt == 4 then
						_G.ptz.device.stopbits = tonumber(item) or 1
					end
					cnt = cnt + 1
				end
			end
			CONFIGS.Submit()
		end

		if _M.ptz_obj then
			local answer = "OK " .. (_G.ptz.device.device:match("/dev/(.+)") or "default") .. " " .. _G.ptz.device.baudrate .. "," .. _G.ptz.device.databits .. ","
			if _G.ptz.device.device.parity == "odd" then
				answer = answer .. "O,"
			elseif _G.ptz.device.device.parity == "even" then
				answer = answer .. "E,"
			else
				answer = answer .. "N,"
			end
			answer = answer .. _G.ptz.device.stopbits .. "\r\n"
			ctx:Send(answer)
		else
			ctx:Send("OK none\r\n")
		end
	elseif cmd == "SER_CLOSE" then
		CONFIGS.BeginUpdate()
		_G.ptz.device.device = "null"
		CONFIGS.Submit()
		ctx:Send("OK\r\n")
	elseif cmd == "PING" then
		ctx:Send("OK PONG\r\n")
	elseif cmd == "SELF" then
		if ctx.idr then
			ctx:Send("OK " .. ctx.idr .. "\r\n" )
		else
			ctx:Send("ERROR Bad connection\r\n")
		end
	elseif cmd == "CLIENTS" then
		local k,v
		local cnt = 0
		local msg = ''
		local matching = tokens[1] or "dead"

		for k,v in pairs(SOCKET_SVC.clients) do
			if v.stage ~= matching then
				cnt = cnt + 1
				msg = msg .. ' '
				if v == ctx then
					msg = msg .. "+"
				end
				msg = msg .. (v.idr or 'unknown' )
			end
		end
		ctx:Send("OK " .. cnt .. msg .. "\r\n" )
	elseif cmd == "KILL" then
		if not tokens[1] then
			ctx:Send( "ERROR Bad argument\r\n" )
		elseif tokens[1] == '#' then
			ctx:Send("OK\r\n")
			ctx:Close()
		elseif tokens[1] == '+' then
			local k,v
			for k,v in pairs(SOCKET_SVC.clients) do
				if v ~= ctx then
					v:Close()
				end
			end
			ctx:Send("OK\r\n")
		elseif tokens[1] == '*' then
			ctx:Send("OK\r\n")
			local k,v
			for k,v in pairs(SOCKET_SVC.clients) do
				v:Close()
			end
		else
			local tgt = SOCKET_SVC.clients[ tokens[1] ]
			if not tgt then
				ctx:Send("ERROR Invalid kill target\r\n" )
			else
				ctx:Send("OK\r\n")
				tgt:Close()
			end
		end
	elseif cmd == "INFO" then
		ctx:Send( "OK " .. SOCKET_SVC.version .. "\r\n" )
	elseif cmd == "BYE" then
		ctx:Send("OK\r\n")
		ctx:Close()
	elseif cmd == "START" then
		ctx.stage = "established"
		ctx.disableTx = false
		ctx.disableRx = false
		if tokens[1] then
			local flag = tokens[1]:upper()
			if flag == "TO" then
				ctx.disableRx = true
			end
			if flag == "RO" then
				ctx.disableTx = true
			end
		end
		ctx:Send("OK\r\n")
	else
		ctx:Send( "ERROR Invalid command\r\n" )
	end
end

----

local function ProcessClientRx( ctx, data )
	if ctx.disableTx then return end
	local ser = _M.ptz_obj and _M.ptz_obj:GetSerialObject() or nil
	if not ser then return end
	ser:write(data)
end

----

local function CreateClientContext( client )
	local peer_ip, peer_port = client:getpeername()
	if peer_port then
		peer_ip = peer_ip .. ":" .. peer_port
	end

	local ctx = {
		idr = peer_ip,
		client = client,
		stage = "prepare",
		txWatcher = nil,
		rxWatcher = nil,
		txBuffer = ''
	}

	function ctx:Close()
		if self.txWatcher then
			if self.txWatcher:is_pending() then
				self.txWatcher:clear_pending()
			end
			self.txWatcher:stop( SOCKET_SVC.loop )
			self.txWatcher = nil
		end

		if self.rxWatcher then
			if self.rxWatcher:is_pending() then
				self.rxWatcher:clear_pending()
			end
			self.rxWatcher:stop( SOCKET_SVC.loop )
			self.rxWatcher = nil
		end

		if self.client then
			self.client:close()
			self.client = nil
		end

		self.stage = "dead"
	end

	function ctx:Send( buf )
		if not buf or #buf == 0 or not self.client then
			return
		end

		self.txBuffer = self.txBuffer .. buf
		if not self.txWatcher then
			self.txWatcher = ev.IO.new( function(loop, watcher)
				local cnt, err = self.client:send( self.txBuffer )
				if not cnt then
					print( "SERIAL_SOCKET: [" .. self.idr .. "] Tx ERROR & CLOSED -", err )
					self:Close()
					return
				end

				-- drop sent bytes.
				self.txBuffer = self.txBuffer:sub(cnt+1)

				if #self.txBuffer == 0 then
					watcher:stop( loop )
					self.txWatcher = nil
				end
			end, self.client:getfd(), ev.WRITE )

			if self.txWatcher then
				self.txWatcher:start( SOCKET_SVC.loop )
			end
		end
	end

	ctx.rxWatcher = ev.IO.new( function( loop, watcher )
		if ctx.stage == "prepare" then
			while true do
				local cmd, err = ctx.client:receive( "*l" )
				if not cmd then
					if err == "timeout" then
						return
					else
						print( "SERIAL_SOCKET: [" .. ctx.idr .. "] Rx ERROR & CLOSED -", err )
						ctx:Close()
						return
					end
				end
				ProcessClientNego( ctx, cmd )
			end
		else
			while true do
				local data, err, part = ctx.client:receive(1024)
				if not data then
					if err == "timeout" then
						if part then
							ProcessClientRx( ctx, part )
						end
						break
					else
						print( "SERIAL_SOCKET: [" .. ctx.idr .. "] Rx ERROR & CLOSED -", err )
						ctx:Close()
						return
					end
				else
					ProcessClientRx( ctx, data )
				end
			end
		end
	end, client:getfd(), ev.READ )

	if not ctx.rxWatcher then
		print( "SERIAL_SOCKET: ERROR - Fail to create client [" .. peer_ip .. "] Rx watcher!" )
		ctx = nil
		return nil
	end

	ctx.rxWatcher:start( SOCKET_SVC.loop )
	return ctx
end

----

local function ServiceListen( loop, watcher )
	-- New accepting, but first clear all dead client
	local k,v
	for k,v in pairs( SOCKET_SVC.clients ) do
		if v.stage == "dead" or not v.client then
			SOCKET_SVC.clients[k] = nil
		end
	end
	
	collectgarbage()

	local client = SOCKET_SVC.svc_socket:accept()
	if not client then
		print( "SERIAL_SOCKET: ERROR - Fail to accept new client" )
		return
	end

	client:settimeout(0)

	local ctx = CreateClientContext( client )
	if ctx then
		SOCKET_SVC.clients[ ctx.idr ] = ctx
	end
end

----

function SOCKET_SVC.Start( port )
	if SOCKET_SVC.svc_socket then
		print( "SERIAL_SOCKET: The service is already started." )
		return false
	end
	
	if not SOCKET_SVC.loop then
		print( "SERIAL_SOCKET: The ev-loop is not initialized so could not start the service." )
		return false
	end

	SOCKET_SVC.svc_socket = socket.bind( "*", port )
	if not SOCKET_SVC.svc_socket then
		SOCKET_SVC.svc_socket = nil
		print( "SERIAL_SOCKET: Fail to bind socket service at port:", port )
		return false
	end

	SOCKET_SVC.svc_socket:settimeout(0)

	SOCKET_SVC.svc_io = ev.IO.new( ServiceListen, SOCKET_SVC.svc_socket:getfd(), ev.READ )
	if not SOCKET_SVC.svc_io then
		print( "SERIAL_SOCKET: Fail to create service socket watcher." )
		SOCKET_SVC.svc_socket:close()
		SOCKET_SVC.svc_socket = nil
		return false
	end

	print( "SERIAL_SOCKET: Service started, listen on port:", port )
	SOCKET_SVC.svc_io:start( SOCKET_SVC.loop )
	return true
end

----

function SOCKET_SVC.Stop()
	if SOCKET_SVC.svc_io then
		if SOCKET_SVC.svc_io:is_pending() then
			SOCKET_SVC.svc_io:clear_pending()
		end
		SOCKET_SVC.svc_io:stop( SOCKET_SVC.loop )
		SOCKET_SVC.svc_io = nil
	end

	local k,v
	for k,v in pairs(SOCKET_SVC.clients) do
		v:Close()
	end

	SOCKET_SVC.clients = {}

	if SOCKET_SVC.svc_socket then
		SOCKET_SVC.svc_socket:close()
		SOCKET_SVC.svc_socket = nil
	end

	print( "SERIAL_SOCKET: Serivce stopped." )
end
---------------------------------------------------

return _M



