-------------------------------------------------
-- Network-manager functions
-------------------------------------------------
local coaf = require "coaf.core"
local thread = require "coaf.thread"
local _dbg = require "coaf.debug"
local lfs = require "lfs"

local _did, special = pcall( require, "nm_special" )
local cfg_tools
if not _did or type(special) ~= "table" then
	special = nil
elseif type(special.GetTools) == "function" then
	cfg_tools = special.GetTools()
end

local IFCONFIG = cfg_tools and cfg_tools.IFCONFIG or "/sbin/ifconfig"
local ROUTE = cfg_tools and cfg_tools.ROUTE or "/sbin/route"
local IP = cfg_tools and cfg_tools.IP or "/sbin/ip"
local DHCLIENT = cfg_tools and cfg_tools.DHCLIENT or "/sbin/dhclient"

local _M = {
	gatheredDNS = {},
	dhclients = {}
}

-------------------------------------------------

local function dhclientWaitter(threadObj, ifname, pid)
	coaf.WaitPID(pid)
	local opt = _M.dhclients[ifname]
	if opt then
		opt.waitThread = nil
		opt.pid = nil
		opt.pid_file = "/var/run/dhclient-" .. ifname .. ".pid"
	end
end

-------------------------------------------------

function _M.DHCPClientStarted( ifname )
	if _M.dhclients[ifname] then
		return true
	else
		return false
	end
end

function _M.KillDHCPClient( ifname )
	local opt = _M.dhclients[ifname]
	if opt then
		if opt.pid then
			os.execute( "/bin/kill " .. opt.pid )
		end

		if opt.pid_file then
			os.execute( "/bin/kill $(cat " .. opt.pid_file .. ")" )
		end

		_M.dhclients[ifname] = nil
	end
end

-------------------------------------------------

-- Enumerate all network interface names
function _M.EnumNetworks()
	local intfs = {}
	local ifname
	local f, l
	for ifname in lfs.dir( "/sys/class/net" ) do
		if ifname ~= "." and ifname ~= ".." and ifname ~= "lo" then
			table.insert( intfs, ifname )
			f = io.open( "/sys/class/net/" .. ifname .. "/carrier", "r" )
			if f then
				l = f:read("l")
				if l then
					intfs[ifname] = tonumber(l)
				end
				f:close()
			end
		end
	end
	return intfs
end

-- Get network linking status
function _M.GetLinkStatus(ifname)
	local f = io.open( "/sys/class/net/" .. ifname .. "/operstate", "r" )
	if not f then
		return "notexist"
	end
	local s = f:read("l")
	f:close()

	if not s then
		return "notexist"
	elseif s == "down" then
		f = io.open( "/sys/class/net/" .. ifname .. "/carrier", "r" )
		if not f then
			return "disabled"
		else
			s = f:read("l")
			f:close()
			if not s then
				return "disabled"
			else
				return "down"
			end
		end
	else
		f = io.open( "/sys/class/net/" .. ifname .. "/carrier", "r" )
		if not f then
			return "error"
		else
			s = f:read("l")
			f:close()
			if not s then
				return "error"
			elseif tonumber(s) == 1 then
				return "up"
			else
				return "down"
			end
		end
	end
end

-- Get link details
function _M.GetLinkDetail(ifname)
	local hw_ifname = ifname:match( "(.+):.*$" )
	if not hw_ifname then
		hw_ifname = ifname
	end

	local status = _M.GetLinkStatus(hw_ifname)
	local ret = {
		status = status,
		address = "0.0.0.0",
		netmask = "255.255.255.255",
		gw = "",
		mac = ""
	}

--	if status == "up" then
--[[ Parse this:
eth1      Link encap:Ethernet  HWaddr 6E:39:D0:5B:41:FB  
          inet addr:192.168.1.168  Bcast:192.168.1.255  Mask:255.255.255.0
          BROADCAST MULTICAST  MTU:1500  Metric:1
          RX packets:623 errors:0 dropped:0 overruns:0 frame:0
          TX packets:83 errors:0 dropped:0 overruns:0 carrier:0
          collisions:0 txqueuelen:1000 
          RX bytes:63174 (61.6 KiB)  TX bytes:19678 (19.2 KiB)
          Interrupt:55 Base address:0x4000 
]]
		local f = io.popen( IFCONFIG .. " " .. ifname, "r" )
		if f then
			local matchStr
			local ip, mask
			local l = f:read("l")
			while l do
				matchStr = l:match("HWaddr%s+(%x%x:%x%x:%x%x:%x%x:%x%x:%x%x)")
				if matchStr then
					ret.mac = matchStr
				else
					ip, mask = l:match("inet addr:([0-9]+%.[0-9]+%.[0-9]+%.[0-9]+).*Mask:([0-9]+%.[0-9]+%.[0-9]+%.[0-9]+)")
					if ip then
						ret.address = ip
						ret.netmask = mask
					end
				end	
				l = f:read("l")
			end
			f:close()
		end

--[[ Parse this:
default via 10.168.0.94 dev eth0 
default via 10.168.0.94 dev eth0 src 10.168.0.93
10.168.0.64/27 dev eth0  src 10.168.0.93 
192.168.1.0/24 dev eth1  src 192.168.1.168 
192.168.2.0/24 dev eth0  src 192.168.2.131 
]]
		f = io.popen( IP .. " route", "r" )
		if f then
			local l = f:read("l")
			local gw
			while l do 
				gw = l:match( "default.*via%s*([0-9]+%.[0-9]+%.[0-9]+%.[0-9]+).*dev%s*" .. hw_ifname )
				if gw then
					if hw_ifname ~= ifname then --Alias checking
						if l:match("src%s*([0-9]+%.[0-9]+%.[0-9]+%.[0-9]+)") == ret.address then
							ret.gw = gw
							break
						end
					else
						ret.gw = gw
						break
					end
				end
				l = f:read("l")
			end
			f:close()
		end
	-- end
	return ret
end

-- Update DNS configurations
function _M.UpdateDNS()
	local mergeDNS = {}
	local k, cfg
	local hasDHCP
	for k, cfg in pairs(_M.gatheredDNS) do
		if type(cfg) == "table" then
			local i,item
			for i,item in ipairs(cfg) do
				if not mergeDNS[ item ] then
					mergeDNS[ item ] = true
					table.insert( mergeDNS, item ) -- To keep the order
				end
			end
		elseif cfg == "dhcp" then
			hasDHCP = true
		end
	end
	
	if #mergeDNS > 0 or not hasDHCP then -- if no DNS set but has DHCP, keep the resolv.conf
		local f = io.open( "/etc/resolv.conf", "w" )
		if not f then
			return false, "Fail to open resolv.conf"
		else
			for k,cfg in ipairs(mergeDNS) do
				f:write( "nameserver " .. cfg .. "\n" )
			end
			f:close()
			return true
		end
	end
end

function _M.UpdateRouter(ifname,gw)
	local ret = _M.GetLinkDetail(ifname)
	if not ret or ret.status ~= "up" or ret.address == "0.0.0.0" then
		if ret.gw ~= "" then
			os.execute( ROUTE .. " del default dev " .. ifname .. " > /dev/null 2>&1" )
		end
	else
		gw = gw or ""
		if ret.gw ~= gw then
			os.execute( ROUTE .. " del default dev " .. ifname .. " > /dev/null 2>&1" )
			-- With valid gateway address
			if gw ~= "" and gw ~= "0.0.0.0" and gw ~= "255.255.255.255" then
				local metric = 0
				local isEth = ifname:match("^eth")
				if isEth then -- If eth0 ~ ethN, the metric is 0,100,200,...
					local ethNum = tonumber( ifname:match("eth(%d+)") )
					if ethNum and ethNum > 0 then
						metric = ethNum*100
					end
					local alias = tonumber( ifname:match(".+:(%d+)") )
					if alias then
						metric = metric + (alias+1)*10
					end
				else -- Other interfaces, default metric is 1000
					metric = 1000 -- as default metric
				end
				os.execute( ROUTE .. " add default gw " .. gw .. " dev " .. ifname .. ( metric == 0 and "" or (" metric " .. metric) ) )
			end
		end
	end
end

-- Configure and start DHCP service for specified interface.
function _M.StartDHCPService(ifname, cfg)
	if special and special.StartDHCPService then
		--local nu = require "kiloview.netUtil"
		local self_ip = cfg.address
		--local self_prefix = nu.ip4MaskToPrefix(cfg.netmask)
		local self_prefix = 24

		local ip_prefix = self_ip:match("(.+)%..*$")
		local ip_start = tonumber(cfg.dhcp_server.dynamic_start)
		local ip_end = tonumber(cfg.dhcp_server.dynamic_end)
		if not ip_start or ip_start < 1 then
			ip_start = 100
		end
		if not ip_end or ip_end > 255 then
			ip_end = 255
		end
		if ip_end < ip_start then
			local tmp = ip_end
			ip_end = ip_start
			ip_start = tmp
		end
		ip_start = ip_prefix .. "." .. ip_start
		ip_end = ip_prefix .. "." .. ip_end

		local gw
		if cfg.dhcp_server.gw == "self" then
			gw = cfg.address
		elseif not cfg.dhcp_server.gw or cfg.dhcp_server.gw == "" or cfg.dhcp_server.gw == "0.0.0.0" then
			gw = ""
		else
			gw = cfg.dhcp_server.gw
		end

		--TODO: DNS gather.

		special.StartDHCPService( ifname, self_ip, self_prefix, {
			dynamic_start = ip_start,
			dynamic_end = ip_end,
			gw = gw,
			dns = {} --TODO.
		})
	end
end

-- Configure ethernet interface.
function _M.ConfigEthernet(ifname, cfg, forAlias)
	if not cfg then
		return nil, "Invalid network config options"
	end
	if ifname == "lo" or not lfs.attributes( "/sys/class/net/" .. (forAlias or ifname or "invalid") ) then
		return nil, "Invalid network interface"
	end
	
	if not forAlias and special and special.StopDHCPService then
		special.StopDHCPService(ifname)
	end

	os.execute( ROUTE .. " del default dev " .. ifname .. " > /dev/null 2>&1" )
	local should_cfg_router = false
	local should_cfg_dns = false
	if cfg.enable then
		if cfg.mac and cfg.mac:match("%x%x:%x%x:%x%x:%x%x:%x%x:%x%x") then
			os.execute( IFCONFIG .. " " .. ifname .. " hw ether " .. cfg.mac )
		end

		os.execute( IFCONFIG .. " " .. ifname .. " up" )
		if cfg.method == "dhcp" then
			_M.gatheredDNS[ ifname ] = "dhcp" -- Mark it's DHCP

			if not _M.DHCPClientStarted( ifname ) then
				local st, posix_unistd = pcall(require,"posix.unistd")
				local pid = coaf.Fork( function() --Since dhclient may block the execution.
					if st and posix_unistd then
						if type(DHCLIENT) == "string" then
							posix_unistd.execp( DHCLIENT, { [0] = DHCLIENT, "-pf", "/var/run/dhclient-" .. ifname .. ".pid", ifname })
						elseif type(DHCLIENT) == "function" then
							local name, args = DHCLIENT( "exec", ifname, "/var/run/dhclient-" .. ifname .. ".pid" )
							posix_unistd.execp( name, args )
						end
						--If arrive here, error is occused.
						os.exit(1)
					else
						if type(DHCLIENT) == "string" then
							os.execute( DHCLIENT .. " " .. ifname )
						elseif type(DHCLIENT) == "function" then
							local cmd = DHCLIENT( "cmd", ifname )
							os.execute( cmd )
						end
						os.exit(0)
					end
				end )

				local waitThread = thread.New( dhclientWaitter )
				waitThread:Start(ifname, pid)
				_M.dhclients[ ifname ] = { waitThread = waitThread, pid = pid }
				coaf.Sleep(1) 
				--TODO: Improve this!!! for DHCP case, some dhcp client may clear all IP settings.
				--      so I must sleep 1 seconds for next IP assign (e.g. Alias). But this is not
				--      a very good method.
			end
		else
			if not forAlias then _M.KillDHCPClient(ifname) end

			if cfg.address and cfg.address ~= "" then
				os.execute( IFCONFIG .. " " .. ifname .. " " .. cfg.address .. " " .. (cfg.netmask and cfg.netmask ~= "" and ("netmask " .. cfg.netmask) or "") )
				should_cfg_router = true
				should_cfg_dns = true
			end
		end
	else
		os.execute( IFCONFIG .. " " .. ifname .. " down > /dev/null 2>&1" )	--down interface, so all router items relied should be deleted.
		if forAlias then
			return true
		else
			_M.KillDHCPClient(ifname)
			_M.gatheredDNS[ ifname ] = nil
			_M.UpdateDNS()
			return true
		end
	end

	if should_cfg_router then
		_M.UpdateRouter( ifname, cfg.gw )
	end

	if should_cfg_dns and not forAlias then
		if type(cfg.dns) == "table" and #cfg.dns > 0 then
			_M.gatheredDNS[ ifname ] = cfg.dns
		else
			_M.gatheredDNS[ ifname ] = nil
		end
		_M.UpdateDNS()
	end

	if not forAlias and cfg.method ~= "dhcp" and cfg.dhcp_server and cfg.dhcp_server.enable then
		_M.StartDHCPService(ifname, cfg)
	end

	return true
end

-- Configure ethernet alias.
function _M.ConfigEthernetAlias(ifname, cfg)
	if not cfg then
		return false, "Invalid ethernet alias config"
	end
	--_dbg.Info( "NM", "Configure ethernet alias for:", ifname .. ":0" )
	local r1 = _M.ConfigEthernet( ifname .. ":0", cfg, ifname )
	local cfg2 = {
		enable = cfg.enable and cfg.address2 ~= "" and cfg.address2 ~= "0.0.0.0",
		address = cfg.address2,
		netmask = cfg.netmask2,
		gw = cfg.gw2
	}
	--_dbg.Info( "NM", "Configure ethernet alias for:", ifname .. ":1" )
	local r2 = _M.ConfigEthernet( ifname .. ":1", cfg2, ifname )
	return r1 or r2
end

-- Configure interlink network.
function _M.ConfigInterlink(cfg)
	if cfg and cfg.ifname then
		return _M.ConfigEthernet( cfg.ifname, cfg )
	else
		return false, "Invalid configure for interlink"
	end
end

------------------------------------------------

function _M.Init()
	os.execute( IFCONFIG .. " lo up" )
end

return _M

