local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local event = require "coaf.event"

require "kmCore"

local _M = {
	file_location = "/tmp/hls",
	recorder_group = nil,
	created_sessions = {
	}
}

----------------------------------------------

local function tsFragmentEvent( topic, msg, msgType )
	if topic and msg and msgType == "table" and msg.action == "close-file" then
		--Topic may be 'ts-fragment.<session-name>':
		local session_name = topic:match(".-%.(.+)$")
		if not session_name then return end
		local session = _M.created_sessions[ session_name ]
		if not session or not session.files or not session.streamer then 
			return 
		end
		--Record files.
		table.insert( session.files, {
			sequence = session.sequence,
			file = msg.file,
			size = msg.size_b,
			duration = msg.time_msec
		} )
		session.sequence = session.sequence + 1
		
		--If too many file, remove oldest one.
		if #session.files > session.streamer.maxSegments then
			os.execute( "/bin/rm -Rf " .. session.streamer.location .. "/" .. session.files[1].file )
			table.remove( session.files, 1 )
		end
	end
end

----------------------------------------------

local hls_meta = {}

function hls_meta:AddSession( idr, mediaSource, session_name, options )
	if not mediaSource or not session_name or session_name == "" then
		_dbg.Error( "HLS", "Invalid HLS arguments (MediaSource/Session name ...). Stream:", self.stream_id, ", IDR:", idr )
		return nil
	end

	if self.session then
		_dbg.Warn( "HLS", "The HLS session is already created. Stream:", self.stream_id, ", IDR:", idr )
		return self.session
	end

	if _M.created_sessions[ session_name ] then
		_dbg.Error( "HLS", "The session '" .. session_name .. "' is already created by others! Stream:", self.stream_id, ", IDR:", idr )
		return nil
	end

	
	self.session = _M.recorder_group:AddSession( "hls:" .. self.stream_id .. ":" .. idr, mediaSource, 4*1024*1024, 256*1024 )
	if not self.session then
		_dbg.Error( "HLS", "Fail to create HLS ts fragment session for Stream:", self.stream_id, ", IDR:", idr )
		return nil
	end

	self.location = _M.file_location .. "/" .. session_name
	self.segmentTime = options and options.segmentTime or 5000
	self.segmentTime = math.ceil(self.segmentTime/1000) --To second(s)
	self.maxSegments = options and options.maxSegments or 6

	os.execute( "/bin/mkdir -p " .. self.location )
	os.execute( "/bin/rm -Rf " .. self.location .. "/*.ts" )

	_M.recorder_group:SetOption( self.session, {
		path = _M.file_location .. "/" .. session_name,
		type = "hls", --NOTE: The new version of moon_streamer should set type to 'hls'(not 'ts') to correct the timestamp problem.
		prefix = "SEG",
		multiple = true,
		maxSize = "0",
		maxTime = '' .. self.segmentTime, --NOTE: This parameter should in 'second' unit.
		muxrate = -1
	})

	_M.created_sessions[ session_name ] = {
		streamer = self,
		sequence = 0,
		files = {}
	}--To save file lists.

	_dbg.Info( "HLS", "Created HLS session, for Stream:", self.stream_id, ", IDR:", idr, ", Session:", session_name )
	self.started = false
	self.session_idr = idr
	self.session_name = session_name
	return self.session
end

function hls_meta:RemoveSession()
	if self.session then
		self:StopSession()
		_M.created_sessions[ self.session_name ] = nil
		_M.recorder_group:RemoveSession( self.session )
		_dbg.Info( "HLS", "HLS session is destroied, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", Session:", self.session_name )
		self.segmentTime = nil
		self.maxSegments = nil
		self.session = nil
		self.session_idr = nil
		self.session_name = nil
		collectgarbage()

		os.execute( "/bin/rm -Rf " .. self.location .. "/*.ts" )
		self.location = nil
	end
	return true
end

function hls_meta:StartSession()
	if self.session and not self.started then
		self.event_topic = _G.streamer_event:Subscribe( "fragment." .. self.session_name, tsFragmentEvent )
		_M.recorder_group:HandleEvent( self.session, "streamer", "fragment." .. self.session_name )
		_M.recorder_group:StartSession( self.session )
		self.started = true
		_dbg.Info( "HLS", "HLS session is started, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", Session:", self.session_name )
	end
	return self.started
end

function hls_meta:StopSession()
	if self.session and self.started then
		_M.recorder_group:StopSession( self.session )
		if self.event_topic then
			_G.streamer_event:Unsubscribe( self.event_topic )
			self.event_topic = nil
		end
		self.started = false
		_dbg.Info( "HLS", "HLS session is stopped, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", Session:", self.session_name )
	end
	return true
end

function hls_meta:Destroy()
	self:RemoveSession()
end

hls_meta.__index = hls_meta
----------------------------------------------

function _M.GetStreamer( stream_id )
	stream_id = tonumber(stream_id)
	stream_id = string.format( "%02d", stream_id )

	if not _M.recorder_group then
		_M.recorder_group = KM_STREAMING.Recorder.createNew( "hls-ts-fragment" )
		if not _M.recorder_group then
			_dbg.Error( "HLS", "Fail to create TS fragment recorder group!" )
			return nil
		end
	end

	local streamer = {
		stream_id = stream_id,
	}

	return setmetatable( streamer, hls_meta )
end

----------------------------------------------
-- Get m3u8 contents (in table format).
-- The web service should get this, correct file url location, then
-- concat them together to a m3u8 file.
----------------------------------------------

function _M.GetM3U8( session_name )
	if not session_name or not _M.created_sessions[ session_name ] then
		return nil
	end

	local files = _M.created_sessions[session_name].files
	if not files or #files < 1 then
		return nil
	end

	local r = {
		"#EXTM3U",
		"#EXT-X-VERSION:3",
		"#EXT-X-ALLOW-CACHE:YES",
		"#EXT-X-MEDIA-SEQUENCE:", --Concat later.
		"#EXT-X-TARGETDURATION:", --Concat later.
	}

	local seq = files[1].sequence or 0
	r[4] = r[4] .. seq
	if seq == 0 then
		table.insert(r, "#EXT-X-DISCONTINUITY")
	end

	local max_dur = 0
	local i
	for i=1, #files do
		table.insert(r, "#EXTINF:" .. math.floor(files[i].duration/1000) .. "." .. string.format("%03d", files[i].duration%1000))
		if files[i].duration > max_dur then
			max_dur = files[i].duration
		end
		table.insert(r, files[i].file)
	end

	if max_dur <= 0 or max_dur < _M.created_sessions[session_name].streamer.segmentTime*1000 then
		max_dur = _M.created_sessions[session_name].streamer.segmentTime*1000
	end
--	r[5] = r[5] .. math.floor(max_dur/1000) .. "." .. string.format("%03d", max_dur%1000)
--	r[5] = r[5] .. math.abs(math.floor(-max_dur/1000))
--	NOTE: iOS does not support "xxx.xxx" style.
	r[5] = r[5] .. math.ceil(max_dur/1000)
	return r
end
----------------------------------------------

return _M


