local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local timedatectl = require "kiloview.systemd-timedate"
local thread = require "coaf.thread"

require "kmCore"

local _M = {
	each_pusher_session_limits = 4,
	pusher_groups = {
		["00"] = {},
		["01"] = {}
	},
}

local function ntp_state()
	local timeinfo = timedatectl.GetTimeDateInfo()
	local ntp_en = timeinfo and timeinfo["ntp_enabled"] or "no"
	local ntp_sync = timeinfo and timeinfo["ntp_sync"] or "no"
	return ntp_en, ntp_sync
end

----------------------------------------------
-- This is old version RTMP pusher, if all tested, net_pusher will
-- replace this pusher so it will be noused.

local rtmp_pusher_meta = {}

local function session_pending_task(thrObj, rtmp_pusher, mediaSource, targetAddr, options)
	local ready = false
	while not thrObj.terminated do
		coaf.Sleep(2)
		local _ignore, sync = ntp_state()
		if sync == "yes" then
			_dbg.Info( "RTMP-PUSH", "NTP synchronized! Creating the RTMP session..." )
			ready = true
			break
		end
	end
	
	rtmp_pusher.session_backend_task = nil

	if ready then
		rtmp_pusher.session = rtmp_pusher.group.group:AddSession( mediaSource, targetAddr, options )
		if not rtmp_pusher.session then
			_dbg.Error( "RTMP-PUSH", "Fail to create RTMP pusher session for Stream:", rtmp_pusher.stream_id, ", IDR:", rtmp_pusher.session_idr )
			return false
		end
	
		_dbg.Info( "RTMP-PUSH", "Created RTMP push session, for Stream:", rtmp_pusher.stream_id, ", IDR:", rtmp_pusher.session_idr, ", URL:", targetAddr )
		return true
	else
		_dbg.Warn( "RTMP-PUSH", "Session backend task is terminated." )
		return false
	end
end

function rtmp_pusher_meta:AddSession( idr, mediaSource, targetAddr, options )
	if not mediaSource or not targetAddr or targetAddr == "" then
		_dbg.Error( "RTMP-PUSH", "Invalid RTMP pusher arguments (MediaSource/TargetAddress ...). Stream:", self.stream_id, ", IDR:", idr )
		return nil
	end

	if self.session or self.session_backend_task then
		_dbg.Warn( "RTMP-PUSH", "The RTMP pusher session is already created. Stream:", self.stream_id, ", IDR:", idr )
		return self.session or self.session_backend_task
	end

	--Now, parsing the URL to get some arguments.
	local protocol, orig_url_tail
	local user_in_url, pass_in_url

	local s

	protocol, s = targetAddr:match( "(%w+)://(.+)$" )
	if not protocol then
		_dbg.Error( "RTMP-PUSH", "Invalid target URL of:", targetAddr, "(Stream:", self.stream_id, ", IDR:", idr, ")" )
		return nil
	end
	protocol = protocol:lower()
	user_in_url, pass_in_url = s:match( "([^/]-):([^/]*)@.+$" )
	if not user_in_url then
		local s1,s2 = s:match("(.-)/(.+)")
		if s1 then
			user_in_url = s1:match( "([^/]-)@./+$" )
		end
	end
	if user_in_url then
		orig_url_tail = s:match( ".-@(.+)$" )
	else
		orig_url_tail = s
	end

	if type(options) ~= "table" then
		options = {}
	end

	if user_in_url and not options["user"] then
		options["user"] = user_in_url
	end
	if pass_in_url and not options["pass"] then
		options["pass"] = pass_in_url
	end

	--Formal the URL
	targetAddr = protocol .. "://" .. orig_url_tail

	local ntpTag = tonumber(options["enableNtp"])
	if ntpTag and ntpTag > 0 then
		local ntp_en, ntp_sync
		ntp_en, ntp_sync = ntp_state()
		if ntp_en ~= "yes" then --Not set NTP, so disable it.
			_dbg.Warn( "RTMP-PUSH", "You are setting NTP timestamp but the NTP is not configured! for Stream:", self.stream_id, ", IDR:", idr, ", URL:", targetAddr )
			options["enableNtp"] = 0
		elseif ntp_sync ~= "yes" then
			self.session_backend_task = thread.New( session_pending_task )
			if self.session_backend_task then

				self.session_idr = idr
				self.formalUrl = targetAddr
				self.savedOptions = options
				self.savedMediaSource = mediaSource

				_dbg.Info( "RTMP-PUSH", "Backend wait NTP synchronization and create RTMP push session, for Stream:", self.stream_id, ", IDR:", idr, ", URL:", targetAddr )
				self.session_backend_task:Start(self, mediaSource, targetAddr, options)
			end
			return self.session_backend_task
		end
	end

	self.session = self.group.group:AddSession( mediaSource, targetAddr, options )
	if not self.session then
		_dbg.Error( "RTMP-PUSH", "Fail to create RTMP pusher session for Stream:", self.stream_id, ", IDR:", idr )
		return nil
	end

	_dbg.Info( "RTMP-PUSH", "Created RTMP push session, for Stream:", self.stream_id, ", IDR:", idr, ", URL:", targetAddr )
	self.session_idr = idr
	self.formalUrl = targetAddr
	self.savedOptions = options
	self.savedMediaSource = mediaSource

	return self.session
end

function rtmp_pusher_meta:RemoveSession()
	if self.session_backend_task then
		_dbg.Info( "RTMP-PUSH", "Destroying the backend task, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.formalUrl )
		self.session_backend_task:Terminate()
		self.session_backend_task = nil
	end

	if self.session then
		self.group.group:RemoveSession( self.session )
		_dbg.Info( "RTMP-PUSH", "RTMP push session is destroied, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.formalUrl )
	end

	self.session = nil
	self.session_idr = nil
	self.formalUrl = nil
	self.savedOptions = nil
	self.savedMediaSource = nil
	collectgarbage()

	return true
end

function rtmp_pusher_meta:StartSession()
	if self.session or self.session_backend_task then
		return true
	end
	if not self.formalUrl or not self.savedMediaSource then
		_dbg.Warn( "RTMP-PUSH", "No session created for Stream:", self.stream_id )
		return false
	end

	local ntpTag = tonumber(self.savedOptions["enableNtp"])

	if ntpTag and ntpTag > 0 then
		local ntp_en, ntp_sync
		ntp_en, ntp_sync = ntp_state()
		if ntp_en ~= "yes" then --Not set NTP, so disable it.
			_dbg.Warn( "RTMP-PUSH", "You are setting NTP timestamp but the NTP is not configured! for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.formalUrl )
			options["enableNtp"] = 0
		elseif ntp_sync ~= "yes" then
			self.session_backend_task = thread.New( session_pending_task )
			if self.session_backend_task then
				_dbg.Info( "RTMP-PUSH", "Backend wait NTP synchronization and start RTMP push session, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.formalUrl )
				self.session_backend_task:Start(self, self.savedMediaSource, self.formalUrl, self.savedOptions)
			end
			return self.session_backend_task
		end
	end

	self.session = self.group.group:AddSession( self.savedMediaSource, self.formalUrl, self.savedOptions )
	if not self.session then
		_dbg.Error( "RTMP-PUSH", "Fail to create RTMP pusher session for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.formalUrl )
		return false
	end
	_dbg.Info( "RTMP-PUSH", "RTMP push session is started, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.formalUrl )
	return true
end

function rtmp_pusher_meta:StopSession()
	if self.session_backend_task then
		_dbg.Info( "RTMP-PUSH", "Destroying the backend task, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.formalUrl )
		self.session_backend_task:Terminate()
		self.session_backend_task = nil
	end

	if self.session then
		self.group.group:RemoveSession( self.session )
		self.session = nil
		_dbg.Info( "RTMP-PUSH", "RTMP push session is stopped, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.formalUrl )
		collectgarbage()
	end
	return true
end

function rtmp_pusher_meta:Destroy()
	self:RemoveSession()
	if self.group.counter > 0 then
		self.group.counter = self.group.counter - 1
	end
end

rtmp_pusher_meta.__index = rtmp_pusher_meta

----------------------------------------------

function _M.GetStreamer( stream_id )
	stream_id = tonumber(stream_id)
	stream_id = string.format( "%02d", stream_id )

	local fetched_group
	local pgroups = _M.pusher_groups[ stream_id ]
	if not pgroups then
		_dbg.Error( "RTMP-PUSH", "Invalid stream id of:", stream_id )
		return nil
	end

	local i
	for i=1, #pgroups do
		if pgroups[i].counter < _M.each_pusher_session_limits then
			pgroups[i].counter = pgroups[i].counter + 1
			fetched_group = pgroups[i]
			break
		end
	end

	if not fetched_group then
		local idr = "rtmp:" .. stream_id .. ":" .. #pgroups + 1
		local rtmp_pusher_group = KM_STREAMING.Rtmp.Pushing.createNew( idr )
		if not rtmp_pusher_group then
			_dbg.Error( "RTMP-PUSH", "Fail to create RTMP pusher for stream:", stream_id )
			return nil
		else
			_dbg.Info( "RTMP-PUSH", "A new RTMP push group '" .. idr .. "' is created." )
		end
		fetched_group = {
			counter = 1,
			group = rtmp_pusher_group
		}
		table.insert(pgroups, fetched_group)
	end

	local pusher = {
		stream_id = stream_id,
		group = fetched_group
	}

	return setmetatable( pusher, rtmp_pusher_meta )
end

----------------------------------------------
return _M


