-- The SERVER global module
require "features"
require "linuxUsers"
local client = require "coaf.client"
local _dbg = require "coaf.debug"
local cfgM = require "kiloview.configManager"

local NGINX = "/usr/local/openresty/nginx/sbin/nginx"
local NGINX_CONF_LOC = "/usr/local/openresty/nginx/conf"
local NGINX_CONF_TEMPLATE = "nginx.conf.default"
local FLAG_HTTP_PORT = "@HTTP_PORT@"
local FLAG_HTTPS_PORT = "@HTTPS_PORT@"
local FLAG_ONVIF_PORT = "@ONVIF_PORT@"

local NGINX_CERT_KEY = "server.key"
local NGINX_CERT_CRT = "server.crt"


-- A special way to special languages.
local r, forceLangs
r, forceLangs = pcall( require, "kiloview.forceLangs" )
if not r then
	forceLangs = nil
end

local _M = {
	config_location = "/usr/local/openresty/nginx/conf",
	config_file = "nginx.conf",
	auth_file = "htpasswd"
}

local MAX_USER_ALLOW = 10
local MAX_ROLE_ALLOW = 10

----------------------------------------------------------

local function GetCoAFProxy( objPath, service, loc, timeout )
	local proxy
	local cli
	cli = client.New( service, loc or "local", nil, timeout or 3 ) --3 seconds timeout.
	if not cli then
		_dbg.Error( "WEB_DAEMON", "Fail to create connection to service: " .. (loc or "local") .. "/" .. (service or "nil") )
		return nil
	end
	proxy = cli:CreateProxy(objPath, true) --Exclusive
	if not proxy then
		_dbg.Error( "WEB_DAEMON", "Fail to create object proxy for '" .. (objPath or "unknown") .. "'" )
		return nil
	end
	return proxy
end

----------------------------------------------------------
local function genPassword(user,pass)
	local rfile = io.popen( "/usr/bin/openssl passwd -crypt '" .. pass .. "'", "r" )
	if rfile then
		local ln = rfile:read("*l")
		local r,st,code = rfile:close()
		if r and st == "exit" and code == 0 and ln then
			return user .. ":" .. ln
		else
			return nil
		end
	else
		return nil
	end
end

_M.UpdateAuth = function()
	if not _G.webConfig then
		return false
	else
		local file = io.open( _M.config_location .. "/super_users.default", "r" )
		local super_users = {}
		if file then
			local ln = file:read("*l")
			while ln do
				if not ln:match("^[ \t]*#") then --not comments.
					local u,k = ln:match("^[ \t]*(.-)[ \t]*:[ \t]*(.-)[ \t\r]*$")
					if u and k then
						super_users[u] = k
					end
				end
				ln = file:read("*l")
			end
			file:close()
		end

		file = io.open( _M.config_location .. "/" .. _M.auth_file, "w" )
		if file then
			for u,k in pairs(super_users) do
				file:write(u .. ":" .. k .. "\n")
			end
			local r
			if not super_users["admin"] then
				r = genPassword( "admin", _G.webConfig.adminUsers.admin or "admin" )
				if r then
					file:write( r )
					file:write( "\n" )
				end
			end

			if not super_users["administrator"] then
				r = genPassword( "administrator", _G.webConfig.adminUsers.administrator or "abc123!@#" )
				if r then
					file:write( r )
					file:write( "\n" )
				end
			end

			if not super_users["root"] then
				r = genPassword( "root", _G.webConfig.adminUsers.root or "Ki1oWeIv15900d" )
				if r then
					file:write( r )
					file:write( "\n" )
				end
			end

			local i
			for i=1,MAX_USER_ALLOW do
				if _G.webConfig.users[i].username ~= "" and not super_users[_G.webConfig.users[i].username] then
					r = genPassword( _G.webConfig.users[i].username,  _G.webConfig.users[i].password )
					if r then
						file:write( r )
						file:write( "\n" )
					end
				end
			end
			file:close()
			return true
		else
			return false
		end
	end
end
----------------------------------------------------------
_M.CheckPrivileges = function( user, ...)
	local privs = {...}
	local rslt = {}
	local privOpt = nil

	if user == "admin" then
		privOpt = _G.webConfig.adminPrivileges
	elseif user == "administrator" or user == "root" then
		local i
		for i=1, #privs do
			table.insert( rslt, true )
		end
		return table.unpack(rslt)
	elseif user ~= "" then
		local i, role
		for i=1,MAX_USER_ALLOW do
			if _G.webConfig.users[i].username == user then
				role = _G.webConfig.users[i].role
				break
			end
		end

		if not role then
			return nil
		elseif role == "" or role == "default" then
			privOpt = _G.webConfig.defaultPrivileges
		elseif role == "admin" then
			privOpt = _G.webConfig.adminPrivileges
		else
			for i=1, MAX_ROLE_ALLOW do
				if _G.webConfig.roles[i].name == role then
					privOpt = _G.webConfig.roles[i].privileges
					break
				end
			end
		end
	end

	if not privOpt then
		return nil
	else
		local i,x
		for i=1, #privs do
			x = privOpt[ privs[i] ]
			if x and x ~= 0 then
				table.insert( rslt, true )
			else
				table.insert( rslt, false )
			end
		end

		return table.unpack(rslt)
	end
end

----------------------------------------------------------
_M.ChangeAdmin = function( user, pass )
	if (user == "admin" or user == "administrator") and type(pass) == "string" then
		_G.CONFIGS.BeginUpdate()
		_G.webConfig.adminUsers[user] = pass
		_G.CONFIGS.Submit()
		_M.UpdateAuth()
		return true
	else
		return false, "invalid"
	end
end
----------------------------------------------------------
_M.AddUser = function( user, pass, defLangs, role )
	if user == "admin" or user == "administrator" or user == "root" then
		return false, "reserved"
	elseif type(user) ~="string" or type(pass) ~= "string" then
		return false, "invalid"
	elseif user == "" then
		return false, "invalid"
	else
		local i
		local pos
		for i=1,MAX_USER_ALLOW do
			if _G.webConfig.users[i].username == user then
				return false, "exist"
			elseif _G.webConfig.users[i].username == "" and not pos then
				pos = i
			end
		end
		if not pos then
			return false, "full"
		else
			_G.CONFIGS.BeginUpdate()
			_G.webConfig.users[pos].username = user
			_G.webConfig.users[pos].password = pass
			_G.webConfig.users[pos].defaultLanguage = defLangs or ""
			_G.webConfig.users[pos].role = role or "default"
			_G.CONFIGS.Submit()
			_M.UpdateAuth()
			return true
		end
	end
end
----------------------------------------------------------
_M.GetUsers = function()
	local results={}
	local i
	for i=1,MAX_USER_ALLOW do
		if _G.webConfig.users[i].username ~= "" then
			table.insert( results, { 
				username = _G.webConfig.users[i].username,
				password = _G.webConfig.users[i].password,
				defaultLanguage = _G.webConfig.users[i].defaultLanguage,
				role = _G.webConfig.users[i].role
			})
		end
	end
	return results
end
----------------------------------------------------------
_M.DelUser = function( user )
	if user == "admin" or user == "administrator" or user == "root" then
		return false, "reserved"
	elseif type(user) ~="string" or user == "" then
		return false, "invalid"
	else
		local i
		local pos
		for i=1,MAX_USER_ALLOW do
			if _G.webConfig.users[i].username == user then
				pos = i
				break
			end
		end
		if not pos then
			return false, "notexist"
		else
			_G.CONFIGS.BeginUpdate()
			_G.webConfig.users[pos].username = ""
			_G.webConfig.users[pos].password = ""
			_G.webConfig.users[pos].defaultLanguage = ""
			_G.webConfig.users[pos].role = ""
			_G.CONFIGS.Submit()
			_M.UpdateAuth()
			return true
		end
	end	
end
----------------------------------------------------------
_M.ChangeUser = function( user, pass, defLangs, role )
	if user == "admin" or user == "administrator" or user == "root" then
		if pass == "~dont~change" then
			return true
		else
			return _M.ChangeAdmin( user, pass )
		end
	elseif type(user) ~="string" or user == "" then
		return false, "invalid"
	else
		local i
		local pos
		for i=1,MAX_USER_ALLOW do
			if _G.webConfig.users[i].username == user then
				pos = i
				break
			end
		end
		if not pos then
			return false, "notexist"
		else
			_G.CONFIGS.BeginUpdate()
			if type(pass) == "string" and pass ~= "~dont~change" then
				_G.webConfig.users[pos].password = pass
			end
			if type(defLangs) == "string" then
				_G.webConfig.users[pos].defaultLanguage = defLangs
			end
			if type(role) == "string" then
				_G.webConfig.users[pos].role = role
			end
			_G.CONFIGS.Submit()
			_M.UpdateAuth()
			return true
		end
	end	
end
----------------------------------------------------------
local privOptions = {
	"SYS",
	"LOG",
	"TIME",
	"DVLP",
	"LD_PSET",
	"SAV_PSET",
	"ADJ_AV",
	"OSD",
	"NET",
	"WEB",
	"ONVIF",
	"SIP",
	"TELNET",

	"SER",
	"PTZ",
	"PTZ_CTRL",
	"ENC_SET",
	"REC_SET",
	"STRM_CHO",
	"STRM_USR",
	"STRM_CTRL",
	"STRM_CFG",
}
----------------------------------------------------------
_M.AddRole = function(role)
	local i
	local pos

	if role == "admin" or role == "default" then
		return false, "reserved"
	elseif type(role) ~="string" or role == "" then
		return false, "invalid"
	else
		for i=1,MAX_ROLE_ALLOW do
			if _G.webConfig.roles[i].name == role then
				return false, "exist"
			elseif _G.webConfig.roles[i].name == "" and not pos then
				pos = i
			end
		end
	end

	if not pos then
		return false, "full"
	else
		_G.CONFIGS.BeginUpdate()
		_G.webConfig.roles[pos].name = role
		for i=1,#privOptions do
			_G.webConfig.roles[pos].privileges[privOptions[i]] = _G.webConfig.defaultPrivileges[privOptions[i]]
		end
		_G.CONFIGS.Submit()
		return true
	end
end
----------------------------------------------------------
_M.DelRole = function( role )
	if role == "admin" or role == "default" then
		return false, "reserved"
	elseif type(role) ~="string" or role == "" then
		return false, "invalid"
	else
		for j = 1, MAX_USER_ALLOW do
			if _G.webConfig.users[j].role == role then
				return false, "used"
			end
		end
		local i
		local pos
		for i=1,MAX_ROLE_ALLOW do
			if _G.webConfig.roles[i].name == role then
				pos = i
				break
			end
		end
		if not pos then
			return false, "notexist"
		else
			_G.CONFIGS.BeginUpdate()
			_G.webConfig.roles[pos].name = ""
			for i=1,#privOptions do
				_G.webConfig.roles[pos].privileges[privOptions[i]] = 0
			end
			_G.CONFIGS.Submit()
			return true
		end
	end	
end
----------------------------------------------------------
_M.ChangeRole = function( role, privs, updateOnly )
	if type(role) ~="string" or role == "" or type(privs) ~= "table" then
		return false, "invalid"
	end

	local i
	local obj

	if role == "admin" then
		obj = _G.webConfig.adminPrivileges
	elseif role == "default" then
		obj = _G.webConfig.defaultPrivileges
	else
		for i=1,MAX_ROLE_ALLOW do
			if _G.webConfig.roles[i].name == role then
				obj = _G.webConfig.roles[i].privileges
				break
			end
		end
	end

	if not obj then
		return false, "notexist"
	else
		_G.CONFIGS.BeginUpdate()
		local k,v
		if not updateOnly then
			for k,v in ipairs(privOptions) do
				obj[v] = 0
			end
		end

		for k,v in pairs(privs) do
			obj[k] = v
		end
		_G.CONFIGS.Submit()
		return true
	end
end
----------------------------------------------------------
_M.GetRoles = function()
	local results={}
	local i
	for i=1,MAX_ROLE_ALLOW do
		if _G.webConfig.roles[i].name ~= "" then
			table.insert( results, _G.webConfig.roles[i].name)
		end
	end
	return results
end
----------------------------------------------------------

_M.GetJsonPrivileges = function(role)
	local privObj
	if role == "admin" then
		privObj = _G.webConfig.adminPrivileges
	elseif role == "default" then
		privObj = _G.webConfig.defaultPrivileges
	elseif role ~= "" then
		local i
		for i=1, MAX_ROLE_ALLOW do
			if _G.webConfig.roles[i].name == role then
				privObj = _G.webConfig.roles[i].privileges
				break
			end
		end
	end

	if not privObj then
		return "null"
	else
		local privs = "{"
		local x
		for x=1, #privOptions do
			if x > 1 then
				privs = privs .. ","
			end
			privs = privs .. privOptions[x] .. ":" .. (privObj[privOptions[x]] or 0)
		end
		privs = privs .. "}"
		return privs
	end
end

----------------------------------------------------------
_M.GetRolesJsonDetail = function()
	local results={}
	local i
	for i=1,MAX_ROLE_ALLOW do
		if _G.webConfig.roles[i].name ~= "" then
			local privs = "{"
			local x
			for x=1, #privOptions do
				if x > 1 then
					privs = privs .. ","
				end
				privs = privs .. privOptions[x] .. ":" .. (_G.webConfig.roles[i].privileges[privOptions[x]] or 0)
			end
			privs = privs .. "}"

			table.insert( results, {
				name = _G.webConfig.roles[i].name,
				json_privs = privs
			})
		end
	end
	return results
end
----------------------------------------------------------
_M.DefaultLanguage = function()
	if forceLangs then
		if _G.webConfig and _G.webConfig.defaultLanguage ~= "" then
			return _G.webConfig.defaultLanguage
		else
			return forceLangs.DEFAULT
		end
	end
	
	local defLang
	if _G.webConfig then
		defLang = _G.webConfig.defaultLanguage
	end
	
	return cfgM.CheckDefaultLanguage(defLang)
end
----------------------------------------------------------
_M.CheckValidLanguage = function(lang)
	if forceLangs then
		local i
		for i=1, #forceLangs.VALIDS do
			if forceLangs.VALIDS[i] == lang then
				return lang
			end
		end
		return forceLangs.VALIDS[1]
	end
	
	return cfgM.CheckDefaultLanguage(lang)
end
----------------------------------------------------------

_M.AllLanaguages = function()
	if forceLangs then
		return forceLangs.VALIDS
	end
	
	return cfgM.GetAllLanguages()
end
----------------------------------------------------------
_M.UpdateSystemConfigs = function()
	--Just try reload the settings.
	local r,msg = pcall( dofile, "/usr/share/lua/5.3/sys_setting.lua" )
	if r then
		if type(_G.KV.SERIAL_HISTORIES) == "table" and #_G.KV.SERIAL_HISTORIES > 0 then
			_G.KV.SERIAL_NO = _G.KV.SERIAL_HISTORIES[ #_G.KV.SERIAL_HISTORIES ].serial
		end

		if type(_G.KV.LANGUAGE_CONFIGS) == "table" and #_G.KV.LANGUAGE_CONFIGS > 0 then
			_G.KV.LANGUAGE_SUPPORT = _G.KV.LANGUAGE_CONFIGS[ #_G.KV.LANGUAGE_CONFIGS ]
		end
		return true
	else
		return false
	end
end
----------------------------------------------------------
_M.GetSerialNo = function()
	return _G.KV.SERIAL_NO or "0000000000"
end
----------------------------------------------------------
_M.CheckHardwareInfo = function()
	local done,hwChk = pcall( require, "kiloview." .. _G.DEVICE_TYPE .. ".hardwareCheck" )
	if done and type(hwChk) == "table" then
		hwChk.Check()
		return {
			hardwareVersion = hwChk.hardwareVersion,
			hardwareVersionCode = hwChk.hardwareVersionCode,
			hardwareVersionSwCode = hwChk.hardwareVersionSwCode,
			init = true
		}
	else
		return nil
	end
end
----------------------------------------------------------
_M.VerifyLinuxUser = function(user,pass)
	return SYSTEM.LINUX.VerifyUser( user,pass )
end
----------------------------------------------------------
_M.UpdateNginxConfig = function()
	local onvif_port = 8080
	local onvifProxy = GetCoAFProxy( "/", "onvif", "local", 15 )
	if onvifProxy then
		local r, cc, get = onvifProxy:Get({ ["onvif.control.port"] = 0 })
		if r and get then
			onvif_port = get["onvif.control.port"] or 8080
		end
		onvifProxy:Destroy()
	end

	os.execute( "/bin/sed -e 's/" .. FLAG_HTTP_PORT .. "/" .. _G.webConfig.port .. "/g' " .. 
				"-e 's/" .. FLAG_HTTPS_PORT .. "/" .. _G.webConfig.https_port .. "/g' " ..
				"-e 's/" .. FLAG_ONVIF_PORT .. "/" .. onvif_port .. "/g' " .. 
				NGINX_CONF_LOC .. "/" .. NGINX_CONF_TEMPLATE .. " > " .. NGINX_CONF_LOC .. "/nginx.conf" )
end

----------------------------------------------------------

_M.StartNginx = function()
	local lfs = require "lfs"
	--Install cert
	if not lfs.attributes( NGINX_CONF_LOC .. "/" .. NGINX_CERT_KEY ) then
		os.execute( "/bin/cp -Rf " .. NGINX_CONF_LOC .. "/" .. NGINX_CERT_KEY .. ".default " .. NGINX_CONF_LOC .. "/" .. NGINX_CERT_KEY )
		os.execute( "/bin/sync" )
	else
		local f = io.open( NGINX_CONF_LOC .. "/" .. NGINX_CERT_KEY, "r" )
		local valid
		if f then
			local line = f:read("*l")
			f:close()
			if line and line:match("%-BEGIN") then
				valid = true
			end
		end
		if not valid then
			os.execute( "/bin/cp -Rf " .. NGINX_CONF_LOC .. "/" .. NGINX_CERT_KEY .. ".default " .. NGINX_CONF_LOC .. "/" .. NGINX_CERT_KEY )
			os.execute( "/bin/sync" )
		end
	end
	if not lfs.attributes( NGINX_CONF_LOC .. "/" .. NGINX_CERT_CRT ) then
		os.execute( "/bin/cp -Rf " .. NGINX_CONF_LOC .. "/" .. NGINX_CERT_CRT .. ".default " .. NGINX_CONF_LOC .. "/" .. NGINX_CERT_CRT )
		os.execute( "/bin/sync" )
	else
		local f = io.open( NGINX_CONF_LOC .. "/" .. NGINX_CERT_CRT, "r" )
		local valid
		if f then
			local line = f:read("*l")
			f:close()
			if line and line:match("%-BEGIN") then
				valid = true
			end
		end
		if not valid then
			os.execute( "/bin/cp -Rf " .. NGINX_CONF_LOC .. "/" .. NGINX_CERT_CRT .. ".default " .. NGINX_CONF_LOC .. "/" .. NGINX_CERT_CRT )
			os.execute( "/bin/sync" )
		end
	end
	_M.UpdateAuth()
	os.execute( NGINX ) --Start nginx web server.
end

----------------------------------------------------------
_M.ReloadWebService = function()
	_G.WEB_CONFIG_CHANGED = false
	_M.UpdateNginxConfig()
	os.execute( NGINX .. " -s stop" )
	os.execute( NGINX )
end

----------------------------------------------------------
_M.GetHwAddress = function()
	local f = io.popen("/sbin/ifconfig eth0")
	if f then
		local mac
		local line = f:read("l")
		while line do
			mac = mac or line:match("HWaddr (.+)")
			line = f:read("l")
		end
		f:close()
		return mac or ""
	end
	return ""
end
----------------------------------------------------------

return _M

