local rpc = require "kiloview.ringRpc"
local nu = require "kiloview.netUtil"

local _M = {}

--Get default wifi device.
function _M.GetDefaultWifi(objId)
	local r, devices = rpc.Call( "NETWORK_MANAGER.GetDevicesBy", "*", "WIFI" )
	if not r or not devices or #devices < 1 then
		return nil
	elseif not objId or objId == "" or objId == "/" then
		return devices[1]
	else
		local k,v
		for k,v in pairs(devices) do
			if v.Object == objId then
				return v
			end
		end
		return nil
	end
end

--Get wifi scanlist
function _M.GetWifiScanlist(deviceObj)
	local r, ssidList = rpc.Call( "WIFI.WifiScanList", deviceObj )
	if not r or not ssidList then
		return nil
	else
		return ssidList
	end
end

--Get Wifi connections (saved)
function _M.GetConnections(obj)
	--local objPath = _M.utils.url_encode(obj)

	local r, devInfo = rpc.Call( "NETWORK_MANAGER.GetDeviceInfo", obj, {GENERIC=true} )	--No options, so only get base informations.
	if not r or type(devInfo) ~= "table" then
		return nil
	end

	local activeUuid = ""
	if devInfo.ActiveConnection and devInfo.ActiveConnection ~= "" and devInfo.ActiveConnection ~= "/" then
		local activeConnInfo
		r, activeConnInfo = rpc.Call( "NETWORK_MANAGER.GetActivateConnectionDetails", devInfo.ActiveConnection )
		if r and activeConnInfo then
			activeUuid = activeConnInfo[devInfo.ActiveConnection].Uuid
		end
	end

	local connections = {}

	if type(devInfo.AvailableConnections) == "table" then
		local connDetails
		--r, connDetails = rpc.Call( "NETWORK_MANAGER.GetConnectionDetails", devInfo.AvailableConnections )
		r, connDetails = rpc.Call( "NETWORK_MANAGER.ListConnectionAndGetDetails", { interface=devInfo.Name } )
		if r and connDetails then
			local k,v 
			for k,v in pairs(connDetails) do
				local title
				local secinfo
				local addrResolv
				local isCurrent = false
				
				title = v["connection"]["id"]
				if not title or title == "" then
					title = v["802-11-wireless"]["ssid"] or "UNKNOWN"
				end

				if v["connection"]["uuid"] and v["connection"]["uuid"] == activeUuid then
					isCurrent = true
				end

				if v["802-11-wireless-security"] then
					local kmgmt = v["802-11-wireless-security"]["key-mgmt"]
					if kmgmt == "none" then
						secinfo = "wep"
					elseif kmgmt == "ieee8021x" then
						secinfo = "dynamic-wep"
					elseif kmgmt == "wpa-none" then
						secinfo = "adhoc-psk"
					elseif kmgmt == "wpa-psk" then
						secinfo = "wpa-psk"
					elseif kmgmt == "wpa-eap" then
						secinfo = "wpa-eap"
					else
						secinfo = "unknown"
					end
				else
					secinfo = "open"
				end

				if v["ipv4"] and v["ipv4"]["method"] then
					if v["ipv4"]["method"] == "auto" then
						addrResolv = "dhcp"
					else
						addrResolv = "manual"
					end
				end

				table.insert( connections, {
					title = title,
					secinfo = secinfo,
					addrResolv = addrResolv,
					isCurrent = isCurrent,
					connObj = k,
					uuid = v["connection"]["uuid"] or ""
				})
			end	--For
		end --If get details
	end -- If Avaliable connections got.
	
	return connections
end

function _M.GetConnection(obj, specConn)
	if not obj or not specConn then
		return nil, "Invalid arguments"
	end

	local r, devInfo = rpc.Call( "NETWORK_MANAGER.GetDeviceInfo", obj, {WIFI = true} )	--Get WIFI informations, so I can get MAC.
	if not r or not devInfo then
		return nil, "Fail to get wifi device info"
	end

	local devMac = devInfo.WIFI.HwAddress	--Identify the WIFI device.
	local ssid = ""
	local bssid = ""
	local dhcp = true
	local ip4addr = ""
	local ip4mask = ""
	local ip4gateway = ""
	local ip4dns1 = ""
	local ip4dns2 = ""
	local psk = nil
	local autoconn = true
	local secrets = "wpa-psk"
	local uuid = ""

	local r, connInfo = rpc.Call( "NETWORK_MANAGER.GetConnectionDetails", specConn )
	if not r or not connInfo or not connInfo[specConn] then
		return nil, "Fail to get connection"
	end

	connInfo = connInfo[specConn]

	ssid = connInfo["802-11-wireless"]["ssid"]
	if type(ssid) == "table" then	--Bytes array
		ssid = string.char( table.unpack(ssid) )
	end

	bssid = connInfo["802-11-wireless"]["bssid"]
	if type(bssid) == "table" and #bssid >= 6 then
		bssid = string.format( "%02X:%02X:%02X:%02X:%02X:%02X", bssid[1],bssid[2],bssid[3],bssid[4],bssid[5],bssid[6] )
	else
		bssid = ""
	end

	local kmgt = connInfo["802-11-wireless-security"] and connInfo["802-11-wireless-security"]["key-mgmt"] or "none"
	local auth_alg = connInfo["802-11-wireless-security"] and connInfo["802-11-wireless-security"]["auth-alg"] or ""
	local wep_key_type = connInfo["802-11-wireless-security"] and connInfo["802-11-wireless-security"]["wep-key-type"] or 0

	if kmgt == "wpa-psk" then
		secrets = "wpa-psk"
	elseif kmgt == "wpa-none" then
		secrets = "wpa-none"
	elseif kmgt == "ieee8021x" then
		secrets = "ieee8021x"
	elseif kmgt == "none" then
		if auth_alg == "shared" then
			secrets = "none-share"
		elseif auth_alg == "open" then
			secrets = "none-open"
		else
			secrets = "none-none"
		end

		if wep_key_type == 2 then
			secrets = secrets .. "-128"
		end
	end


	local id = connInfo["connection"] and connInfo["connection"]["id"] or ssid
	if id then
		--[[
		local fname = "/etc/NetworkManager/system-connections/" .. id
		local file = io.open( fname, "r" )
		print( file, fname )
		if file then
			local l = file:read("l")
			while l do
				psk = l:match( "^psk=(.*)$" )
				if psk then
					break
				end
				l = file:read("l")
			end
			file:close()
		end
		]]
		local fname = "/etc/NetworkManager/system-connections/" .. id
		local pp = io.popen( "/bin/cat " .. fname .. " | /bin/grep \"psk=\"", "r" )
		if pp then
			local l = pp:read("l")
			if l then
				psk = l:match( "psk=(.*)" )
			end
			pp:close()
		end
	end

	if not psk then
		psk = connInfo["802-11-wireless-security"] and connInfo["802-11-wireless-security"]["psk"] or "~dont~change"
	end

	uuid = connInfo["connection"]["uuid"]
	if type(connInfo["connection"]["autoconnect"]) ~= "nil" then
		--remotedostring("print(...)", "XXX", type(connInfo["connection"]["autoconnect"]) );
		autoconn = connInfo["connection"]["autoconnect"]
	else
		autoconn = true
	end
	--remotedostring("print(...)", 1, autoconn )

	if autoconn and type(autoconn) ~= "boolean" then
		autoconn = (autoconn ~= 0) and (autoconn ~= "false") and (autoconn ~= "no" )
	end

	--remotedostring("print(...)", 2, autoconn )

	dhcp = connInfo["ipv4"]["method"] and connInfo["ipv4"]["method"] == "auto" or false

	if connInfo["ipv4"]["address-data"] and #connInfo["ipv4"]["address-data"] > 0 then
		ip4addr = connInfo["ipv4"]["address-data"][1]["address"]
		local pfx = connInfo["ipv4"]["address-data"][1]["prefix"]
		ip4mask = nu.ip4PrefixToMask(pfx)
		ip4gateway = connInfo["ipv4"]["gateway"]

		if connInfo["ipv4"]["dns"] and #connInfo["ipv4"]["dns"] > 0 then
			ip4dns1 = connInfo["ipv4"]["dns"][1] or ""
			ip4dns2 = connInfo["ipv4"]["dns"][2] or ""
		end
	end
	
	return {
		DeviceObject = obj,
		Object = specConn,
		MAC = devMac,
		SSID = ssid,
		BSSID = bssid,
		DHCP = dhcp and 1 or 0,
		IP = ip4addr,
		Mask = ip4mask,
		Prefix = pfx,
		Gateway = ip4gateway,
		Dns1 = ip4dns1,
		Dns2 = ip4dns2,
		Psk = psk,
		AutoConnect = autoconn and 1 or 0,
		Security = secrets,
		Uuid = uuid
	}
end

--Set/new Wifi connection
function _M.SetConnection( deviceObj, new, args )
	local r, devInfo = rpc.Call( "NETWORK_MANAGER.GetDeviceInfo", deviceObj, {WIFI = true} )	--Get WIFI informations, so I can get MAC.
	if not r or not devInfo then
		return false, "ERROR: Fail to get WIFI device informations"
	end

	local devName = devInfo.Name

	local connSettings = {
		["connection"] = {},
		["802-11-wireless"] = {},
		["ipv4"] = {}
	}

	connSettings["connection"]["type"] = "802-11-wireless"
	connSettings["connection"]["autoconnect"] = (args.AutoConnect == 1 or args.AutoConnect == '1')

	if new then	--New a connection.
		if not args.SSID or args.SSID == "" then
			return false, "Invalid SSID"
		end

		connSettings["connection"]["id"] = args.Title or args.SSID
		connSettings["802-11-wireless"]["ssid"] = args.SSID
	else --Change a connection
		if args.Uuid and args.Uuid ~= "" then
			connSettings["connection"]["uuid"] = args.Uuid
			local r, detail = rpc.Call( "NETWORK_MANAGER.ListConnectionAndGetDetails", {uuid=args.Uuid} )
			if r and detail then
				local k,v
				local found
				for k,v in pairs(detail) do
					found = v
					break
				end
				
				if found then
					connSettings["connection"]["id"] = found["connection"]["id"]
					connSettings["802-11-wireless"]["ssid"] = found["802-11-wireless"]["ssid"]
				end
			end
		end
	end

	connSettings["connection"]["interface-name"] = devName

	local psk = args.Psk
	if psk == "~dont~change" then
		psk = nil
	end

	if args.Security then
		if args.Security == "none-none" then
			connSettings["802-11-wireless-security"] = nil	--No security setting
		elseif args.Security == "none-open" then
			connSettings["802-11-wireless-security"] = {
				["key-mgmt"] = "none",
				["auth-alg"] = "open",
				["wep-key-type"] = 1,
				["wep-key0"] = psk
			}
		elseif args.Security == "none-share" then
			connSettings["802-11-wireless-security"] = {
				["key-mgmt"] = "none",
				["auth-alg"] = "shared",
				["wep-key-type"] = 1,
				["wep-key0"] = psk
			}
		elseif args.Security == "none-open-128" then
			connSettings["802-11-wireless-security"] = {
				["key-mgmt"] = "none",
				["auth-alg"] = "open",
				["wep-key-type"] = 2,
				["wep-key0"] = psk
			}
		elseif args.Security == "none-share-128" then
			connSettings["802-11-wireless-security"] = {
				["key-mgmt"] = "none",
				["auth-alg"] = "shared",
				["wep-key-type"] = 2,
				["wep-key0"] = psk
			}
		--[[
		elseif args.Security == "ieee8021x" then
			connSettings["802-11-wireless-security"] = {
				["key-mgmt"] = "ieee8021x",
				["auth-alg"] = "shared",
				["wep-key-type"] = 1,
				["wep-key0"] = psk
			}
		]]
		elseif args.Security == "wpa-none" then
			connSettings["802-11-wireless-security"] = {
				["key-mgmt"] = "wpa-none",
				["psk"] = psk,
				["pairwise"] = { "none" }
			}
		else
			connSettings["802-11-wireless-security"] = {
				["key-mgmt"] = "wpa-psk",
				["psk"] = psk,
				["pairwise"] = { "ccmp", "tkip" }
			}
		end
	end

	if args.DHCP and tonumber(args.DHCP) == 1 then	--DHCP
		connSettings["ipv4"]["method"] = "auto"
	else
		connSettings["ipv4"]["method"] = "manual"
		if args.IP and args.IP ~= "" then
			connSettings["ipv4"]["address-data"] = {{address=args.IP, prefix=tonumber(args.Prefix) or nu.ip4MaskToPrefix(args.Mask or "255.255.255.0")}}
		end

		connSettings["ipv4"]["gateway"] = args.Gateway or ""
		if args.Dns1 and args.Dns1 == "" then
			args.Dns1 = nil
		end

		if args.Dns2 and args.Dns2 == "" then
			args.Dns2 = nil
		end

		connSettings["ipv4"]["dns"] = { args.Dns1, args.Dns2 }
		if #connSettings["ipv4"]["dns"] < 1 then
			connSettings["ipv4"]["dns"] = nil
		end
	end


	if new then
		local r, o = rpc.Call( "NETWORK_MANAGER.AddConnection", connSettings )
		if r and o then
			return true
		else
			return false, "ERROR: Add connection fail"
		end
	else
		local r, o = rpc.Call( "NETWORK_MANAGER.ChangeConnectionDetails", args.Object, connSettings )
		if r and o then
			return true
		else
			return false, "ERROR: Change connection fail"
		end
	end
end

--Forget/Delete a connection
function _M.DeleteConnection( conPath )
	if not conPath then
		return false, "ERROR: Invalid connection object"
	end

	local r, code = rpc.Call( "NETWORK_MANAGER.DeleteConnection", conPath )
	if r and code then
		return true
	else
		return false, "ERROR: Fault to delete wifi connection"
	end
end

--Connect a connection
function _M.ConnectConnection(obj, conPath )
	if not obj then
		return false, "ERROR: Invalid Wifi device"
	end

	if not conPath then
		return false, "ERROR: Invalid connection object"
	end

	local r, code = rpc.Call( "NETWORK_MANAGER.ConnectDevice", obj, conPath )
	if r and code then
		return true
	else
		return false, "ERROR: Connect fail"
	end
end

--Disconnect WIFI
function _M.DisconnectConnection( obj )
	if not obj then
		return false, "ERROR: Invalid Wifi object"
	end

	local r, code = rpc.Call( "NETWORK_MANAGER.DeactivateDevice", obj )
	if r and code then
		return true
	else
		return false, "ERROR: Disconnect fail"
	end
end

--Get wifi detail
function _M.GetWifiDetails( device )
	if not device then
		return nil
	end

	local rslt = {}
	
	local r,devInfo = rpc.Call( "NETWORK_MANAGER.GetDeviceInfo", device.Object, { WIFI = true, IP4 = true } )
	if devInfo and devInfo.WIFI and devInfo.WIFI.WirelessCapabilities then
		rslt.Capabilities = devInfo.WIFI.WirelessCapabilities
	else
		rslt.Capabilities = ""
	end

	if devInfo and devInfo.WIFI and devInfo.WIFI.ActiveAccessPoint then
		rslt.ConnectedSSID = devInfo.WIFI.ActiveAccessPoint.Ssid
		rslt.ConnectingSpeed = devInfo.WIFI.ActiveAccessPoint.MaxBitrate/1000
		rslt.SignalStrength = devInfo.WIFI.ActiveAccessPoint.Strength
							 
		if devInfo.IP4 and devInfo.IP4.Addresses and #devInfo.IP4.Addresses > 0 then
			rslt.IP = devInfo.IP4.Addresses[1].address
			rslt.Prefix = devInfo.IP4.Addresses[1].prefix
			rslt.Gateway = devInfo.IP4.Gateway
		else
			rslt.IP = ""
			rslt.Prefix = 0
			rslt.Gateway = ""
		end
	else
		rslt.ConnectedSSID = ""
		rslt.ConnectingSpeed = 0
		rslt.SignalStrength = 0
		rslt.IP = ""
		rslt.Prefix = 0
		rslt.Gateway = ""
	end
	
	return rslt
end

--------------------------------------------------------

return _M


