#!/usr/bin/env lua
local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"
local service = require "coaf.service"

local socket = require "cosocket.core"
local roll = require "coaf.log.rollfile"

local event = require "coaf.event"

-- Change current dir to app location
coaf.ChDir(coaf.GetAppDir(arg))

------------------------------------------------
local cfg = require("nas_config").Load()

local withLog = false
local _x=1
while _x <= #arg do
	if arg[_x] == "--log" then
		withLog = true
	end
	_x = _x+1
end

-- First, initialize the log file.
if withLog then
	os.execute( "/bin/mkdir -p /var/log" )
	local logF = roll.Create("/var/log/nas_manager.log")
	if not logF or not logF:init() then
		_dbg.Error("NAS", "Fail to open log file: /var/log/nas_manager.log, so the log messages should be output to console.")
	else
		_dbg.AddLogger("logfile", logF, nil, nil)
	end
end


------------------------------------------------
_G.g_service = nil
------------------------------------------------

local nas = {
	mounts = {}
}

local function mountThread(thrObj, id, ctx, item)
	--TODO: Check NAS type
	local nas_mounter = require "nas_NFS"
	local mnt_type = "nfs"

	if item.type and item.type ~= "" and item.type ~= "nfs" then
		if item.type == "cifs" then
			nas_mounter = require "nas_CIFS"
			mnt_type = "cifs"
		end
	end

	_dbg.Info( "NAS", "Start mounting NAS/" .. mnt_type .. " item of ", id , "..." )
	while not thrObj.terminated do
		local mnt,name = nas_mounter.Mount( _G.g_event, id, item )
		if mnt then
			ctx.status = "mounted"
			ctx.mount_thread = nil
			ctx.name = name
			ctx.mount = mnt
			break
		else
			_dbg.Error( "NAS", "Mount NAS/" .. mnt_type .. " item of ", id , " fail! with message:", name )
		end
		coaf.Sleep(3)	--Sleep 3 seconds and try again.
	end

	if ctx.status == "mounted" then
		_dbg.Info( "NAS", "NAS/" .. mnt_type .. " item of ", id , " is mounted at ", ctx.mount, ", name is:", ctx.name )
		os.execute( "/usr/bin/disk-attach.lua " .. ctx.mount .. " nas " .. mnt_type .. " " .. ctx.name )
	else
		_dbg.Info( "NAS", "NAS/" .. mnt_type .. " item of ", id , " is not mounted!" )
	end
end

function nas.StartMount(id)
	local cfgItem = cfg[id]
	if not cfgItem then
		return false, "Invalid NAS item id"
	else
		local ctx = { status = "mounting" }
		ctx.mount_thread = thread.New( mountThread )
		ctx.type = cfgItem.type
		nas.mounts[id] = ctx
		ctx.mount_thread:Start( id, ctx, cfgItem )
		return true
	end
end

function nas.GetList()
	local rslt = {}
	local k,v 
	for k,v in pairs(cfg) do
		local item = { id = k, name = v.name, type=v.type, host = v.host, volumn = v.volumn, options = v.options }
		local ctx = nas.mounts[k]
		if ctx then
			item.status = ctx.status
			item.name = ctx.name or item.name
			item.mount = ctx.mount
		else
			item.status = "unmount"
		end
		table.insert(rslt,item)
	end
	table.sort(rslt, function(a,b)
		return a.id < b.id
	end)
	return rslt
end

function nas.Add( id, name, type, host, volumn, options )
	if not id or not type or not host or not volumn then
		return false, "Invalid NAS arguments"
	end

	if cfg[id] then
		return false, "The NAS item of " .. (id or "?") .." is already exist!"
	end

	cfg[id] = {
		name = name ~= "" and name or id,
		type = type ~= "" and type or "nfs",
		host = host,
		volumn = volumn,
		options = options
	}

	cfg:__TRIGGER(true)
	nas.StartMount(id)
	return true
end

function nas.Remove( id )
	local ctx = nas.mounts[id]
	if ctx then
		if ctx.mount_thread then
			ctx.mount_thread:Terminate()
			ctx.mount_thread = nil
		end

		if ctx.status == "mounted" then
			os.execute( "/usr/bin/disk-dettach.lua " .. ctx.mount )
			local mounter = require "nas_NFS"
			if ctx.type and ctx.type ~= "" and ctx.type ~= "nfs" then
				if ctx.type == "cifs" then
					mounter = require "nas_CIFS"
				end
			end
			mounter.UMount(_G.g_event, id)
		end
		nas.mounts[id] = nil
	end

	cfg[id] = nil
	cfg:__TRIGGER(true)

	return true
end

function nas.Init()
	local k,v
	for k,v in pairs(cfg) do
		nas.StartMount(k)
	end
end

------------------------------------------------

coaf.Init( "nas_manager" )
-- Initialize service.
service.InitSocket()

---------------------------------------------------------

local function Main(args)
	_G.g_event = event.createNew( "nas_manager", true ) -- Don't start slave thread, since we never use it.
	g_service = service.New( "nas_manager", "unix:/var/run/nas.sock" )
	if g_service then
		g_service:CreateCoServerProxy()
		g_service:ExposeObject("/", nas )
	else
		_dbg.Error( "NAS", "Fail to create CoService for nas_manager" )
	end
	nas.Init()
end

coaf.Start(Main,arg)

_dbg.Print( "NAS", "nas_manager service is terminated." )

