local _dbg = require "coaf.debug"
local protocol = require "coaf.rpc.protocol"
local compa = require "coaf.compatible"

--Try require 'coaf.core', but not necessary.
local _flag, coaf = pcall(require, "coaf.core")
if not _flag then
	coaf = nil
end

local _M = {
	timeoutUnit = 1	--NOTE: Since the openresty socket socket timeout is based ms, if that,
					--      you must set it to be 1000.
}

local adapter_meta = {}

function adapter_meta:Close()
	if self.connection then
		self.connection:close()
		self.connection = nil
	end
end

function adapter_meta:Send(content)
	if not content or not self.connection then
		return false
	end
	local sn = self.SN_seed
	if sn >= 99999999 then
		self.SN_seed = 0
	else
		self.SN_seed = sn + 1
	end

	local pkg
	if type(content) == "table" then
		content.SN = sn
		if not content.headers then
			content.headers = { VIA = self.via_src }
		else
			content.headers.VIA = self.via_src
		end
		pkg = table.concat( protocol.FormatContent(content) )
	elseif type(content) == "string" then
		pkg = content
	end
		
	while pkg and true do
		local status, err = self.connection:send( pkg )
		if not status then
			if err ~= "busy" then
				return status, err
			elseif coaf then
				_dbg.Info("CoAF.Client", "Simple adapter - sending is busy, try again.")
				coaf.Sleep(0.01)
			end
		else
			return status
		end
	end
end

function adapter_meta:RecvACK(matchSN,timeout)
	if not self.connection then
		return nil, "invalid"
	end

	local parser = protocol.NewParser()

	while true do
		if timeout then
			self.connection:settimeout(timeout*_M.timeoutUnit)
		end
		
		local data, err, partial = self.connection:receive(parser.next_fetch)
		
		if timeout then
			self.connection:settimeout(self.timeout*_M.timeoutUnit)
		end

		if not data then
			if err ~= "timeout" and err ~= "busy" then
				_dbg.Info("CoAF.Client", "Simple adapter - recv error:", err)
				return nil, err
			end
			if partial and #partial > 0 then
				data = partial
			elseif err == "timeout" then
				return nil, err
			else --NOTE: here! Very important: (BUG found)
			-- In coaf environment, 'busy' means the socket is using by others.
			-- In this case, if this task (coroutine, or main task) does not yield
			-- out, it will fall in a dead loop: Other task will never be waked up
			-- and this loop will always continue ('busy' is always get).
				if coaf then
					coaf.Sleep(0.01)
				end
			end
		end

		if data then
			local done, flag = parser:Push(data)
			if done and flag == "complete" then
				local content = parser.content
				parser:Clear() -- To accept new contents.

				if not content.SN or not content.CMD then
					_dbg.Warn("CoAF.Client", "Simple adapter - Invalid message got (without SN or CMD).")
					if matchSN == "dryrun" then
						return true, "bad"
					end
				elseif not content.CMD:match("^ACK/") then
					if content.CMD == "NOTIFY" then --Process NOTIFY
						if self.proxy then
							self.proxy:__ProcessNotifier(content)
						end
						if matchSN == "dryrun" then
							return true, "notify"
						end
					else
						local ack = {
							SN = content.SN,
							CMD = "ACK/" .. content.CMD,
							ARGS = content.ARGS,
							headers = {
								RESULT = "ERROR",
								MSG = "Invalid request",
								SVC = content.headers["SVC"],
								VIA = content.headers["VIA"],
								LOC = content.headers["LOC"],
								OBJ = content.headers["OBJ"]
							}
						}
						self:Send( table.concat(protocol.FormatContent(ack)) )
						if matchSN == "dryrun" then
							return true, "request"
						end
					end
				elseif matchSN == "dryrun" then
					return true, "ack"
				elseif not matchSN or tonumber(content.SN) == matchSN then
					return content
				end 
			end -- End of if completed.
		end --End of get line.
	end --loop.
	return nil, "invalid"
end

function adapter_meta:ForkConnection()
	local newConn = self.socket.tcp()
	newConn:settimeout(self.timeout*_M.timeoutUnit)
	local status, err = newConn:connect(self.address, self.port)
	if not status then
		_dbg.Error("CoAF.Client", "Simple adapter - Fail to connect target address while forking connection")
		return nil
	end
	local new_adapter = {
		socket = self.socket,
		address = self.address,
		port = self.port,
		connection = newConn,
		timeout = self.timeout,
		--via_src = ...
		SN_seed = 0,
	}
	local token = tostring(new_adapter)
	token = token:match(".*:%s*(.*)$")
	new_adapter.via_src = token
	setmetatable(new_adapter, adapter_meta)
	return new_adapter
end

function adapter_meta:ConnectService(serviceId, loc, authKey, authUniq)
	local content = {
		CMD = "CONNECT",
		headers = {
			SVC = serviceId,
			LOC = loc,
			UNIQ = authUniq
		}
	}
	if not self:Send(content) then
		_dbg.Error("CoAF.Client", "Simple adapter - Fail to send CONNECT message")
		return false
	end

	local ack = self:RecvACK(content.SN)
	if not ack then
		_dbg.Error("CoAF.Client", "Simple adapter - Fail to recv ACK/CONNECT")
		return false
	end

	if not ack.headers or not ack.headers.RESULT then
		_dbg.Error("CoAF.Client", "Simple adapter - Invalid CONNECT response")
		return false
	end

	if not ngx or not ngx.socket then -- Check in openresty environment?
	--TODO: if not in openresty, then try authorization.
		if authKey and ack.headers.RESULT == "AUTH_REQ" and ack.headers.CHA then
			require "cocrypto"

			--_dbg.Info("CoAF.Client", "[AUTH] Simple adapter - Remote require authorization, challenge is:", ack.headers.CHA)

			local tip = CoAF_CRYPTO.get_challenge()
			local dig = CoAF_CRYPTO.finger_print( ack.headers.CHA .. tip .. authKey )

			content = {
				CMD = "CONNECT",
				headers = {
					SVC = serviceId,
					LOC = loc,
					UNIQ = authUniq,
					CHA = ack.headers.CHA,
					TIP = tip,
					DIG = dig
				}
			}
			if not self:Send(content) then
				_dbg.Error("CoAF.Client", "[AUTH] Simple adapter - Fail to send CONNECT message")
				return false
			end
		
			ack = self:RecvACK(content.SN)
			if not ack then
				_dbg.Error("CoAF.Client", "[AUTH] Simple adapter - Fail to recv ACK/CONNECT")
				return false
			end
		
			if not ack.headers or not ack.headers.RESULT then
				_dbg.Error("CoAF.Client", "[AUTH] Simple adapter - Invalid CONNECT response")
				return false
			end
		end
	end

	if ack.headers.RESULT ~= "OK" and ack.headers.RESULT ~= "ALT-ADDR" then
		_dbg.Error("CoAF.Client", "Simple adapter - CONNECT service fail:", ack.headers.RESULT, ack.headers.MSG)
		return false
	end

	if ack.headers.ALT_ADDR and ack.headers.ALT_ADDR ~= "" and type(ack.data) == "table" then
		table.insert( ack.data, 1, "return ")
		local func, flag = load(table.concat( ack.data ))
		if not func then
			_dbg.Warn("CoAF.Client", "Simple adapter - Connect service done but got invalid alternative address list")
			if ack.RESULT == "ALT-ADDR" then
				return false
			else
				return true
			end
		end
		local alt_list
		flag, alt_list = pcall(func)
		if not flag or type(alt_list) ~= "table" then
			_dbg.Warn("CoAF.Client", "Simple adapter - Connect service done but fail to parse alternative address list")
			if ack.RESULT == "ALT-ADDR" then
				return false
			else
				return true
			end
		end

		local i, v, err
		local addr, port
		for i,v in pairs(alt_list) do

			local newConn = self.socket.tcp()
			if v:match("^unix:") then
				addr = v
				port = 0 --NOTE: don't be nil because openresty does not accept it.
			else
				addr, port = v:match("(.*):([0-9]+)")
				if not addr then
					addr = v
					port = 14000
				else
					port = tonumber(port) or 14000
				end
			end

			flag, err = newConn:connect(addr, port)
			if flag then
				_dbg.Info("CoAF.Client", "Simple adapter - Connected to alternative address:", addr, port)
				self.connection:close()
				self.connection = newConn
				self.address = addr
				self.port = port
				return true
			else
				newConn:close()
				newConn = nil
			end
		end

		if ack.RESULT == "ALT-ADDR" then
			_dbg.Error("CoAF.Client", "Simple adapter - Fail to connect alternative addresses")
			return false
		else
			_dbg.Warn("CoAF.Client", "Simple adapter - Fail to connect alternative addresses, use current (proxied) connection")
			return true
		end
	elseif ack.RESULT == "ALT-ADDR" then
		_dbg.Error("CoAF.Client", "Simple adapter - No valid alternative addresses but coserver requires connect to alternative address")
		return false
	else
		return true
	end
end

function adapter_meta:BindProxy(proxy)
	self.proxy = proxy
end

adapter_meta.__index = adapter_meta

function _M.New(socket, addr, defaultTimeout)
	local conn = socket.tcp()
	local orgAddr = addr
	local port
	if not addr:match("^unix:") then
		local ip
		ip, port = addr:match("(.*):([0-9]+)")
		if ip then
			if ip == '' or ip == '*' then
				addr = "127.0.0.1"
			else
				addr = ip
			end
			port = tonumber(port) or 14000
		else
			port = 14000
		end
		port = tonumber(port) or 14000
	else
		port = 0 --NOTE: openresty must accept a number port.
	end 

	defaultTimeout = defaultTimeout or 30
	conn:settimeout(defaultTimeout*_M.timeoutUnit)

	local status, err = conn:connect(addr,port)
	if not status then
		conn:close()
		_dbg.Error("CoAF.Client", "Simple adapter - Fail to connect:", orgAddr, err)
		return nil, err
	end

	local adapter = {
		socket = socket,
		address = addr,
		port = port,
		connection = conn,
		SN_seed = 0,
		timeout = defaultTimeout,
	}

	local token = tostring(adapter)
	token = token:match(".*:%s*(.*)$")
	adapter.via_src = token
	setmetatable(adapter, adapter_meta)
	return adapter
end


return _M
