--[==[

The Co-Application Framework (CoAF) RPC service protocol.

Copyright (C) 2018, KILOVIEW Electronics, Co., Ltd.

]==]

local _dbg = require "coaf.debug"
local pretty = require "pl.pretty"

--------------------------------------------------------------------------
--[==[
CoRPC protocol is declared as:

1. Basic concepts

CoRPC protocol is a text based protocol. It has 3 parts of a whole protocol
content: A) Instruction; B) Headers; C) Content.

SPECIAL NOTE:
In anywhere, if a line is '--#KEEP-ALIVE#--', you *MUST* drop it since it's
a keep-alive message and no used.

A) Instruction format is:

@<SN> COMMAND [arg1 arg2 ...]\r\n

<SN> is a serial number, to indicate request sequence.
COMMAND is declared as:

* CONNECT	:	Connect to a service, followed connecting service id;
* PING		:	PING target, without command arguments;
* CALL		:	Call a remote function, followed call method name;
* STATIC_CALL:	same as call, but it's obj.Call(...) not obj:Call(...)
* GET		:	Get property, without command arguments;
* SET		:	Set property value, without command arguments;
* SUBS		:	Subscribe topics, without command arguments;
* UNSUBS	:	Unsubscribe topics, without command arguments;
* NOTIFY	:	Notify subscribe message, followed topic name;
* STREAM	:	Data streaming.

if COMMAND is started with 'ACK/' (e.g. ACK/CONNECT, ACK/CALL), that's a
ACK command. The ACK command SN *MUST* be same as the request command SN.
Except for STREAM: ACK SN is ignored for data transfer.

B) Headers:

HeaderName: value\r\n
HeaderName2: value2\r\n

Some important headers:

	SVC: service-id
	LOC: service-location
	VIA: proxy1 proxy2 ...
	OBJ: /path/to/object
	LEN: Content-length
	RESULT: OK|Error
	MSG: Error messages

C) Content:

After last line of headers, *MUST* put a empty line, then
followed the content.

Content is a Lua style table, represent all call arguments for
current instruction. It *MUST* be a single table, or empty should
be '{}'.

The last line of content *MUST* be "--#EOF#--".

E.g.:

@103 CALL functionName\r\n
SVC: 0:codec.video\r\n
OBJ: /\r\n
LEN: 42\r\n
\r\n
{
	"String for arg1",
	123,
	{
		field1 = "Value1",
		field2 = 2
	}
}\r\n
--#EOF#--\r\n

For acking:
@103 ACK/CALL functionName\r\n
SVC: 0:codec.video\r\n
OBJ: /\r\n
LEN: 22\r\n
\r\n
{"OK"}\r\n
--#EOF#--\r\n

]==]
--------------------------------------------------------------------------

--------------------------------------------------------------------------
local _M = {
	LF = "\r\n"
}
--------------------------------------------------------------------------

local parser_meta = {}

function parser_meta:Clear()
	self.content = {}
	self.flag = "init"
	self.next_fetch = "*l"
	self.content_len = 0
	self.orig_content = {}
end

function parser_meta:EnableOriginal(en)
	self.enable_orig = en
end

function parser_meta:Push(data)
	if data == "--#KEEP-ALIVE#--" then
		return true, "accept"
	end

	if self.flag == "init" then
		if data:sub(1,1) == '@' then  -- command/submit instruction
			--data = data .. " "
			local sn,cmd,args = data:match("@(.-)%s+(.-)%s+(.*)$")
			if not sn then
				sn, cmd = data:match("@(.-)%s+(.*)$")
			end

			if sn then
				self.content.SN = sn
				self.content.CMD = cmd
				self.content.ARGS = args
				self.flag = "header"
				
				if self.enable_orig then
					table.insert( self.orig_content, data )
					table.insert( self.orig_content, _M.LF )
				end

				return true, "accept"
			else
				_dbg.Warn("CoAF.Protocol", "Invalid instruction of @")
				return false, "invalid"
			end
		else
			return false, "invalid"
		end
	elseif self.flag == "header" then
		if data == '' then
			if self.enable_orig then
				table.insert( self.orig_content, _M.LF )
			end

			self.flag = "content"
			if self.content_len > 0 then
				self.next_fetch = self.content_len
			else
				self.next_fetch = "*l" --I don't know, so must check lines.
			end
		else
			if self.enable_orig then
				table.insert( self.orig_content, data )
				table.insert( self.orig_content, _M.LF )
			end

			local opt, val = data:match("(.-):%s*(.*)")
			if not opt then
				_dbg.Detail("CoAF.Protocol", "Bad header line:", data)
				return true, "invalid"
			else
				if not self.content.headers then
					self.content.headers = {}
				end
				self.content.headers[opt] = val
				if opt == "LEN" then
					self.content_len = tonumber(val)
				end 
				return true, "accept"
			end
		end
	elseif self.flag == "content" then
		local len = #data
		if not self.content.data then
			self.content.data = { data }
		else
			table.insert( self.content.data, data )
		end

		if self.enable_orig then table.insert( self.orig_content, data ) end

		if self.content_len > 0 then
			self.next_fetch = self.next_fetch - len
			if self.next_fetch <= 0 then
				self.flag = "complete"
				return true, "complete"
			else
				return true, "accept"
			end
		else -- Since receive("*l") drops LF. 
			if self.enable_orig then table.insert( self.orig_content, _M.LF ) end
			table.insert( self.content.data, _M.LF)
			if data == "--#EOF#--" then
				self.flag = "complete"
				return true, "complete"
			else
				return true, "accept"
			end
		end
	else
		return false, "outrange"
	end
end

parser_meta.__index = parser_meta

-----------------------------------------------

function _M.NewParser()
	return setmetatable({
		content = {},
		orig_content = {},
		flag = 'init',
		next_fetch = "*l",
		content_len = 0
	}, parser_meta)
end

-----------------------------------------------

function _M.FormatContent(content)
	local LF = _M.LF
	if not content or not content.SN or not content.CMD then
		return nil
	end

	local rslt = {}

	table.insert(rslt, "@")
	table.insert(rslt,content.SN)
	table.insert(rslt,' ')
	table.insert(rslt,content.CMD)
	if content.ARGS and #content.ARGS > 0 then
		table.insert(rslt, ' ')
		table.insert(rslt, content.ARGS)
	end
	table.insert(rslt,LF)
	
	if type(content.headers) == "table" then
		local k,v
		for k,v in pairs(content.headers) do
			if k:lower() ~= "len" then  -- Ignore 'LEN' header, because of I need calculate it again.
				table.insert(rslt,k)
				table.insert(rslt,': ')
				table.insert(rslt,v)
				table.insert(rslt,LF)
			end
		end
	end

	local eof_flag = "--#EOF#--" .. LF
	local cont_len = #eof_flag

	if type(content.parameters) == "table" then
		local argData = pretty.write(content.parameters,'\t')
		cont_len = cont_len + #argData + #LF
		table.insert(rslt, "LEN: ")
		table.insert(rslt, cont_len)
		table.insert(rslt, LF)
		table.insert(rslt, LF)
		table.insert(rslt, argData)
		table.insert(rslt, LF)
		table.insert(rslt, eof_flag)
	elseif not content.data then
		cont_len = cont_len + 2 + #LF
		table.insert(rslt, "LEN: ")
		table.insert(rslt, cont_len)
		table.insert(rslt, LF)
		table.insert(rslt, LF)
		table.insert(rslt, "{}")
		table.insert(rslt, LF)
		table.insert(rslt, eof_flag)
	else
		local lenIdx
		table.insert(rslt, "LEN: ")
		table.insert(rslt, 0) -- stub here.
		lenIdx = #rslt
		table.insert(rslt, LF)
		table.insert(rslt, LF)

		local no_eof
		local data_size
		if type(content.data) == "string" then
			data_size = #content.data
			if data_size == 0 then
				table.insert(rslt, '{}')
				data_size = 2
			else
				table.insert(rslt, content.data)
				cont_len = 0
				no_eof = true
			end
		elseif type(content.data) == "table" then
			local i, v
			data_size = 0
			for i,v in ipairs(content.data) do
				table.insert(rslt, v)
				data_size = data_size + #v
			end
			if data_size == 0 then
				table.insert(rslt,"{}")
				data_size = 2
			else
				no_eof = true
				cont_len = 0
			end
		else
			table.insert(rslt,"{}")
			data_ize = 2
		end

		if not no_eof then
			cont_len = cont_len + data_size + #LF
			rslt[lenIdx] = cont_len
			table.insert(rslt, LF)
			table.insert(rslt, eof_flag)
		else
			cont_len = cont_len + data_size
			rslt[lenIdx] = cont_len
		end
	end

	return rslt --So you can use table.concat() to output string.
end

------------------------------------------------

return _M
