-- This disk manager and recorder module.
local DBUS_SERVICE = nil
local _DECL = require "kiloview.dbus.sunshine_record"
local lfs = require "lfs"
local pretty = require "pl.pretty"
pcall(require, "features")

-------------------------

local _M = {
	Identifier = "RECORD_MAN",
	FriendlyName = "Disk and record Manager",
	Version = "2.0",
	ConfigPrefix = "record_m",

	ALL_DISKS = {},	--Record all disks and them's mounts.

	RECORDING = {
		[0] = {},
		[1] = {}
	},

	SEQUENCE_COUNTER = 0,
}

local StreamsMap = {
	[0] = "main",
	[1] = "sub"
}

-------------------------

local getset = require "kiloview.getSetHelper"

----------------------------------------------------
-- General module interface
----------------------------------------------------

function _M:Activate()
	--Nothing
end

function _M:Start()
end

function _M:Stop()
end

function _M:Reset()
--TODO: Nothing
end

function _M:Terminate()
	os.exit(0)
end

function _M:Control(action,params)
	print('Control()', action,params)
	if action == "activate" then
		self:Activate()
		return 0, {}
	elseif action == "start" then
		self:Start()
		return 0, {}
	elseif action == "stop" then
		self:Stop()
		return 0, {}
	elseif action == "Reset" then
		self:Reset()
		return 0, {}
	elseif action == "terminate" then
		self:Terminate()
		return 0, {}
	else
		return -1, {}
	end
end

----------------------------------------------------
-- General configure interface
----------------------------------------------------

function _M:Set(params)
	print('Set()', params)
	local cnt = getset.SetConfigureValues( self.ConfigPrefix, params )
	return cnt
end

function _M:SetEnv(params)
	print('SetEnv()', params)
	local cnt = getset.SetValues( "_ENV", params )
	return cnt
end

function _M:Get(inParams)
	print('Get()', inParams)
	local cnt = getset.GetValues( self.ConfigPrefix, inParams )
	return cnt,inParams
end

function _M:GetEnv(inParams)
	print('GetEnv()', inParams)
	local cnt = getset.GetValues( "_ENV", inParams )
	return cnt,inParams
end

function _M:RestoreDefault()
	--TODO:
	return 0
end

function _M:LoadConfigureFile(filename)
	print('LoadConfigureFile()', filename)
	CONFIGS.BeginUpdate()
	local r = pcall( dofile, filename )
	CONFIGS.Submit()
	if r then
		return 1
	else
		return 0
	end
end

----------------------------------------------------
-- Disk interface
----------------------------------------------------

function _M:DiskAttach(device, type, filesys, volName)
	print('DiskAttach()', device,type,filesys,volName)
	if not device or device == "" or not type or type == "" or not filesys or filesys == "" then
		print( "ERROR: Disk attach: invalid arguments" )
		return
	end

	print( "A disk is attached: DEVICE='" .. device .. "', TYPE='" .. type .. "', FILE_SYS='" .. filesys .. "', VOLUMN_NAME='" .. (volName or "") .. "'" )

	if self.ALL_DISKS[device] then
		print( "Current device is already attached, so remount it. (A manually dettach should occus)" )
		self:DiskDettach(device)
	end

	print( "Mounting device..." )
	local r, msg = self:MountDisk( device, type, filesys )
	if not r then
		print( "Mount ERROR: ", msg )
		return
	else
		print( "Mount DONE.")
	end

	--Attached, so initialize the recording database.
	self.InitRecordDB( device )
	self.ALL_DISKS[device].VOLNAME = volName
	local done, total, used, free, percent = self:GetDiskSpaces( device, self.ALL_DISKS[device].TYPE == "nas" )
	if done then
		print( "Disk '" .. device .. "' info: TOTAL=" .. total .. " MB, USED=" .. used .. " MB, FREE=" .. free .. " MB, Percent=" .. percent .. "%" )
		self.ALL_DISKS[device].TOTAL = total
		self.ALL_DISKS[device].USED = used
		self.ALL_DISKS[device].FREE = free
		self.ALL_DISKS[device].PERCENT = percent
	else
		self.ALL_DISKS[device].TOTAL = 0
		self.ALL_DISKS[device].USED = 0
		self.ALL_DISKS[device].FREE = 0
		self.ALL_DISKS[device].PERCENT = 100
	end

	--Recording is disabled.
	if _G.DISABLE_RECORD then
		return
	end

	local k, v
	for k,v in pairs(StreamsMap) do
		if record_m.stream[k].auto_record ~= 0 then	--Auto start recording, so try start it
			record_m.stream[k].start = 1
		end
	end
end

function _M:DiskDettach(device)
	print('DiskDettach()', device)
	print( "A disk is dettached: DEVICE='" .. device .. "'" )
	local k,v
	for k,v in pairs( self.RECORDING) do
		if (v.Status == "starting" or v.Status == "started" or v.Status == "stopping") and v.DiskDevice == device then
			print( "Recording stream '" .. StreamsMap[k] .. "' is depended by this device, so the recording should be stopped." )
			record_m.stream[k].start = 0
		end
	end

	self:UmountDisk(device)
	print( "Disk is un-mounted." )
end

function _M:DiskList()
	local result = {}
	local k,v
	local cc = 0
	for k,v in pairs(self.ALL_DISKS) do
		result[k] = v.VOLNAME or ""
		cc = cc + 1
	end
	return cc, result
end

function _M:DiskDetail(device)
	print('DiskDetail()', device)
	local r = self.ALL_DISKS[device]
	if not r then
		return -1, {}
	else
		return 0, r
	end
end

function _M:DiskRecordList(device,filter)
	print('DiskRecordList()', device,filter)
	if not device then
		return -1, {{}}
	end

	local r = self.ALL_DISKS[device]
	if not r then
		return -1, {{}}
	else
		local rr = self.CheckRecordList( device, filter )
		if not rr then
			return -1, {{}}
		else
			return #rr, rr
		end
	end
end

----------------------------------------------------
-- Record interface
----------------------------------------------------

function _M:RegisterWorker( stream )
	print('RegisterWorker()', stream)
	local k,v
	local streamIdx
	if not stream or stream == "any" then
		streamIdx = -1
	elseif stream == "main" or stream == "0" then
		streamIdx = 0
	elseif stream == "sub" or stream == "1" then
		streamIdx = 1
	else
		streamIdx = -1
	end

	for k,v in pairs( self.RECORDING) do
		if streamIdx == -1 or streamIdx == k then
			if v.Status == "starting" then	--Pending, that 'streamer' module may not get the start message, so try again.
				self.StartRecord( k == 0 and "main" or "sub" )
			elseif v.Status == "started" then --Why a started job but new worker registered? I think the worker may crashed and start again!
											  --So I need stop current first, then try start again!
				self.StopRecord( k == 0 and "main" or "sub" )
				self.StartRecord( k == 0 and "main" or "sub" )
			elseif v.Status == "stopping" then --Pending, that 'streamer' module may not get the stop message, so try again.
				self.StopRecord( k == 0 and "main" or "sub" )
			end
		end
	end
end

function _M:AckStartRecord( sequence_id, ackCode )
	print('AckStartRecord()', sequence_id, ackCode)
	local k,v
	for k,v in pairs( self.RECORDING) do
		if v.Status == "starting" and v.SequenceID == sequence_id then
			if ackCode == "done" or ackCode == "started" then
				print( "INFO: The recording sequence of '" .. sequence_id .. "' is acked started." )
				v.Status = "started"
				v.ErrorAction = ""
				v.Error = ""
			else
				print( "INFO: The recording sequence of '" .. sequence_id .. "' is acked start error: " .. ackCode )
				v.Status = "stopped"
				v.ErrorAction = "start"
				v.Error = ackCode or "unknown"
			end
			break
		end
	end
	--TODO: recording status.
end

function _M:AckStopRecord( sequence_id, ackCode )
	print('AckStopRecord()', sequence_id, ackCode)
	local k,v
	for k,v in pairs( self.RECORDING) do
		if v.Status == "stopping" and v.SequenceID == sequence_id then
			if ackCode == "done" or ackCode == "stopped" then
				print( "INFO: The recording sequence of '" .. sequence_id .. "' is acked stopped." )
				v.Status = "stopped"
				v.ErrorAction = ""
				v.Error = ""
			else
				print( "INFO: The recording sequence of '" .. sequence_id .. "' is acked stop error: " .. ackCode )
				v.Status = "stopped"
				v.ErrorAction = "stop"
				v.Error = ackCode or "unknown"
			end

			v.CURRENT_FILE = nil
			v.CURRENT_START_TIME = nil
			v.CURRENT_END_TIME = nil
			v.START_TIME = nil
			v.END_TIME = nil
			v.FORMAT = nil
			v.FILES = nil
			v.action = nil
			collectgarbage()
			break
		end
	end
end

function _M:UpdateRecord( sequence_id, action, filename, format, time, size_kb )
	print('UpdateRecord()', sequence_id, action,filename, format, time, size_kb)
	--print( "INFO: Update recording: SequenceID='" .. sequence_id .. "', Action='" .. action .. "', Filename='" .. filename .. "' ..." )
	local k,v
	for k,v in pairs(self.RECORDING) do
		-- If matched sequence-id and current is started.
		if v.Status == "started" and v.SequenceID == sequence_id then
			v.action = action
			if action == "new" then
				v.CURRENT_FILE = filename
				v.CURRENT_START_TIME = tonumber(time)
				v.START_TIME = v.START_TIME or tonumber(time)
				v.END_TIME = tonumber(time) or v.START_TIME
				v.CURRENT_END_TIME = v.END_TIME
				v.FORMAT = format or "ts"
				if not v.FILES then
					v.FILES = { [filename] = size_kb or 0 }
				else
					v.FILES[ filename ] = size_kb or 0
				end
			else -- action == "update" or action == "end"
				v.CURRENT_FILE = filename --Just update it.
				v.FORMAT = format or v.FORMAT or "ts"
				v.START_TIME = v.START_TIME or tonumber(time)
				v.END_TIME = tonumber(time) or v.START_TIME
				v.CURRENT_END_TIME = v.END_TIME
				if not v.FILES then
					v.FILES = { [filename] = size_kb or 0 }
				else
					v.FILES[ filename ] = size_kb or 0
				end
			end

			v.action = action

			--TODO: Update the meta file.
			local diskMount = self.ALL_DISKS[v.DiskDevice].MOUNT
			if diskMount then
				local file
				os.execute( "/bin/mkdir -p " .. diskMount .. "/RECORD/" .. sequence_id )
				file = io.open( diskMount .. "/RECORD/" .. sequence_id .. "/.START", "w" )
				if file then
					file:write( v.START_TIME .. "\n" )
					file:close()
				end

				file = io.open( diskMount .. "/RECORD/" .. sequence_id .. "/.END", "w" )
				if file then
					file:write( v.END_TIME .. "\n" )
					file:close()
				end

				file = io.open( diskMount .. "/RECORD/" .. sequence_id .. "/.FORMAT", "w" )
				if file then
					file:write( v.FORMAT .. "\n" )
					file:close()
				end
			end
			return "ok"
		elseif v.SequenceID == sequence_id then
			if action == "stop" or action == "terminate" then
				record_m.stream[k].start = 0
				--_M.AckStopRecord( sequence_id, "stopped" )
				return "ok"
			end

			print( "***WARNING: A recording task is not 'started' but '" .. v.Status .. "', but record update message received!" )
			v.Error = "status-error"
			v.ErrorAction = "update-record"
			return "error"
		end
	end

	print( "***WARNING: Invalid sequence-id but record update message received!" )
	return "error"
end

function _M:GetStatus(stream)
	print('GetStatus()', stream)
	local streamIdx = 0
	if stream == "sub" or tonumber(stream) == 1 then
		streamIdx = 1
	end

	if self.RECORDING[streamIdx].Status ~= "started" then
		if not self.RECORDING[streamIdx].Status then
			return "idle",{}
		elseif self.RECORDING[streamIdx].Error and self.RECORDING[streamIdx].Error ~= "" then
				return "error:" .. (self.RECORDING[streamIdx].ErrorAction or "") .. ":" .. self.RECORDING[streamIdx].Error, {}
		else
			return self.RECORDING[streamIdx].Status, {}
		end
	else
		local items = self.CheckRecordList( self.RECORDING[streamIdx].DiskDevice, self.RECORDING[streamIdx].SequenceID )
		if items and items[1] then
			local v = self.RECORDING[streamIdx]
			items[1].CURRENT_FILE = v.CURRENT_FILE or ""
			items[1].CURRENT_START_TIME = v.CURRENT_START_TIME or 0
			items[1].CURRENT_END_TIME = v.CURRENT_END_TIME or 0
		end
--[[
		local detailInfo = {
			PATH = self.ALL_DISKS[v.DiskDevice].MOUNT .. "/RECORD/" .. v.SequenceID,
			STREAM = streamIdx == 0 and "main" or "sub",
			SEQUENCE = v.SequenceID,
			FILES = "",
			FORMAT = v.FORMAT or "ts",
			START_TIME = v.START_TIME or 0,
			END_TIME = v.END_TIME or 0,
			CURRENT_FILE = v.CURRENT_FILE or "",
			SIZE = 0
		}

		if v.FILES then
			local kk,vv
			for kk,vv in pairs(v.FILES) do
				detailInfo.FILES = detailInfo.FILES .. kk .. "=" .. (vv or 0) .. "\n"
				detailInfo.SIZE = detailInfo.SIZE + vv --Total size.
			end
		end
]]
		return "started", items and items[1] or {}
	end
end

----------------------------------------------------
-- Disk mount/umount
----------------------------------------------------
function _M:MountDisk( device, type, filesys)
	print('MountDisk()', device,type,filesys)
	if not device or not type or not filesys then
		return false, "Invalid arguments"
	end

	if  self.ALL_DISKS[device] then
		return false, "Already mount"
	end

	if type == "nas" then --Since NAS is already mounted.
		self.ALL_DISKS[device] = {
			DEVICE = deivce,
			TYPE = type,
			FILESYS = filesys,
			MOUNT = device
		}
		return true
	end

	local mountPath = "/media/" .. type .. "/" .. device
	local ret, status, code = os.execute( "/bin/mkdir -p " .. mountPath  )
	if not ret or status ~= "exit" or code ~= 0 then
		return false, "Fail to create mount point"
	end

	if filesys == "ntfs" then
		ret, status, code = os.execute( "/usr/bin/ntfs-3g -o utf8,big_writes,gid=100,umask=0 /dev/" .. device .. " " .. mountPath )
		if not ret or status ~= "exit" or code ~= 0 then
			--Try umount then mount, try it again.
			print( "*** First mount chance is fail. Try umount first, then try again." )
			os.execute( "/bin/umount -l /dev/" .. device )

			ret, status, code = os.execute( "/usr/bin/ntfs-3g -o utf8,big_writes,gid=100,umask=0 /dev/" .. device .. " " .. mountPath )
			if not ret or status ~= "exit" or code ~= 0 then
				return false, "Fail to mount NTFS device at '" .. mountPath .. "'"
			end
		end
	else
		ret, status, code = os.execute( "/bin/mount -t " .. filesys .. " -o utf8,gid=100,umask=0 /dev/" .. device .. " " .. mountPath )

		if not ret or status ~= "exit" or code ~= 0 then
			--Try umount then mount, try it again.
			print( "*** First mount chance is fail. Try umount first, then try again." )
			os.execute( "/bin/umount -l /dev/" .. device )
			ret, status, code = os.execute( "/bin/mount -t " .. filesys .. " -o utf8,gid=100,umask=0 /dev/" .. device .. " " .. mountPath )
			if not ret or status ~= "exit" or code ~= 0 then
				return false, "Fail to mount " .. filesys .. " device at '" .. mountPath .. "'"
			end
		end
	end

	self.ALL_DISKS[device] = {
		DEVICE = deivce,
		TYPE = type,
		FILESYS = filesys,
		MOUNT = mountPath
	}
	return true
end

----------------------------------------------------

function _M:UmountDisk( device )
	print('UmountDisk()', device)
	if not device then
		return false
	end

	local diskItem = self.ALL_DISKS[device]
	if not diskItem then
		return true
	end

	if diskItem.type ~= "nas" then
		os.execute( "/bin/umount -l " .. diskItem.MOUNT )
	end

	self.ALL_DISKS[device] = nil
	return true
end

----------------------------------------------------

function _M:GetDiskSpaces( device, isNAS )
	
	print('GetDiskSpaces()', device,isNAS)
	local file
	if isNAS then
		file = io.popen("/bin/df -m " .. device .. ' | /bin/grep /', "r")
	else
		file = io.popen("/bin/df -m /dev/" .. device .. ' | /bin/grep /',"r")
	end

	if not file then
		return false, "Fail to get disk spaces"
	end

	local line = file:read("a")
	local flag,total,used,free,percent = line:match( "(.-)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%%%s+" )
	local exit = file:close()

	if exit then
		return true, tonumber(total), tonumber(used), tonumber(free), tonumber(percent)
	else
		return false, "Fail to get disk spaces"
	end
end

----------------------------------------------------

function _M.Init( dbusService )
	print('Init()', dbusService)
	DBUS_SERVICE = dbusService
end

----------------------------------------------------
function _M.PickDisk( req )
	print('PickDisk()', req)
	req = req or "largest"

	local reqType = nil
	if req:match("sdmmc") then
		reqType = "sdmmc"
	elseif req:match( "usb" ) then
		reqType = "usb"
	elseif req:match( "nas" ) then
		reqType = "nas"
	end

	local reqMatch = "largest"
	if req:match( "first" ) then
		reqMatch = "first"
	elseif req:match( "mark" ) then
		reqMatch = "mark"
	end

	local shouldBe = {}
	local shouldCount = 0
	local largestDev
	local largestSize = 0
	local k,v
	for k,v in pairs( _M.ALL_DISKS ) do
		if not reqType or reqType == v.TYPE then
			if reqMatch == "first" then
				return k
			elseif reqMatch == "largest" then
				if not largestDev or v.FREE > largestSize then
					largestDev = k
					largestSize = v.FREE
				end
			else
				shouldBe[k] = v.MOUNT
				shouldCount = shouldCount + 1
			end
		end
	end

	if reqMatch == "largest" then
		return largestDev
	elseif reqMatch == "mark" then
		if shouldCount == 0 then
			return nil
		else
			local firstOne
			local file

			for k,v in pairs(shouldBe) do
				if not firstOne then
					firstOne = k
				end

				file = io.open( v .. "/RECORD_HERE.txt", "r" )
				if file then
					file:close()
					return k
				end
				file = io.open( v .. "/RECORD_HERE", "r" )
				if file then
					file:close()
					return k
				end
			end
			return firstOne
		end
	else
		return nil
	end
end
----------------------------------------------------

function _M.StartRecord( stream )
	print('StartRecord()', stream)

	_M.StopRecord(stream)

	if not stream or not DBUS_SERVICE then
		print( "ERROR: Start recording with invalid arguments or no valid DBUS service." )
		return false
	end

	local streamIdx = 0
	local streamFlag = "M"

	stream = stream:lower()
	if stream == "sub" or stream == "1" then
		stream = "sub"
		streamFlag = "S"
		streamIdx = 1
	else
		stream = "main"
	end

	local recordItem = _M.RECORDING[streamIdx]
	if not recordItem then
		return false
	end

	local alreadyStart = false
	if recordItem.Status == "starting" or recordItem.Status == "started" then
		print( "Start record [" .. stream .. "]: Recording is already started, just update it. Sequence-ID is [" .. recordItem.SequenceID .. "]" )
		alreadyStart = true
		recordItem.ErrorAction = ""
		recordItem.Error = ""
	else
		local now = os.date("*t")
		local nowStr = string.format("%04d%02d%02d_%02d%02d%02d_%04d", now.year, now.month, now.day, now.hour, now.min, now.sec, _M.SEQUENCE_COUNTER)
		recordItem.SequenceID = streamFlag .. "-" .. nowStr
		_M.SEQUENCE_COUNTER = _M.SEQUENCE_COUNTER + 1
		print( "Start record [" .. stream .. "]: Prepair to start recording, *NEW* Sequence-ID is [" .. recordItem.SequenceID .. "]" )
		recordItem.Status = "starting" -- Start pending.
		recordItem.ErrorAction = ""
		recordItem.Error = "" --Clear error status.
		recordItem.DiskDevice = nil	--Drop saved, if exists.

		recordItem.CURRENT_FILE = nil
		recordItem.CURRENT_START_TIME = nil
		recordItem.CURRENT_END_TIME = nil

		recordItem.START_TIME = nil
		recordItem.END_TIME = nil
		recordItem.FORMAT = nil
		recordItem.FILES = nil
		recordItem.action = nil
	end

	recordItem.DiskPolicy = record_m.stream[streamIdx].disk_policy	--Just update.

	local format = record_m.stream[streamIdx].format
	if format == "" then
		format = "ts"
	end

	local prefix = record_m.stream[streamIdx].file_prefix

	if recordItem.DiskDevice then
		--Check is it exist yet?
		if not _M.ALL_DISKS[recordItem.DiskDevice] then
			print( "WARNING: Current recording disk '" .. recordItem.DiskDevice .. "' is lost, try select another!" )
			recordItem.DiskDevice = nil
		end
	end

	if not recordItem.DiskDevice then
		--Auto pick a disk
		recordItem.DiskDevice = _M.PickDisk( record_m.stream[streamIdx].disk_type )
		if not recordItem.DiskDevice then
			print( "ERROR: No valid disk, recording cannot be started." )
			if alreadyStart then
				_M.StopRecord( stream )
			end
			recordItem.Status = "stopped"
			recordItem.ErrorAction = "start"
			recordItem.Error = "nodisk"
			return false
		else
			print( "Selected recording disk: '" .. recordItem.DiskDevice .. "'" )
		end
	end

	if not _M.ReserveDiskSpace( streamIdx, 300 ) then --Before record, must try keep 100MB spaces
		print( "ERROR: No disk space." )
		if alreadyStart then
			_M.StopRecord( stream )
		end
		recordItem.Status = "stopped"
		recordItem.ErrorAction = "start"
		recordItem.Error = "nospace"
		return false
	end

	local diskPath = _M.ALL_DISKS[recordItem.DiskDevice].MOUNT .. "/RECORD"

	os.execute( "/bin/mkdir -p " .. diskPath .. "/" .. recordItem.SequenceID )

	local limit_type = record_m.stream[streamIdx].limit_type
	local limit = 0
	if limit_type == "time" or limit_type == "time-loop" then
		limit = record_m.stream[streamIdx].limit_time
	elseif limit_type == "size" or limit_type == "size-loop" then
		limit = record_m.stream[streamIdx].limit_size
	else
		limit_type = "none"
	end

	print( "NOTIFY: Notify the recording is started." )
	DBUS_SERVICE:EmitSignal( _DECL.iRecordEvents, "NotifyStartRecord", _M.ALL_DISKS[recordItem.DiskDevice].FILESYS, stream, recordItem.SequenceID,
							  format, diskPath, prefix, limit_type, limit )

	return true
end

----------------------------------------------------

function _M.StopRecord( stream )
	print('StopRecord()', stream)
	if not stream or not DBUS_SERVICE then
		print( "ERROR: Stop recording with invalid arguments or no valid DBUS service." )
		return false
	end

	local streamIdx = 0
	stream = stream:lower()
	if stream == "sub" or stream == "1" then
		stream = "sub"
		streamIdx = 1
	else
		stream = "main"
	end

	local recordItem = _M.RECORDING[streamIdx]
	if not recordItem then
		return false
	end

	if recordItem.Status == "starting" or recordItem.Status == "started" then
		print( "Stop record [" .. stream .. "]: Sequence-ID is [" .. ( recordItem.SequenceID or "nil" ) .. "]" )
		recordItem.Status = "stopping" --stopping
	elseif recordItem.Status ~= "stopping" then	--If stopping, need notify again.
		print( "Stop record [" .. stream .. "]: Already stopped." )
		return true
	end

	print( "NOTIFY: Notify stop recording." )
	DBUS_SERVICE:EmitSignal( _DECL.iRecordEvents, "NotifyStopRecord", stream, recordItem.SequenceID or "" )

	return true
end

----------------------------------------------------
function _M.InitRecordDB( device )
	print('InitRecordDB()', device)
	if not device then
		print( "ERROR: Invalid disk device while initialize recording." )
		return false
	end

	local mountPoint = _M.ALL_DISKS[device].MOUNT
	if not mountPoint then
		print( "ERROR: Invalid disk mounts while initialize recording." )
		return false
	end

	local path = mountPoint .. "/RECORD"
	local ret, status, code = os.execute( "/bin/mkdir -p " .. path )
	if not ret or status ~= "exit" or code ~= 0 then
		print( "ERROR: Fail to initialize recording database." )
		return false
	end
	return true
end
----------------------------------------------------
function _M.CheckRecordList( device, filter )
	print('CheckRecordList()', device)
	if not device then
		print( "ERROR: Invalid disk device while initialize recording." )
		return false
	end

	if not _M.InitRecordDB(device) then
		return nil
	end

	local mountPoint = _M.ALL_DISKS[device].MOUNT
	local path = mountPoint .. "/RECORD"
	local result = {}
	local attrs
	local line
	local pathname

	local allEntries = {}
	local entry
	for entry in lfs.dir( path ) do
		if entry ~= "." and entry ~= ".." and (not filter or entry:match(filter)) then
			line = path .. "/" .. entry
			attrs = lfs.attributes(line)
			if attrs and attrs.mode == "directory" then
				table.insert( allEntries, {
					TIME = attrs.change,
					FULL = line,
					DIRNAME = entry
				})
			end
		end
	end

	table.sort( allEntries, function(a,b)
		return a.TIME < b.TIME
	end)

	local k,v
	for k,v in pairs( allEntries ) do
		line = v.FULL
		pathname = v.DIRNAME

		local stream = pathname:match( "([msMS]*)-[0-9_%-%:]+" )

		if stream then
			stream = stream:upper()
			local sequence = pathname
			local item
			local startTime = nil
			local endTime = nil
			local format = nil
			local snapfile = nil
			local file

			file = io.open( line .. "/.START", "r" )
			if file then
				startTime = tonumber( file:read("l") ) or 0
				file:close()
			end

			file = io.open( line .. "/.END", "r" )
			if file then
				endTime = tonumber( file:read("l") ) or 0
				file:close()
			end

			file = io.open( line .. "/.FORMAT", "r" )
			if file then
				format = file:read("l")
				file:close()
			end

			file = io.open( line .. "/SNAPSHOT.jpg", "r" )
			if file then
				snapfile = line .. "/SNAPSHOT.jpg"
				file:close()
			end

			local filelist = {}
			local totalSize = 0
			local fileCounts = 0
			local cmd = '/bin/ls -l ' .. line .. ' | /bin/egrep "\\.+[ts]|[mp4]|[mkv]|[mov]|[avi]|[264]|[inc]$"'

			file = io.popen(cmd)

			while (1) do
				local s = file:read('l'	)
				if (s == nil) then break end
				local file_size,file_name = string.match(s,'%w+%s+%d+%s+%w+%s+%w+%s+(%d+)%s+%w+%s+%d+%s+%d+%:?%d+%s+(.+)')
				if (file_name) then
					filelist[file_name] = file_size
					fileCounts = fileCounts + 1
					totalSize = totalSize + file_size
				else
					print(cmd,s,file_name)
				end
			end

			file:close()

			local rec = {
				DEVICE = device,
				PATH = line,
				STREAM = (stream == "M" and "main" or "sub"), --NOTE: only consider main and sub.
				SEQUENCE = sequence,
				FILES = "",
				FILE_COUNT = fileCounts,
				START_TIME = startTime,
				END_TIME = endTime,
				FORMAT = format,
				SIZE = totalSize,
				SNAPSHOT = snapfile
			}

			local kkk,vvv
			for kkk,vvv in pairs(filelist) do
				rec.FILES = rec.FILES .. kkk .. "=" .. (vvv or 0) .. "\n"
			end

			local recordItem = _M.RECORDING[ stream == "M" and 0 or 1 ]
			if (recordItem.Status == "started" or recordItem.Status == "starting") and recordItem.SequenceID == sequence then
				rec._CURRENT = true
			end

			table.insert( result, rec )
		end --If get stream id and sequence.
	end --For each entry.

	return result
end
----------------------------------------------------

function _M.ReserveDiskSpace( streamIdx, sizeMB )
	print('ReserveDiskSpace()', streamIdx,sizeMB)
	if streamIdx < 0 or streamIdx > 1 then
		return nil
	end

	local recordItem = _M.RECORDING[streamIdx]
	if not recordItem or not recordItem.DiskDevice then
		return nil
	end

	local free = _M.ALL_DISKS[recordItem.DiskDevice].FREE or 0
	if free >= sizeMB then
		return free
	end
	if free <= 0 then
		return nil
	end
	local limit_type={
		["size-loop"] = "Limit size and automatically cut multiple files",
		["time-loop"] = "Limit time and automatically cut multiple files",
	}
	if record_m.stream[streamIdx].disk_policy == "overwrite" and limit_type[record_m.stream[streamIdx].limit_type]  then
		local rec_list = _M.CheckRecordList( recordItem.DiskDevice )
		--First, try drop current stream records.
		local i
		local stream = "main"
		if streamIdx > 0 then
			stream = "sub"
		end

		for i=1, #rec_list do
			if not rec_list[i]._CURRENT then	--Not current file.
				if rec_list[i].STREAM == stream then	--Matched.
					print( "*** REMOVE(Overwrite) old records: " .. rec_list[i].PATH )
					local r,s,c = os.execute( "/bin/rm -Rf " .. rec_list[i].PATH )
					if r and s == "exit" and c == 0 then
						rec_list[i]._KICKED = true
						free = free + (rec_list[i].SIZE/1024)	--To MB(size is KB)
					end
				end
			else
				local size = 0
				local j = 0
				for k=1, #rec_list do
						if rec_list[k]._CURRENT and rec_list[k].FILE_COUNT > 1 then
								if rec_list[k].SIZE >= size then
										size = rec_list[k].SIZE
										j = k
								end
						end
				end
				if j ~= 0 then
					local l= rec_list[j].FILES
					list = {}
					for w in l:gmatch("%w+-%d+.[ts|mp4|mov|mkv|avi]+=%d+") do table.insert(list, w) end
					if #list == 0 then
							for w in l:gmatch("-%d+.[ts|mp4|mov|mkv|avi]+=%d+") do table.insert(list, w) end
					end
					table.sort(list)
					if #list > 0 then
						local file = list[1]
						local file_size = string.gmatch(string.gmatch(file,"(=%d+)")(),"(%d+)")()
						file = string.gmatch(file,"%w+-%d+.[ts|mp4|mov|mkv|avi]+")() or string.gmatch(file,"-%d+.[ts|mp4|mov|mkv|avi]+")()
						print( "*** REMOVE(Overwrite) old records: " .. rec_list[j].PATH.."/"..file )
						local r,s,c = os.execute( "/bin/rm -Rf " .. rec_list[j].PATH.."/"..file )
						if r and s == "exit" and c == 0 then
								free = free + file_size / 1024         --To MB(size is KB)
						end
					end
				end
			end

			if free >= sizeMB then
				return free
			end
		end

		--Then, try drop other streams record.
		for i=1, #rec_list do
			if not rec_list[i]._CURRENT and not rec_list[i]._KICKED then	--Not current file, and not processed.
				print( "*** REMOVE(Overwrite) old records: " .. rec_list[i].PATH )
				local r,s,c = os.execute( "/bin/rm -Rf " .. rec_list[i].PATH )
				if r and s == "exit" and c == 0 then
					rec_list[i]._KICKED = true
					free = free + (rec_list[i].SIZE/1024)	--To MB(size is KB)
				end
			end

			if free >= sizeMB then
				return free
			end
		end

		return nil
	else
		return nil
	end
end

----------------------------------------------------

-- Step check to update recording [ Generally, I need to reserve disk spaces ]
-- The master control program need call this period, ~about 5 seconds each time.

function _M.StepCheck()
	print('StepCheck()')
	local k,v
	--Update the disk free spaces first.
	for k,v in pairs(_M.ALL_DISKS) do
		local done, total, used, free, percent = _M:GetDiskSpaces( k, v.TYPE == "nas" )
		if done then
			--print( "Disk '" .. device .. "' info: TOTAL=" .. total .. " MB, USED=" .. used .. " MB, FREE=" .. free .. " MB, Percent=" .. percent .. "%" )
			v.TOTAL = total
			v.USED = used
			v.FREE = free
			v.PERCENT = percent
		end
	end

	for k,v in pairs(_M.RECORDING) do
		if v.Status == "started" then
			local limit_type = record_m.stream[k].limit_type
			
			if(v.action == nil) then
				if(v.action_timeout) then v.action_timeout = v.action_timeout + 1
				else v.action_timeout = 0 end
			end

			if v.action_timeout == 4 or (v.action == 'end' and (limit_type == 'size' or limit_type == 'time')) then
				_M.StopRecord( k == 0 and "main" or "sub" )
				v.action_timeout = nil
			end

			if not _M.ReserveDiskSpace( k, 300 ) then
				print( "WARNING: Insufficent disk spaces for stream " .. k .. " recording, so stop the recording!" )
				v.ErrorAction = "recording"
				v.Error = "nospace"
				_M.StopRecord( k == 0 and "main" or "sub" )
			end
		end
	end
end

----------------------------------------------------
return _M
