--local coaf = require "coaf.core"
local _dbg = require "coaf.debug"

local META = {
	max_packet_size = 4096
}

function META:Connect(...)
	local thisObj = self.__owner

	local content = {
		CMD = "STREAM",
		ARGS = "connect",
		--SN will be auto filled by Send(...)
		headers = {
			SVC = thisObj.__internal_client.serviceId,
			LOC = thisObj.__internal_client.location,
			OBJ = thisObj.__internal_objectPath,
			--VIA will be auto filled by Send(...)
		},
		parameters = {...}
	}

	if not thisObj.__internal_connection:Send(content) then
		_dbg.Error("CoAF.Client", "Proxy - fail send STREAM connect message")
		return false
	end

	self.connected = false
	local start_time = os.time()
	local ack
	while true do
		ack = thisObj.__internal_connection:RecvACK()
		if not ack then
			_dbg.Error("CoAF.Client", "Proxy - fail get ACK/STREAM message while connect")
			return
		end
		if ack.CMD == "ACK/STREAM" and ack.ARGS == "connect" and tonumber(ack.SN) == content.SN then
			break
		end
		if (os.time() - start_time) < 0 or (os.time() - start_time) > 15 then
			_dbg.Error( "CoAF.Client", "Proxy - Connect STREAM timeout" )
			return false
		end
	end

	if not ack.headers or not ack.headers.RESULT then
		_dbg.Error("CoAF.Client", "Proxy - invalid ACK/STREAM connect message")
		return false
	end

	if ack.headers.RESULT ~= "OK" then
		_dbg.Info("CoAF.Client", "Proxy - Connect STREAM error:", ack.headers.RESULT, ack.headers.MSG )
		return false
	end

	self.connected = true
	return true
end

function META:Disconnect()
	if not self.connected then
		return true
	end

	local thisObj = self.__owner

	local content = {
		CMD = "STREAM",
		ARGS = "disconnect",
		--SN will be auto filled by Send(...)
		headers = {
			SVC = thisObj.__internal_client.serviceId,
			LOC = thisObj.__internal_client.location,
			OBJ = thisObj.__internal_objectPath,
			--VIA will be auto filled by Send(...)
		}
	}

	if not thisObj.__internal_connection:Send(content) then
		_dbg.Error("CoAF.Client", "Proxy - fail send STREAM disconnect message")
		return false
	end

	local start_time = os.time()
	while true do
		local ack = thisObj.__internal_connection:RecvACK()
		if not ack then
			_dbg.Error("CoAF.Client", "Proxy - fail get ACK/STREAM message while disconnect")
			return
		end
		if ack.CMD == "ACK/STREAM" and ack.ARGS == "disconnect" --[[and tonumber(ack.SN) == content.SN]] then
			break
		end
		if (os.time() - start_time) < 0 or (os.time() - start_time) > 15 then
			_dbg.Error( "CoAF.Client", "Proxy - Wait disconnect STREAM timeout, force it disconnected." )
			break
		end
	end

	self.connected = false
	return true
end

function META:Read(size_pattern, timeout)
	local cont,flag,size = self:__InternalRead(size_pattern)
	if not cont and flag == "partial" then
		local thisObj = self.__owner
		if not self.connected then
			return nil, "closed", 0
		end

		while true do
			local ack, err = thisObj.__internal_connection:RecvACK(nil, timeout)
			if not ack then
				return nil, err, 0
			end
			if ack.CMD == "ACK/STREAM" and ack.ARGS == "disconnect" then
				self.connected = false
				return nil, "closed", 0
			end
			if ack.CMD == "ACK/STREAM" and ack.ARGS == "content" then
				if ack.data and #ack.data > 0 then
					if type(ack.data) == "table" then
						local i
						for i=1, #ack.data do
							table.insert( self.data, {
								data = ack.data[i],
								size = #ack.data[i],
								read_pos = 1
							})
						end
					else
						table.insert( self.data, {
							data = ack.data,
							size = #ack.data,
							read_pos = 1 
						})
					end
					break
				end
			end
		end
		return self:__InternalRead(size_pattern)
	else
		return cont,flag,size
	end
end

function META:__InternalRead(size_pattern)
	local ret

	local size = tonumber(size_pattern)
	if size then
		if size > 0 then
			local i,item
			local actual_size = 0
			local this_size
			local drop_to
			for i,item in ipairs(self.data) do
				if item.read_pos <= item.size then
					this_size = (item.size - item.read_pos) + 1
					if this_size > size then
						this_size = size
						if not ret then
							ret = { item.data:sub( item.read_pos, item.read_pos + this_size - 1 ) }
						else
							table.insert( ret, { item.data:sub( item.read_pos, item.read_pos + this_size - 1) } )
						end
						item.read_pos = item.read_pos + this_size
					elseif item.read_pos == 1 then
						if not ret then
							ret = { item.data }
						else
							table.insert( ret, item.data )
						end
						drop_to = i
					else
						if not ret then
							ret = { item.data:sub( item.read_pos ) }
						else
							table.insert( ret, item.data:sub( item.read_pos ) )
						end
						drop_to = i
					end
					actual_size = actual_size + this_size
					size = size - this_size
					if size == 0 then
						break
					end
				else
					drop_to = i
				end
			end

			if drop_to then
				for i=1, drop_to do
					table.remove( self.data, 1 )
				end
				collectgarbage()
			end

			if not ret then
				return nil, "partial", 0
			end

			if actual_size >= size then
				return table.concat(ret), nil, actual_size
			else
				return table.concat(ret), "partial", actual_size
			end
		else
			return nil, "invalid", 0
		end
	elseif not size_pattern or size_pattern == "*a" or size_pattern == "a" or size_pattern == "*" or size_pattern == "" then
		local i,item
		local actual_size = 0
		local this_size
		for i,item in ipairs(self.data) do
			if item.read_pos <= item.size then
				this_size = (item.size - item.read_pos) + 1
				if item.read_pos == 1 then
					if not ret then
						ret = { item.data }
					else
						table.insert( ret, item.data )
					end
				else
					if not ret then
						ret = { item.data:sub( item.read_pos ) }
					else
						table.insert( ret, item.data:sub( item.read_pos ) )
					end
				end
				actual_size = actual_size + this_size
			end
		end
		if not ret then
			return nil, "partial", 0
		end
		self.data = {}
		collectgarbage()
		return table.concat(ret), nil, actual_size
	elseif size_pattern == "*l" or size_pattern == "l" then
		local i,item
		local actual_size = 0
		local drop_to, ensured
		local token,rem
		local this_size
		for i,item in ipairs(self.data) do
			if item.read_pos <= item.size then
				this_size = (item.size - item.read_pos) + 1
				local token, rem
				if item.read_pos == 1 then
					token,rem = item.data:match("^(.-)[\r\n]+(.*)$")
				else
					token,rem = item.data:sub(item.read_pos):match("^(.-)[\r\n]+(.*)$")
				end

				if token then
					this_size = #token
					if not ret then
						ret = { token }
					else
						table.insert( ret, token )
					end
					if rem == "" then
						drop_to = i
					else
						drop_to = i-1
						item.data = rem
						item.size = #rem
						item.read_pos = 1
					end
					actual_size = actual_size + this_size
					ensured = true
					break
				elseif item.read_pos == 1 then
					if not ret then
						ret = { item.data }
					else
						table.insert( ret, item.data )
					end
				else
					if not ret then
						ret = { item.data:sub(item.read_pos) }
					else
						table.insert( ret, item.data:sub(item.read_pos) )
					end
				end
				actual_size = actual_size + this_size
			else
				drop_to = i
			end
		end

		if drop_to then
			for i=1, drop_to do
				table.remove( self.data, 1 )
			end
			collectgarbage()
		end

		if not ensured then
			return nil, "partial", 0
		end

		return table.concat(ret), nil, actual_size
	else
		return nil, "invalid", 0
	end
end

function META:Write(data,i,j)
	local thisObj = self.__owner
	if not self.connected then
		return -1, "closed"
	end
	if not data then
		return -1, "invalid"
	end

	i = i or 1
	j = j or #data

	local sent = 0
	local pkt, pkt_size
	local cc, flag
	while i <= j do
		pkt_size = j-i+1
		if pkt_size > META.max_packet_size then
			pkt_size = META.max_packet_size
		end

		pkt = data:sub(i, i+pkt_size-1)
		sent = sent + pkt_size
		i = i + pkt_size
		local content = {
			--SN will auto fill.
			CMD = "STREAM",
			ARGS = "content",
			headers = {
				SVC = thisObj.__internal_client.serviceId,
				LOC = thisObj.__internal_client.location,
				OBJ = thisObj.__internal_objectPath,
				--VIA will be auto filled by Send(...)
			},
			data = pkt
		}
		cc, flag = thisObj.__internal_connection:Send(content)
		if not cc then
			return nil, flag
		end
	end
	return sent, nil
end

META.__index = META

return META

