local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"
local services = require "coserver.services"
local actions = require "coserver.actions"
local client = require "coaf.client"

-----------
local _M = {
	PROBE_TIMEOUT = 3,
	PROXY_TIMEOUT = 30,
	PROBE_INTERVAL = 10,

	all_services = {},
	host_items = {},
	host_items_changed = false,
	order_count = 0
}

-----------
local function request_proxy_run(thrObj,svcObj)
	local remote_link, stat
	local conn_daemon = require "coserver.conn_daemon"

	while not thrObj.terminated do
		if svcObj.status ~= "ready" or not svcObj.started then
			thrObj.is_sleeping = true
			coaf.Sleep(5)
			thrObj.is_sleeping = false
		else
			remote_link, stat = conn_daemon.RequestProxy( svcObj.target_addr .. ":" .. svcObj.target_port, 
														svcObj.location,
														svcObj.service,
														30,
														svcObj.dynamic_key )
			if remote_link then
				stat = remote_link:WaitDie()
				_dbg.Info("CoServer.DynService", "Location:", svcObj.location, "proxy is disconnected, state:", stat )
				if stat == "#CoAF-KILL#" and type(remote_link.Destroy) == "function" then
					remote_link:Destroy()
				end
				remote_link = nil
			elseif not thrObj.terminated then
				_dbg.Error("CoServer.DynService", "Location:", svcObj.location, "request proxy error:", stat )
				thrObj.is_sleeping = true
				coaf.Sleep(2)
				thrObj.is_sleeping = false
			end
		end
	end
end
-----------
local function proxy_link_run(thrObj,svcObj)
	local proxy_link, stat
	local conn_daemon = require "coserver.conn_daemon"

	while not thrObj.terminated do
		if svcObj.status ~= "ready" then
			thrObj.is_sleeping = true
			coaf.Sleep(5)
			thrObj.is_sleeping = false
		else
			proxy_link = conn_daemon.NewProxyLink( svcObj.target_addr .. ":" .. svcObj.target_port, 
													30, --Keep-alive timeout
													nil,
													svcObj.self_location,
													svcObj.dynamic_key
												 )
			if proxy_link then
				stat = proxy_link:WaitDie()
				_dbg.Info("CoServer.DynService", "Location:", svcObj.location, "reverse proxy link is disconnected, state:", stat )
				if stat == "#CoAF-KILL#" and type(proxy_link.Destroy) == "function" then
					proxy_link:Destroy()
				end
				proxy_link = nil
			elseif not thrObj.terminated then
				_dbg.Error("CoServer.DynService", "Location:", svcObj.location, "reverse proxy link error:", stat )
				thrObj.is_sleeping = true
				coaf.Sleep(2)
				thrObj.is_sleeping = false
			end
		end
	end
end
-----------
local function removeCoafService( svcObj )
	if svcObj.service_added then
		svcObj.service_added = nil
		services.RemoveService( svcObj.location, svcObj.service )
	end
end

local function addCoafService( svcObj )
	if not svcObj.service_added and svcObj.service_config then
		svcObj.service_added = services.AddService( svcObj.location, svcObj.service, svcObj.service_config )
		if not svcObj.service_added then
			_dbg.Error( "CoServer.DynService", "Add dynamic service:", svcObj.location .. "/" .. svcObj.service, "to coserver fail!" )
		end
	end
end

-----------
local function probe_task_run(thrObj,svcObj)
	_dbg.Info( "CoServer.DynService", "Service:", svcObj.location .. "/" .. svcObj.service, "probe task started." )
	local cocore = require "coserver.core"

	local unique_id = cocore.configs.UNIQUE_ID
	local cli, proxy
	local dynamic_key_ready
	local my_loc
	local need_proxy_link, proxy_link_thread

	local ret
	local should_sleep

	local addr_idx = 1
	local current_addr, current_port, current_authorKey

	svcObj.status = "trying"
	while not thrObj.terminated do
		--Delayed terminate processing.
		if svcObj.flag_delay_term then
			local prev_status = svcObj.status
			svcObj.status = "terminating"
			svcObj.is_sleeping = true
			coaf.Sleep(svcObj.flag_delay_term)
			svcObj.is_sleeping = false
			--Not cancelled, so really terminate.
			if thrObj.terminated or svcObj.flag_delay_term then
				svcObj.flag_delay_term = nil
				break
			else
				svcObj.status = prev_status
			end
		end

		if svcObj.flag_reset_addr then
			svcObj.flag_reset_addr = nil
			addr_idx = 1
			try_count = 0
		end

		local timeout = svcObj.manual and _M.PROXY_TIMEOUT or _M.PROBE_TIMEOUT

		if svcObj.target_addr then
			current_addr = svcObj.target_addr
			current_port = svcObj.target_port or 14000
			current_authorKey = svcObj.authorKey
			--If has address list, assume this time try will fail.
			--   but if connected, it will re-assign these values
			--   later; otherwise, drop current address and try
			--   other addresses in the list in next try.
			if svcObj.address_list and #svcObj.address_list > 0 then
				svcObj.target_addr = nil
				svcObj.target_port = nil
				svcObj.target_authorKey = nil
				should_sleep = 0.5
			else
				should_sleep = _M.PROBE_INTERVAL
			end

		elseif not svcObj.address_list or #svcObj.address_list == 0 then
			svcObj.status = "idle"
			removeCoafService( svcObj )
			current_addr = nil
			should_sleep = _M.PROBE_INTERVAL

		else 
			if addr_idx < 1 or addr_idx > #svcObj.address_list then
				addr_idx = 1
			end
			current_addr = svcObj.address_list[addr_idx].address
			current_port = svcObj.address_list[addr_idx].port or 14000
			current_authorKey = svcObj.address_list[addr_idx].authorKey
			addr_idx = addr_idx + 1

			if addr_idx <= #svcObj.address_list then
				should_sleep = 0.5 --Next try as fast as posible.
			else
				should_sleep = _M.PROBE_INTERVAL
			end
			timeout = svcObj.address_try_timeout or _M.PROBE_TIMEOUT
		end

		if current_addr then
			_M.UpdateHostItem( svcObj.location, current_addr, false ) --Don't force update.
			--Record them, for get status.
			svcObj.last_try_addr = current_addr
			svcObj.last_try_port = current_port
			svcObj.last_try_authorKey = current_authorKey

			my_loc = unique_id
			if type(my_loc) == "function" then
				my_loc = my_loc()
			end
			if not my_loc then
				my_loc = coaf.GetHostname()
			end

			cli = client.New( "_CO_DAEMON.discovery", "local", current_addr .. ":" .. current_port, timeout, current_authorKey, my_loc )
		end

		if cli and not thrObj.terminated then
			_M.UpdateHostItem( svcObj.location, current_addr, true ) --Now force update the address (likely to win)
			proxy = cli:CreateProxy( "/", true )

			--Note: unlikely
			if not proxy then
				svcObj.status = "trying"
				removeCoafService( svcObj )
				cli:Destroy()
				cli = nil
				if should_sleep and not thrObj.terminated then
					svcObj.is_sleeping = true
					coaf.Sleep(should_sleep)
					svcObj.is_sleeping = false
				end
			else
				svcObj.target_addr = current_addr
				svcObj.target_port = current_port
				svcObj.authorKey = current_authorKey

				local alias

				-----------------------------------
				if thrObj.terminated then break end
				-----------------------------------

				ret, alias = proxy.TellMeMyAlias(my_loc)
				if ret and alias then
					svcObj.self_location = alias
				else
					svcObj.self_location = my_loc
				end

				--For debug purpose:
				--[[
				-----------------------------------
				if thrObj.terminated then break end
				-----------------------------------
				local rem_my_addr, rem_my_port
				ret, rem_my_addr, rem_my_port = proxy.TellMeMyAddress(my_loc)
				if ret then
					_dbg.Info( "CoServer.DynService", "Service probed - [" .. svcObj.self_location .."]", rem_my_addr, ":", rem_my_port, " to [" .. svcObj.location .. "]", current_addr, ":", current_port )
				end
				--]]
				-------------------

				local touch_fail
				local answer
				if svcObj.flag_no_bidi then
					ret = true
					answer = "yes" --Don't proxy link.
				else
					-----------------------------------
					if thrObj.terminated then break end
					-----------------------------------
					ret, answer = proxy.CanYouTouchMe( svcObj.self_location )
				end

				if not ret then
					_dbg.Error( "CoServer.DynService", "Touch remote [", svcObj.location, "] with error:", answer, ", service is unreachable." )
					touch_fail = true
				else
					-----------------------------------
					if thrObj.terminated then break end
					-----------------------------------
					dynamic_key_ready, svcObj.dynamic_key = proxy.RequestDynamicAuthorKey( my_loc, svcObj.self_location )

					if not dynamic_key_ready then
						touch_fail = true
						svcObj.dynamic_key = nil
						_dbg.Error( "CoServer.DynService", "Fail to exchange secrete key with remote [", svcObj.location, "], service is unreachable." )
					else --Try get publish_options for this.
						-----------------------------------
						if thrObj.terminated then break end
						-----------------------------------
						local opts
						ret, opts = proxy.GetPublishOptions()
						if ret and opts then
							svcObj.publish_options = opts
						end
					end
				end

				proxy:Destroy()
				cli:Destroy()
				proxy = nil
				cli = nil

				if touch_fail then
					--So try again.
					svcObj.status = "trying"
					removeCoafService( svcObj )

					if should_sleep and not thrObj.terminated then
						svcObj.is_sleeping = true
						coaf.Sleep(should_sleep)
						svcObj.is_sleeping = false
					end
				elseif not thrObj.terminated then
					svcObj.status = "ready"
					addCoafService( svcObj )
					need_proxy_link = (answer == "no" or answer == "exist")

					if need_proxy_link then
						if not proxy_link_thread then
							proxy_link_thread = thread.New( proxy_link_run )
							proxy_link_thread:Start( svcObj )
						elseif proxy_link_thread.is_sleeping then
							proxy_link_thread:Kill("wakeup")
						end
					else
						if proxy_link_thread then
							proxy_link_thread:Terminate()
							proxy_link_thread = nil
						end
					end

					if svcObj.proxy_thread and svcObj.proxy_thread.is_sleeping then
						svcObj.proxy_thread:Kill("wakeup")
					end

					svcObj.is_sleeping = true
					coaf.Sleep(_M.PROBE_INTERVAL)
					svcObj.is_sleeping = false
				end
			--} Touch done.
			end

		elseif not thrObj.terminated then
			svcObj.status = "trying"
			removeCoafService( svcObj )
			if should_sleep then
				svcObj.is_sleeping = true
				coaf.Sleep(should_sleep)
				svcObj.is_sleeping = false
			end
		end
	end

	if proxy_link_thread then
		proxy_link_thread:Terminate()
		proxy_link_thread = nil
	end

	if svcObj.proxy_thread then
		svcObj.proxy_thread:Terminate()
		svcObj.proxy_thread = nil
	end

	svcObj.status = "dead"
	svcObj.task_thread = nil

	removeCoafService( svcObj )

	if proxy then
		proxy:Destroy()
	end
	if cli then
		cli:Destroy()
	end
	proxy = nil
	cli = nil

	_M.all_services[svcObj.location .. "/" .. svcObj.service] = nil
	_dbg.Info( "CoServer.DynService", "Service:", svcObj.location .. "/" .. svcObj.service, "probe task terminated." )
end

-----------
--Note the 'noBidiCheck': if set, do not check the bidi connection.
function _M.AddService(loc, svc, target_addr, target_port, authKey, noBidiCheck)
	if not loc or not svc then
		_dbg.Error( "CoServer.DynService", "Add service but invalid location/service" )
		return nil, "invalid"
	end
	local key = loc .. "/" .. svc
	if _M.all_services[key] then
		local svc_obj = _M.all_services[key]
		if svc_obj.task_thread and svc_obj.flag_delay_term and svc_obj.is_sleeping then
			svc_obj.flag_delay_term = nil

			--If address arguments are nil, keep old values.
			svc_obj.target_addr = target_addr or svc_obj.target_addr
			svc_obj.target_port = target_port or svc_obj.target_port
			svc_obj.authorKey = authKey or svc_obj.authorKey

			svc_obj.task_thread:Kill("wakeup")
			return key
		end
		return nil, "exist"
	end

	local svc_cfg = {
		LOC = loc,
		ID = svc,
		ACTION = actions.DYNAMIC_SERVICE(loc,svc),
		--AUTHOR_KEY = authKey
	}

	--[[
	local svc_cfg = services.AddService( loc, svc, svc_cfg )
	if not svc_cfg then
		_dbg.Error( "CoServer.DynService", "Add dynamic service:", loc .. "/" .. svc, "to coserver fail!" )
		return nil, "fail"
	end
	]]

	local task = thread.New( probe_task_run )
	if not task then
		--services.RemoveService( loc, svc )
		_dbg.Error( "CoServer.DynService", "Add dynamic service:", loc .. "/" .. svc, "but create probe task fail!" )
		return nil, "fail"
	end

	local svc_obj = {
		location = loc,
		service = svc,
		target_addr = target_addr,
		target_port = target_port,
		authorKey = authKey,

		last_try_addr = target_addr,
		last_try_port = target_port,
		last_try_authorKey = authKey,
		
		address_list = nil,

		service_config = svc_cfg,
		service_added = false,

		is_sleeping = false,
		
		flag_reset_addr = false,
		flag_delay_term = nil,
		flag_no_bidi = noBidiCheck,

		status = "idle",
		task_thread = task,
		proxy_thread = nil,

		order = _M.order_count
	}

	_M.order_count = _M.order_count + 1

	_M.all_services[key] = svc_obj
	task:Start( svc_obj )
	return key
end

function _M.SetAddressList(loc, svc, list, try_timeout)
	if not loc or not svc then
		return false, "invalid"
	end
	local key = loc .. "/" .. svc
	local svc_obj = _M.all_services[key]
	if not svc_obj then
		return false
	end
	svc_obj.address_list = list

	--The discovered addresses are always in local area network.
	--So the try connection timeout be 2 seconds.
	svc_obj.address_try_timeout = try_timeout or _M.PROBE_TIMEOUT

	svc_obj.flag_reset_addr = true

	if svc_obj.is_sleeping then
		svc_obj.task_thread:Kill("wakeup")
	end

	return true
end

function _M.AddressListRetry(loc, svc)
	if not loc or not svc then
		return false, "invalid"
	end
	local key = loc .. "/" .. svc
	local svc_obj = _M.all_services[key]
	if not svc_obj then
		return false
	end

	svc_obj.flag_reset_addr = true

	if svc_obj.is_sleeping then
		svc_obj.task_thread:Kill("wakeup")
	end
	return true
end

--NOTE: 
--  if 'delayed' (a number, as seconds) specified, the service
--  will wait delayed time then terminate; In the delayed time,
--  you can cancel the terminating.

function _M.RemoveService(loc, svc, delayed)
	if not loc or not svc then
		_dbg.Error( "CoServer.DynService", "Remove service but invalid location/service" )
		return false, "invalid"
	end

	_M.RemoveHostItem(loc)

	local key = loc .. "/" .. svc
	local svc_obj = _M.all_services[key]
	if not svc_obj then
		return true
	end

	if type(delayed) == "number" and svc_obj.task_thread then
		svc_obj.flag_delay_term = delayed
	else
		delayed = nil
		--Drop flag if previous set.
		svc_obj.flag_delay_term = nil
	end

	--NOTE: This is done by after thread terminated.
	--removeCoafService( svc_obj )

	if delayed then
		if svc_obj.is_sleeping then
			svc_obj.task_thread:Kill("wakeup")
		end
		return true
	end

	if svc_obj.task_thread then
		svc_obj.task_thread:Terminate()
		svc_obj.task_thread = nil
	end

	--NOTE: This is done by after thread terminated.
	--_M.all_services[key] = nil
	
	local hasMore
	local k,v
	for k,v in pairs(_M.all_services) do
		hasMore = true
		break
	end

	if not hasMore then
		_M.order_count = 0
	end

	return true
end

--TODO: improve this.
function _M.GetServiceDetail(loc, svc)
	if not loc or not svc then
		return nil
	end
	local key = loc .. "/" .. svc
	local svc_obj = _M.all_services[key]
	if not svc_obj then
		return nil
	end

	return svc_obj
end

function _M.ClearAllServices()
	local k,svc_obj
	for k,svc_obj in pairs(_M.all_services) do
		_M.RemoveHostItem( svc_obj.location )
		removeCoafService( svc_obj )
		if svc_obj.task_thread then
			svc_obj.task_thread:Terminate()
			svc_obj.task_thread = nil
		end
	end

	_M.all_services = {}
	collectgarbage()
	_M.order_count = 0
end

function _M.GetServices(order)
	if not order then
		return _M.all_services
	end

	local ret = {}
	local k,v
	for k,v in pairs(_M.all_services) do
		ret[k] = v
		table.insert(ret,v)
	end

	table.sort( ret, function(a,b)
		if a == b then
			return false
		elseif not a.order or not b.order then
			return false
		else
			return a.order <= b.order
		end
	end)

	return ret
end

function _M.StartService(loc, svc)
	if not loc or not svc then
		_dbg.Error( "CoServer.DynService", "Start service but invalid location/service" )
		return false, "invalid"
	end
	local key = loc .. "/" .. svc
	local svc_obj = _M.all_services[key]
	if not svc_obj or not svc_obj.task_thread then
		return false, "not-exist"
	else
		svc_obj.started = true
		if not svc_obj.proxy_thread then
			svc_obj.proxy_thread = thread.New( request_proxy_run )
			svc_obj.proxy_thread:Start( svc_obj )
		elseif svc_obj.proxy_thread.is_sleeping then
			svc_obj.proxy_thread:Kill("wakeup")
		end
		return true
	end
end

function _M.StopService(loc, svc)
	if not loc or not svc then
		_dbg.Error( "CoServer.DynService", "Stop service but invalid location/service" )
		return false, "invalid"
	end
	local key = loc .. "/" .. svc
	local svc_obj = _M.all_services[key]
	if not svc_obj or not svc_obj.task_thread then
		return false, "not-exist"
	else
		svc_obj.started = false

		if svc_obj.proxy_thread then
			svc_obj.proxy_thread:Terminate()
			svc_obj.proxy_thread = nil
		end
		return true
	end
end

function _M.UpdateHostItem( name, address, force )
	local changed = false
	local item = _M.host_items[name]
	if not item then
		changed = true
		_M.host_items[name] = address
	elseif force then
		if item ~= address then
			changed = true
			_M.host_items[name] = address
		end
	end
	_M.host_items_changed = _M.host_items_changed or changed
end

function _M.RemoveHostItem( name )
	local changed = false
	if _M.host_items[name] then
		changed = true
		_M.host_items[name] = nil
	end
	_M.host_items_changed = _M.host_items_changed or changed
end

return _M

